Require Import BakerLemma.
Require Import VGC.

Set Implicit Arguments.


(* based on List Processing in Real Time on a Serial Computer by Henry G. Baker *)

(*
Baker compared to Cheney:

- A Baker collector allocates black, at the top of the space.
  B = O U N, where O is old objects, and N is new objects, that
  have been allocated this cycle.  We need this to express the
  isomorphism

- On read, any stale pointers are forwarded and copied if needed

Gray can now contain B,G pointers.  This is the biggest difference.

B, G |- B

B, G, W, M |- G

W, M |- W

phi : M --> (O U G)

*)



Variable scan_per_gc : nat.


(* local register conventions *)
Definition rscan := t6.
Definition rroot := t0.
Definition rgcInfo := s7.
Definition raSave := k0.   (* just temporary.  really should save ra in the heap. *)
Definition rcount := t3.


Definition BAKERLOOP    := 31.
Definition BAKERLOOP2   := 32.
Definition BAKEREXIT    := 33.
Definition RESTOREROOT  := 34.


(* implementation *)


Definition bakerAlloc :=
  (* init gc registers *)
  lw rfrStart  0 rgcInfo;
  lw rfrEnd    4 rgcInfo;
  lw rtoStart  8 rgcInfo;
  lw rtoEnd   12 rgcInfo;
  lw rscan    16 rgcInfo;
  lw rfree    20 rgcInfo;
  lw ralloc   24 rgcInfo;
  addiu rcount zero 0;
  addiu raSave ra 0;

  (* if we haven't run out of space, go directly to the loop *)
  bne rfree ralloc BAKERLOOP;;

  (* if we haven't finished scanning, abort *)
  sltu ast rscan rfree;
  bne ast zero INFLOOP;;

  (* flip *)
  addiu ast rfrStart 0;
  addiu rfrStart rtoStart 0;
  addiu rtoStart ast 0;
  addiu ast rfrEnd 0;
  addiu rfrEnd rtoEnd 0;
  addiu rtoEnd ast 0;

  (* store new space bounds *)
  sw rfrStart  0 rgcInfo;
  sw rfrEnd    4 rgcInfo;
  sw rtoStart  8 rgcInfo;
  sw rtoEnd   12 rgcInfo;

  (* init rfree, rscan, ralloc *)
  addiu rfree rtoStart 0;
  addiu rscan rtoStart 0;
  addiu ralloc rtoEnd 0;

  (* forward root *)
  sw rroot 16 rgcInfo;
  addiu a0 rgcInfo 16;
  jal SCANFIELD RESTOREROOT.

Definition restoreRoot :=
  lw rroot 16 rgcInfo;
  j BAKERLOOP.

Definition bakerLoop :=
  beq rscan rfree BAKEREXIT;;   (* if (rscan == rfree) return *)
  addiu ast zero scan_per_gc;   (* ast := scan_per_gc *)
  sltu ast rcount ast;          (* ast := rcount < scan_per_gc *)
  beq ast zero BAKEREXIT;;      (* if (ast == 0) return *)
  addiu a0 rscan 0;             (* a0 := rscan *)
  jal SCANFIELD BAKERLOOP2.     (* scan first field *)

Definition bakerLoop2 :=
  addiu a0 rscan WORDSIZE;      (* a0 := rscan + 4 *)
  addiu rscan rscan OBJSIZE;    (* rscan := rscan + 8 *)
  addiu rcount rcount 1;        (* rcount := rcount + 1 *)
  jal SCANFIELD BAKERLOOP.      (* scan second field *)

Definition bakerExit :=
  (* if no space left, abort *)
  beq rfree ralloc INFLOOP;;

  (* allocate and initialize object from end of free space *)
  addiu v0 zero 8;
  subu ralloc ralloc v0;
  addiu v0 zero 1;
  sw v0 0 ralloc;
  sw v0 4 ralloc;
  addiu v0 ralloc 0;

  (* store new values for scan, free and alloc *)
  sw rscan  16 rgcInfo;
  sw rfree  20 rgcInfo ;
  sw ralloc 24 rgcInfo ;

  (* return ra and return *)
  addiu ra raSave 0;
  jr ra.




(* specifications *)

Notation Geqv := (fun s => rangeObjs (rget s rscan) (rget s rfree)).


Definition bakerLoopStOk phi W M F A s :=
  rget s rtoStart <= rget s rscan /\
  aligned8 (rget s rscan) (rget s rfree) /\
  scanFieldStOk phi W M F A s.


Definition gcHpPtrMem' base v1 v2 v3 :=
  16 + base |->, (v1::v2::v3::nil).

Definition gcHpPtrMem base :=
  lexists v1, lexists v2, lexists v3,
    gcHpPtrMem' base v1 v2 v3.

Definition bakerLoopPre s :=
  exists phi, exists G, exists W, exists M, exists F,
    set_eqv G (Geqv s) /\
    bakerLoopStOk phi W M F 
      (objHp ((W |_| M) |_| (toSpace s)) G ** gcHpPtrMem (rget s rgcInfo) ** top) s.

Definition bakerLoopGuar s s' :=
  (forall phi G W M F A,
    set_eqv G (Geqv s) ->
    bakerLoopStOk phi W M F 
      (objHp ((W |_| M) |_| (toSpace s)) G ** gcHpPtrMem (rget s rgcInfo) ** A) s ->
    exists phi', exists B, exists B', exists G0, exists G', 
      exists W', exists M', exists F',
	set_eqv W (M' |_| W') /\
	set_eqv F ((B' |_| G') |_| set_add (rget s' ralloc) F') /\
	iso_set phi' M' (B' |_| G') /\
	set_eqv G (B |_| G0) /\
	(set_eqv (B |_| B') (rangeObjs (rget s rscan) (rget s' rscan)) /\
	  rget s rscan <= rget s' rscan) /\
	aligned8 (rget s rfree) (rget s' rfree) /\
	(forall x, set_In x W' -> s' |||- objCopied x s x ** top) /\
	bakerLoopStOk (map_union phi phi') W' (M |_| M') F'
	((foralls x in B, objRespToMap (map_union phi phi') s x) ** 
	  (foralls x in G0, objCopied x s x) **
	  (foralls x in B', objIsToFwded phi phi' s x) **
	  (foralls x in G', objCopiedPhi phi' s x) ** 
	  rget s' v0 |->, (1::1::nil) **
	  gcHpPtrMem' (rget s rgcInfo) (rget s' rscan) (rget s' rfree) (rget s' ralloc) **
	  A) s') /\
  regPres 
  (regNotInList (ralloc::rscan::rfree::rcount::a0::v0::ra::ast::
    rtemp0::rtemp1::rtemp2::raSave0::nil)) s s' /\
  rget s' ra = rget s raSave /\
  rget s ralloc = rget s' ralloc + 8 /\
  rget s' v0 = rget s' ralloc.


Notation baker2Step :=
  (fun s => 
    rset (rset (rset (rset s a0 (4 + rget s rscan)) 
      rscan (8 + rget s rscan)) rcount (1 + rget s rcount)) ra BAKERLOOP).

Definition bakerLoop2Pre s :=
  scanFieldPre (baker2Step s) /\
  forall s', scanFieldGuar (baker2Step s) s' -> bakerLoopPre s'.

Definition bakerLoop2Guar s s'' :=
  exists s', scanFieldGuar (baker2Step s) s' /\ bakerLoopGuar s' s''.

Definition bakerStOk' vh h 
  gcInfo root frStart frEnd toStart toEnd scan free alloc scanh freeh alloch A :=
  exists phi, exists W, exists M,
    aligned8s (toStart::scan::free::alloc::toEnd::nil) /\
    aligned8 frStart frEnd /\
    set_eqv (W |_| M) (rangeObjs frStart frEnd) /\
    let B  := rangeObjs toStart scan in
    let B' := rangeObjs alloc toEnd in
    let G  := rangeObjs scan free in
    let toObjs := B |_| G |_| B' in
      iso_set phi M (B |_| G) /\
      okFieldVal toObjs root /\
      h ||- mapHp M phi ** objHp (W |_| M) W **
	objHp toObjs (B |_| B') ** 
	objHp ((W |_| M) |_| toObjs) G ** 
	(foralls x in rangeObjs free alloc, x |->__) **
	gcInfo |->, (frStart::frEnd::toStart::toEnd::scanh::freeh::alloch::nil) ** A /\
      vh ||- (foralls x in B |_| B', objCopiedH x h x) **
	(foralls x in G |_| W, objRespMapH (map_union phi (Some (A := nat))) h x).


Definition bakerStLoaded vs A s :=
  exists v1, exists v2, exists v3,
    rget vs rroot = rget s rroot /\
    bakerStOk' (heapOf vs) (heapOf s) (rget s rgcInfo) (rget s rroot) 
    (rget s rfrStart) (rget s rfrEnd)
    (rget s rtoStart) (rget s rtoEnd) (rget s rscan) (rget s rfree) (rget s ralloc) v1 v2 v3 A.

Definition bakerStOk vs A s :=
  exists frStart, exists frEnd, exists toStart, exists toEnd, 
    exists scan, exists free, exists alloc,
      rget vs rroot = rget s rroot /\
      bakerStOk' (heapOf vs) (heapOf s) (rget s rgcInfo) (rget s rroot) frStart frEnd 
      toStart toEnd scan free alloc scan free alloc A.

Definition bakerStOkPost vs A s :=
  exists frStart, exists frEnd, exists toStart, exists toEnd, 
    exists scan, exists free, exists alloc,
      let x := rget s v0 in
      let vs' := rset (uhset (uhset vs x 1) (4 + x) 1) v0 x in
      rget vs rroot = rget s rroot /\
      (set_In x (rangeObjs alloc toEnd) /\
	Heap.get (heapOf vs) x = None /\ 
	Heap.get (heapOf vs) (4 + x) = None) /\
      bakerStOk' (heapOf vs') (heapOf s) (rget s rgcInfo) (rget s rroot) frStart frEnd 
      toStart toEnd scan free alloc scan free alloc A.


Notation bakerAllocPre := (fun s => exists vh, bakerStOk vh top s).

Definition bakerVirtGuar (s s':state) := 
  forall objs h,
    s |||- (minObjHp (rget s rroot::nil) objs && Heap.heap_eqv h) ** top ->
    exists objs', exists phi,
      iso_set phi objs objs' /\
      mapX phi (rget s rroot) = Some (rget s' rroot) /\
      s' |||- (foralls x in objs', objIsFwdedH phi h x) ** top.

Definition bakerAllocGuar s s' :=
  (forall vs A,
    bakerStOk vs A s ->
    exists vs',
      bakerVirtGuar vs vs' /\
      bakerStOkPost vs' A s') /\
  calleeSavedOk s s' /\
  rget s ra = rget s' ra.
  (* the usability of this definition relies on implementation details
     being exposed, but that's okay for here. *)

(*
  Definition allocGuar s s' :=
    (forall vs A,
      s |||- repr vs s (rroot::nil) ** A ->
      exists vs', exists x,
	gcGuar vs vs' (rroot::nil) /\
	let vs'' := rset (uhset (uhset vs' x NULL) (4 + x) NULL) v0 x in
	s' |||- repr vs'' s' (v0::rroot::nil) ** A /\
	forall B, vs' |||- B -> vs'' |||- B ** x |->, (NULL::NULL::nil)) /\
    regPres presReg s s'.
*)



Definition restoreRootPre s :=
  exists root,
    s |||- 16 + rget s rgcInfo |-> root ** top /\ 
      exists vh, bakerStLoaded vh top (rset s rroot root).

Definition restoreRootGuar s s' :=
  forall root,
    s |||- 16 + rget s rgcInfo |-> root ** top ->
      (exists vh, bakerStOk vh top s') /\ bakerLoopGuar (rset s rroot root) s'.


Definition bakerExitPre (s:state) := 
  let free := rget s rfree in
  let alloc := rget s ralloc in
    aligned8 free alloc /\
    s |||- gcHpPtrMem (rget s rgcInfo) ** 
      (foralls x in rangeObjs free alloc, x |->__) ** top.


Definition bakerExitGuar (s s':state) :=
  let free := rget s rfree in
  let alloc := rget s ralloc in
  let alloc' := rget s' ralloc in
    (forall A,
      s |||- gcHpPtrMem (rget s rgcInfo) ** (foralls x in rangeObjs free alloc, x |->__) ** A ->
      s' |||- gcHpPtrMem' (rget s rgcInfo) (rget s rscan) (rget s rfree) alloc' ** 
	alloc' |->, (1::1::nil) ** (foralls x in rangeObjs free alloc', x |->__) ** A) /\
  regPres (regNotInList (ralloc::v0::ra::nil)) s s' /\
  alloc = alloc' + 8 /\
  rget s' v0 = alloc' /\
  rget s' ra = rget s raSave /\
  aligned8 free alloc'.


Definition Pbaker : chspec := fun l => 
  if beq_nat l SCANFIELD then
    SomeT (bspeci scanFieldPre scanFieldGuar)
  else if beq_nat l BAKERLOOP then
    SomeT (bspeci bakerLoopPre bakerLoopGuar)
  else if beq_nat l BAKERLOOP2 then
    SomeT (bspeci bakerLoop2Pre bakerLoop2Guar)
  else if beq_nat l BAKEREXIT then
    SomeT (bspeci bakerExitPre bakerExitGuar)
  else if beq_nat l RESTOREROOT then
    SomeT (bspeci restoreRootPre restoreRootGuar)
  else if beq_nat l INFLOOP then
    SomeT (bspeci infLoopPre infLoopGuar)
  else
    NoneT _.


Ltac PbakerSimpl := 
  unfold Pbaker; 
  rewrite beq_true; 
  repeat (rewrite beq_false; [idtac | discriminate]); 
  reflexivity.

Ltac destrBL H := 
  let Le0 := fresh "Le0" in
  let Al1 := fresh "Al1" in
  let Sf := fresh "Sf" in
    (destruct H as [Le0 [Al1 Sf]];
     destrSF Sf).


(* verifying the specs holds *)


Lemma enterLoopOk' : forall vh A s,
  bakerStLoaded vh A s -> 
  exists phi, exists G, exists W, exists M,
    set_eqv G (Geqv s) /\
    set_eqv (W |_| M) (rangeObjs (rget s rfrStart) (rget s rfrEnd)) /\
    bakerLoopStOk phi W M (rangeObjs (rget s rfree) (rget s ralloc)) 
    (objHp ((W |_| M) |_| (toSpace s)) G ** gcHpPtrMem (rget s rgcInfo) ** 
      rget s rgcInfo |->, (rget s rfrStart::rget s rfrEnd::rget s rtoStart::rget s rtoEnd::nil) **
      (let B := rangeObjs (rget s rtoStart) (rget s rscan) in
      let G := rangeObjs (rget s rscan) (rget s rfree) in
      let B' := rangeObjs (rget s ralloc) (rget s rtoEnd) in
	objHp (B |_| G |_| B') (B |_| B')) ** A) s /\
    iso_set phi M ((rangeObjs (rget s rtoStart) (rget s rscan)
           |_| rangeObjs (rget s rscan) (rget s rfree))).
Proof.
  intros vh A s [scan [free [alloc Hp]]].
  destruct Hp as [rootEq [phi [W [M [Al1 [Al2 [WMeq [iso [rootOk [Hp _]]]]]]]]]].
  exists phi; exists (rangeObjs (rget s rscan) (rget s rfree)); exists W; exists M.
  constructor.
  apply set_eqvRefl.
  constructor.
  assumption.
  simpl in Al1; destruct Al1 as [Alstsc [Alscfr [Alfral Alalend]]].
  constructor.
  2:assumption.
  constructor.
  destruct Alstsc; omega.
  constructor.
  assumption.

  assert (EqEq : set_eqv (rangeObjs (rget s rtoStart) (rget s rtoEnd))
    (rangeObjs (rget s rtoStart) (rget s rscan) |_| 
      rangeObjs (rget s rscan) (rget s rfree) |_|
	rangeObjs (rget s rfree) (rget s ralloc) |_|
	  rangeObjs (rget s ralloc) (rget s rtoEnd))).
  rewrite <- (rangeObjsSplit Alfral Alalend).
  assert (ZZ : aligned8 (rget s rfree) (rget s rtoEnd)).
  eapply aligned8Trans; eauto.
  rewrite <- (rangeObjsSplit Alscfr ZZ).
  clear ZZ; assert (ZZ : aligned8 (rget s rscan) (rget s rtoEnd)).
  eapply aligned8Trans; eauto.
  eapply aligned8Trans; eauto.
  rewrite <- (rangeObjsSplit Alstsc ZZ).
  apply set_eqvRefl.

  constructor.
  unfold fromSpace; unfold toSpace.
  apply disjObjsDisjRange with (h := heapOf s).
  hypReplace Hp (sconjAssocR Hp).
  apply (sconjImpl Hp); clear Hp; intros h Hp.
  unfold hpredTrue; rewrite <- WMeq.
  apply fasMerge'.
  apply sconjComm; sconjMatch.
  unfold mapHp in Hp; linLogElim Hp.
  eapply fasImpl; eauto.
  simpl; intros x h' Hp'; linLogElim Hp'; linLogIntro; sconjMatch.
  unfold objHp in Hp.
  eapply fasImpl; eauto.
  simpl; unfold okObjAt; unfold okFieldAt; intros x h' Hp'; linLogElim Hp'; linLogIntro; sconjMatch.
  intros x h' Hp'; linLogElim Hp'; linLogIntro; sconjMatch.

  hypReplace Hp (sconjAssocR Hp).
  hypReplace Hp (sconjAssocR Hp).
  sconjMatch.
  unfold objHp in Hp; linLogElim Hp.
  unfold hpredTrue; setoid_rewrite EqEq.

  assert (Hp' : h ||- (foralls x in rangeObjs (rget s rtoStart) (rget s rscan)
    |_| rangeObjs (rget s ralloc) (rget s rtoEnd), x |->__) **
  (foralls x in rangeObjs (rget s rscan) (rget s rfree), x |->__) **
  (foralls x in rangeObjs (rget s rfree) (rget s ralloc), x |->__)).
  sconjMatch.
  eapply fasImpl; eauto.
  simpl; unfold okObjAt; unfold okFieldAt; intros x h' Hp'; linLogElim Hp'; linLogIntro; sconjMatch.
  eapply fasImpl; eauto.
  simpl; unfold okObjAt; unfold okFieldAt; intros x h' Hp'; linLogElim Hp'; linLogIntro; sconjMatch.
  clear Hp; rename Hp' into Hp.
  setoid_replace (rangeObjs (rget s rtoStart) (rget s rscan)
       |_| rangeObjs (rget s rscan) (rget s rfree)
           |_| rangeObjs (rget s rfree) (rget s ralloc)
               |_| rangeObjs (rget s ralloc) (rget s rtoEnd)) with
  ((rangeObjs (rget s rtoStart) (rget s rscan) |_| rangeObjs (rget s ralloc) (rget s rtoEnd))
    |_| (rangeObjs (rget s rscan) (rget s rfree)) |_| (rangeObjs (rget s rfree) (rget s ralloc))).
  apply fasMerge'.
  sconjMatch.
  apply fasMerge'.
  assumption.
  intros x h' Hp'; linLogElim Hp'; linLogIntro; sconjMatch.
  intros x h' Hp'; linLogElim Hp'; linLogIntro; sconjMatch.

  constructor; intro I.
  destruct (set_union_elim I) as [I' | I']; clear I.
  sui1; sui1; auto.
  destruct (set_union_elim I') as [I | I]; clear I'.
  sui2; sui1; auto.
  destruct (set_union_elim I) as [I' | I']; clear I.
  sui2; sui2; auto.
  sui1; sui2; auto.
  destruct (set_union_elim I) as [I' | I']; clear I.
  destruct (set_union_elim I') as [I | I]; clear I'.
  sui1; auto.
  sui2; sui2; sui2; auto.
  destruct (set_union_elim I') as [I | I]; clear I'.
  sui2; sui1; auto.
  sui2; sui2; sui1; auto.
  
  assumption.
  eapply aligned8Trans; eauto.
  eapply aligned8Trans; eauto.
  eapply aligned8Trans; eauto.

  constructor.
  unfold fromSpace; setoid_rewrite WMeq.
  apply objRangeSetRangeSubset.
  constructor.
  apply set_eqvRefl.

  constructor.
  intros x y Peq NSP.
  destruct iso as [_ [iso _]].
  assert (Z := iso _ _ Peq).
  unfold notSpacePtrP in NSP.
  destruct NSP.
  assert (Hp' : s |||- ((!ptrP y ** top) ** top) ** top).
  destruct (set_union_elim Z) as [I | I]; clear Z.
  sconjLift3 Hp; sconjMatch.
  assert (I' : set_In y (rangeObjs (rget s rtoStart) (rget s rscan)
    |_| rangeObjs (rget s ralloc) (rget s rtoEnd))).
  apply set_union_intro1; assumption.
  hypReplace Hp (objHpElim I' Hp).
  unfold okObjAt in Hp; unfold okFieldAt in Hp; linLogElim Hp; sconjMatch.
  linLogIntro; eapply contElim1; eauto.
  sconjLift4 Hp; sconjMatch.
  hypReplace Hp (objHpElim I Hp).
  unfold okObjAt in Hp; unfold okFieldAt in Hp; linLogElim Hp; sconjMatch.
  linLogIntro; eapply contElim1; eauto.
  linLogElim Hp'.
  apply (not_even_and_odd y).
  apply ptrPeven; assumption.
  assumption.
  apply H.
  clear rootOk Hp.
  apply setRangeIntro.
  destruct Alstsc; destruct Alscfr; destruct Alfral; destruct Alalend.
  destruct (set_union_elim Z) as [I | I]; clear Z.
  destruct (rangeObjsElim I) as [I1 [k I2]]; clear I.
  omega.
  destruct (rangeObjsElim I) as [I1 [k I2]]; clear I.
  omega.

  constructor.
  destruct Alstsc; destruct Alscfr; destruct Alfral; destruct Alalend; omega.
  constructor.
  destruct Alstsc; destruct Alscfr; destruct Alfral; destruct Alalend; omega.
  constructor.
  assumption.

  apply (sconjImplR Hp); clear Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift3 Hp; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift2 Hp.
  clear rootOk.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  apply (objHpWeaken Hp).
  intros x I.
  destruct (set_union_elim I) as [I' | I']; clear I.
  apply set_union_intro1; assumption.
  apply set_union_intro2.
  unfold toSpace.
  clear Hp.

  apply objRangeSetRangeSubset.
  setoid_rewrite EqEq.
  destruct (set_union_elim I') as [I | I]; clear I'.
  apply set_union_intro1; assumption.
  apply set_union_intro2.
  destruct (set_union_elim I) as [I' | I']; clear I.
  apply set_union_intro1; assumption.
  apply set_union_intro2; apply set_union_intro2; assumption.

  unfold gcHpPtrMem; unfold gcHpPtrMem'; linLogIntro.
  instantiate (1 := alloc).
  instantiate (1 := free).
  instantiate (1 := scan).
  sconjLift2 Hp; linLogElim Hp.
  sconjLift7 Hp; sconjLift7 Hp; sconjLift7 Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  assumption.

Qed.


Lemma enterLoopOk : forall vh A s,
  bakerStLoaded vh A s -> 
  exists phi, exists G, exists W, exists M, exists h,
    set_eqv G (Geqv s) /\
    set_eqv (W |_| M) (rangeObjs (rget s rfrStart) (rget s rfrEnd)) /\
    bakerLoopStOk phi W M (rangeObjs (rget s rfree) (rget s ralloc))
    (objHp ((W |_| M) |_| (toSpace s)) G ** gcHpPtrMem (rget s rgcInfo) ** 
      heap_eqv h ** A) s /\
    h ||-
      rget s rgcInfo |->, (rget s rfrStart::rget s rfrEnd::rget s rtoStart::rget s rtoEnd::nil) **
      (let B := rangeObjs (rget s rtoStart) (rget s rscan) in
      let G := rangeObjs (rget s rscan) (rget s rfree) in
      let B' := rangeObjs (rget s ralloc) (rget s rtoEnd) in
	objHp (B |_| G |_| B') (B |_| B')) /\
    iso_set phi M (rangeObjs (rget s rtoStart) (rget s rscan)
           |_| rangeObjs (rget s rscan) (rget s rfree)).
Proof.
  intros vh A s H.
  destruct (enterLoopOk' H) as [phi [G [W [M [Geqv [WMeqv [H' Iso]]]]]]]; clear H.
  exists phi; exists G; exists W; exists M.
  unfold bakerLoopStOk in H'; intuition.
  destrSF H2.
  sconjLift7 Hp; sconjLift7 Hp.
  hypReplace Hp (sconjAssocR Hp).
  destruct (sconjExtract Hp) as [h [Hp1 Hp2]]; clear Hp.
  exists h.
  constructor.
  assumption.
  unfold bakerLoopStOk; unfold scanFieldStOk; intuition.
  clear Hp2 Le1 Le2 NSP Feq H H1 WMdisj WMsub Le3 AlF Geqv.
  sconjLift6 Hp1; sconjLift6 Hp1; sconjLift6 Hp1; sconjLift6 Hp1; sconjLift6 Hp1;
  assumption.
Qed.


Lemma restoreRootOk : iseqOkS Pbaker restoreRootPre restoreRootGuar restoreRoot.
Proof.
  unfold restoreRoot.
  iseqWeakSolver PbakerSimpl.

  unfold restoreRootPre.
  intros s [root [Hp [vh Pre]]].
  rewrite (lwStep Hp); try reflexivity.
  assert (X := enterLoopOk Pre); generalize dependent X; stSimpl; intro;
    destruct X as [phi [G [W [M [h [Geqv [WMeqv [bOk [Hok Iso]]]]]]]]].
  set (F := rangeObjs (rget s rfree) (rget s ralloc)) in * .
  constructor.

  (* show the precondition *)
  exists phi; exists G; exists W; exists M; exists F; constructor.
  assumption.
  unfold bakerLoopStOk, scanFieldStOk in bOk |- *; intuition.
  scm.

  (* show the guar *)
  unfold restoreRootGuar.
  intros s' Guar root' Hp'.
  assert (X := contContEq Hp Hp'); clear Hp'; subst root'.
  constructor.
  2:assumption.
  destruct Guar as [Guar [Req [raEq [allocEq v0Eq]]]].
  generalize dependent allocEq; stSimpl; intros.
  destruct (Guar phi G W M F (Heap.heap_eqv h ** top)) as 
    [phi' [B' [B'' [G' [G'' [W' [M' [F' [W'eq [F'eq [Iso' [G'eq X]]]]]]]]]]]]; clear Guar.
  assumption.
  unfold bakerLoopStOk in bOk |- *; unfold scanFieldStOk in bOk |- *; intuition.

  generalize dependent X; stSimpl; intro.
  destruct X as [[BBeq scLE] [AlF [Wcopy Bok]]].
  generalize dependent Hok; stSimpl; intro.
  clear raEq.
  unfold bakerLoopStOk, scanFieldStOk in Bok;
    generalize dependent Bok; unfold fromSpace, toSpace.
  rewrite <- Req; [stSimpl | regNotInListSolver].
  rewrite <- (Req rfrStart); [stSimpl | regNotInListSolver].
  rewrite <- (Req rfrEnd); [stSimpl | regNotInListSolver].
  rewrite <- (Req rtoEnd); [stSimpl | regNotInListSolver].
  rewrite allocEq in Hok.
  intros [scLt [Al' [frToDisj [WMM'sub [F'eq' [NSP [Le1 [Le2 [Al2 Hp'']]]]]]]]].

  unfold bakerStOk.
  sconjLift10 Hp''; hyr Hp'' (sconjAssocL (sconjInsert Hp'' Hok)); clear Hok.
  sconjLift10 Hp''; sconjLift10 Hp''; sconjLift10 Hp''; 
  sconjLift10 Hp''; sconjLift10 Hp''; sconjLift9 Hp''.
  sconjLift8 Hp''.
  hyr Hp'' (sconjAssocR (sconjAssocR Hp''));
  hyr Hp'' (sconjAssocR (sconjAssocR Hp''));
  hyr Hp'' (sconjAssocR (sconjAssocR Hp'')).
  destruct (sconjExtract Hp'') as [h' [HpA HpV]]; clear Hp''; rename HpA into Hp''.
  set (vh' := Heap.HeapMap.map_mapRange 
    (fun x => match map_union phi phi' x with Some y => y | _ => x end) h').
  exists (vh', rfileOf s').
  lle HpV.
  exists (rget s rfrStart); exists (rget s rfrEnd);
    exists (rget s rtoStart); exists (rget s rtoEnd);
      exists (rget s' rscan); exists (rget s' rfree); exists (rget s' ralloc).
  constructor.
  reflexivity.
  exists (map_union phi phi'); exists W'; exists (M |_| M').
  constructor.
  unfold bakerStLoaded, bakerStOk' in Pre; 
    generalize dependent Pre; stSimpl; intros [v1 [v2 [v3 [_ [phi0 [W0 [M0 X]]]]]]].
  simpl; intuition; simpl in H1; intuition.
  destruct Al' as [k1 Al']; destruct AlF as [k2 AlF];
    destruct H1 as [k3 H1]; destruct H7 as [k4 H7].
  exists ((k4 + k3 + k2) - k1).
  replace (rget s rtoStart + 8 * ((k4 + k3 + k2) - k1)) with
    ((((rget s rtoStart + 8 * k4) + 8 * k3) + 8 * k2) - 8 * k1).
  2:omega.
  rewrite <- H7; rewrite <- H1; rewrite <- AlF.
  rewrite Al'.
  omega.
  eapply aligned8Trans; eauto.
  rewrite allocEq; exists 1; omega.
  constructor.
  unfold bakerStLoaded, bakerStOk' in Pre; 
    generalize dependent Pre; stSimpl; intros [v1 [v2 [v3 [_ [phi0 [W0 [M0 X]]]]]]].
  intuition.
  constructor.
  rewrite (set_eqvUnionSymm M M'); rewrite <- (set_eqvUnionAssoc W' M' M);
    rewrite (set_eqvUnionSymm W' M'); rewrite <- W'eq; assumption.
  intros B0 B0' G0 toObjs.
  simpl in HpV; set (B := rangeObjs (rget s rtoStart) (rget s rscan)) in * ;
  set (Gz := rangeObjs (rget s rscan) (rget s rfree)) in * ;
  set (Bz' := rangeObjs (rget s ralloc) (rget s rtoEnd)) in * .
  subst B0'.

  assert (BBBeq : set_eqv B0 (B |_| B' |_| B'')).
  subst B0 B; rewrite BBeq.
  unfold bakerStLoaded in Pre; unfold bakerStOk' in Pre; 
    generalize dependent Pre; stSimpl; intros [v1 [v2 [v3 [_ [phi0 [W0 [M0 [X [Y _]]]]]]]]].
  assert (AL : aligned8 (rget s rtoStart) (rget s rscan)).
  simpl in X; intuition.
  rewrite (rangeObjsTransEqv AL scLE).
  apply set_eqvRefl.

  assert (FalEq : set_eqv (set_add (rget s' ralloc) F') 
    (rangeObjs (rget s' rfree) (rget s ralloc))).
  rewrite F'eq'.
  apply set_eqvSymm.
  replace (rget s' ralloc) with (rget s ralloc - 8).
  apply rangeObjsEndSplit.
  rewrite allocEq; destruct Al2 as [k Al2]; exists (S k); omega.
  destruct Al2; omega.
  rewrite allocEq; omega.

  assert (GGeq : set_eqv G0 (G' |_| G'')).
  assert (GGeq : set_eqv G0 ((set_diff G (B' |_| B'')) |_| G'')).
  assert (X : aligned8 (rget s rscan) (rget s rfree)).
  destruct Pre as [v1 [v2 [v3 [_ [phi0 [W0 [M0 [X [Y _]]]]]]]]].  
  simpl in X; intuition.
  clear Req.
  subst G0; rewrite Geqv; subst Gz; eapply grayObjExtract; auto.
  instantiate (1 := rget s ralloc).
  instantiate (1 := set_add (rget s' ralloc) F').
  subst F; rewrite F'eq; apply set_eqvRefl.
  assumption.
  apply aligned8LE_R with (z := rget s' rfree); auto;
    eapply aligned8Trans; eauto.
  rewrite allocEq; destruct Al2 as [k Al2]; exists (S k); omega.

  assert (HX : h' ||- top ** top ** fasd B' ** top ** fasd B'' ** fasd G'' ** top).
  scm; apply (fasImpl Hp0); uilleim.
  lle HX; hyr HX (sconjAssocR HX); scm.
  assert (HX' : h' ||- fasd (B' |_| B'') ** fasd G'' ** top).
  lle HX; hyr HX (sconjAssocR HX); scm.
  apply fasdMerge; assumption.
  sconjLift2 HX'; apply (fasFasDisj HX'); auto.
  assert (HX : h' ||- top ** top ** top ** top ** top ** fasd G'' ** 
    fasd (set_add (rget s' ralloc) empty_set)).
  scm.
  apply (fasImpl Hp0); uilleim.
  apply fasIntro; hyr Hp0 (unitIR Hp0); scm; rewrite <- v0Eq; hpunfoldIn Hp0; lli; scm.
  lle HX; hyr Hp'' (sconjInsert Hp'' HX); sconjLift4 Hp''; lle Hp''; sconjLift3 Hp''.
  assert (HpX : s' |||- fasd (set_add (rget s' ralloc) F') ** fasd G'' ** top).
  hyr Hp'' (sconjAssocR Hp''); scm.
  assert (HpX : h0 ||- (rget s' ralloc |->_ ** top) ** (foralls x in F', x |->__)).
  scm.
  assert (II : set_In (rget s' ralloc) (set_add (rget s' ralloc) empty_set)).
  left; reflexivity.
  hyr Hp0 (fasElim II Hp0); lleim.
  clear Hp0; apply fasIntro; scm.
  llMatch; lleim.
  intro I; assert (HpY : h0 ||- (rget s' ralloc |->_ ** top) ** (rget s' ralloc |->_ ** top)).
  scm.
  hyr Hp0 (fasElim I Hp0); lleim.
  lle HpY.
  sconjLift2 HpX; apply (fasFasDisj HpX); illeim.
  setoid_replace G' with (set_diff G (B' |_| B'')).
  assumption.
  rewrite G'eq; rewrite (set_diffUnionDistrib B' G' (B' |_| B''));
    rewrite (setUnionDiffDiff B' B' B''); rewrite (set_diffSelf B').
  assert (Disj : set_disj G' (B' |_| B'')).
    assert (HX : h' ||- top ** top ** fasd B' ** fasd G' ** fasd B'' ** top).
  scm; apply (fasImpl Hp0); uilleim.
  assert (HX' : h' ||- fasd (B' |_| B'') ** fasd G' ** top).
  lle HX; sconjLift3 HX; sconjLift2 HX; hyr HX (sconjAssocR HX); scm.
  apply fasdMerge; assumption.
  sconjLift2 HX'; apply (fasFasDisj HX'); auto.
  rewrite (set_diffDisj Disj).
  simpl; apply set_eqvRefl.

  assert (BG0eqv : set_eqv (B0 |_| G0) ((B |_| G) |_| B'' |_| G'')).
  rewrite GGeq; rewrite BBBeq.
  rewrite (set_eqvUnionSymm B' B''); rewrite <- (set_eqvUnionAssoc B B'' B');
    rewrite (set_eqvUnionAssoc (B |_| B'') B' (G' |_| G'')); 
      rewrite <- (set_eqvUnionAssoc B' G' G''); rewrite <- G'eq.
  rewrite (set_eqvUnionAssoc B B'' (G |_| G''));
    rewrite (set_eqvUnionSymm G G''); rewrite <- (set_eqvUnionAssoc B'' G'' G);
      rewrite (set_eqvUnionSymm (B'' |_| G'') G); rewrite <- (set_eqvUnionAssoc B G (B'' |_| G''));
	apply set_eqvRefl.

  assert (IsoU : iso_set (map_union phi phi') (M |_| M') (B0 |_| G0)).
  rewrite BG0eqv; rewrite Geqv; apply iso_union; auto.
  (* set_disj M M' *)
  assert (Disj : set_disj M W).
  destrBL bOk.
  eapply fasFasDisj.
  apply (sconjImpl Hp0).
  intros hh HHp; apply (mapHpIsObjs HHp).
  instantiate (1 := top).
  intros hh HHp; scm; apply (objHpIsObjs Hp1).
  cbv beta; illeim.
  cbv beta; illeim.
  apply set_disjSymm.
  apply set_unionDisjElimL with W'.
  rewrite <- W'eq; apply set_disjSymm; assumption.
  (* set_disj (B |_| G) (B'' |_| G'') *)
  assert (Disj : set_disj B (B'' |_| G'')).
  apply set_unionDisjElimL with Bz'.
  assert (HX : h' ||- fasd (B |_| Bz') ** top ** top ** top ** fasd B'' ** fasd G'' ** top).
  rewrite <- allocEq in HpV; unfold objHp in HpV; llMatch;
    hpunfoldIn Hp0; lleim.
  assert (HX' : h' ||- fasd (B'' |_| G'') ** fasd (B |_| Bz') ** top).
  lle HX; sconjLift3 HX; sconjLift3 HX; hyr HX (sconjAssocR HX); scm.
  apply fasdMerge; assumption.
  sconjLift2 HX'; apply (fasFasDisj HX'); auto.
  rewrite <- Geqv.
  assert (Disj' : set_disj G (B'' |_| G'')).
  rewrite G'eq.
  assert (HX : h' ||- top ** top ** fasd B' ** fasd G' ** fasd B'' ** fasd G'' ** top).
  scm; unfold objHp in Hp0; apply (fasImpl Hp0); uilleim.
  assert (HX' : h' ||- fasd (B'' |_| G'') ** fasd (B' |_| G') ** top).
  lle HX; hyr HX (sconjAssocR HX); sconjLift3 HX; sconjLift3 HX; 
    hyr HX (sconjAssocR HX); scm; apply fasdMerge; assumption.
  sconjLift2 HX'; apply (fasFasDisj HX'); auto.
  apply set_disj_intro; intros x Ix Ix'; destruct (set_union_elim Ix).
  apply (set_disj_elim Disj x); assumption.
  apply (set_disj_elim Disj' x); assumption.

  constructor.
  assumption.
  constructor.
  rewrite <- Req; [stSimpl | regNotInListSolver];
    destruct Pre as [v1 [v2 [v3 [_ [phi0 [W0 [M0 [_ [_ [_ [_ [X _]]]]]]]]]]]].
  destruct X as [okF | okF]; [left | right].
  assumption.
  generalize dependent okF; stSimpl; intro.
  subst toObjs; rewrite <- (set_eqvUnionAssoc B0 G0 (rangeObjs (rget s' ralloc) (rget s rtoEnd))); 
    rewrite BG0eqv.
  replace (rangeObjs (rget s rtoStart) (rget s rscan)) with B in okF; [idtac | reflexivity].
  replace (rangeObjs (rget s ralloc) (rget s rtoEnd)) with Bz' in okF; [idtac | reflexivity].
  rewrite (set_eqvUnionAssoc B G (B'' |_| G'')).
  rewrite (set_eqvUnionAssoc B (G |_| B'' |_| G'') (rangeObjs (rget s' ralloc) (rget s rtoEnd))).
  destruct (set_union_elim okF) as [I | I]; clear okF.
  sui1; assumption.
  sui2.
  destruct (set_union_elim I) as [I' | I']; clear I.
  sui1; sui1; rewrite Geqv; assumption.
  sui2; subst Bz'; rewrite allocEq in I'; set_InElim;
    apply rangeObjsIntro with (S k); omega.

  rewrite <- (Req rgcInfo); [stSimpl | regNotInListSolver];
    clear Req; rename Bz' into Ba.

  assert (toEq : set_eqv (rangeObjs (rget s rtoStart) (rget s rtoEnd)) ((B |_| G) |_| F |_| Ba)).
  rewrite Geqv; subst Gz F B Ba.
  destruct Pre as [v1 [v2 [v3 [_ [phi0 [W0 [M0 [X [Y _]]]]]]]]]; 
    generalize dependent X; stSimpl; simpl; intuition.
  assert (LE1 : rget s rscan <= rget s rfree).
  destruct H3; omega.
  rewrite (rangeObjsTransEqv H1 LE1).
  assert (X : aligned8 (rget s rtoStart) (rget s rfree)).
  eapply aligned8Trans; eauto.
  assert (LE : rget s rfree <= rget s ralloc).
  destruct H2; omega.
  rewrite <- (set_eqvUnionAssoc (rangeObjs (rget s rtoStart) (rget s rfree))
    (rangeObjs (rget s rfree) (rget s ralloc)) (rangeObjs (rget s ralloc) (rget s rtoEnd))).
  rewrite (rangeObjsTransEqv X LE).
  assert (Le' : rget s ralloc <= rget s rtoEnd).
  destruct H5; omega.
  assert (AlX : aligned8 (rget s rtoStart) (rget s ralloc)).
  eapply aligned8Trans; eauto.
  rewrite (rangeObjsTransEqv AlX Le'); apply set_eqvRefl.

  set (Bz'' := rangeObjs (rget s' ralloc) (rget s rtoEnd)) in * .

  assert (SubBGT : set_subset (B0 |_| G0) toObjs).
  subst toObjs; rewrite <- (set_eqvUnionAssoc B0 G0 Bz''); intros x Ix; 
    sui1; assumption.
  
  assert (HpV' : h' ||- objHp toObjs (B0 |_| Bz'') **
    objHp (W' |_| M |_| M') W' ** objHp ((W' |_| M |_| M') |_| toObjs) G0).
  unfold objHp; sconjLift7of7 HpV; sconjLift4 HpV; sconjLift6 HpV;
    hyr HpV (sconjAssocR (sconjAssocR (sconjAssocR HpV))); scm.
   unfold hpredTrue; rewrite BBBeq; rewrite (set_eqvUnionSymm B (B' |_| B''));
    rewrite (set_eqvUnionAssoc (B' |_| B'') B Bz''); apply objHpMerge.
   hyr Hp0 (sconjAssocL Hp0); scm.
  clear Hp; apply objHpMerge; sconjLift2of2 Hp0; unfold objHp; hpunfoldIn Hp0; hpunfold; llMatch; lle Hp.
  (* B' *)
  assert (I' : set_In x G).
  rewrite G'eq; sui1; assumption.
  assert (HH : s |||- (okObjAt ((W |_| M) |_| B |_| G |_| Ba) x ** top) ** top).
  destruct Pre as [v1 [v2 [v3 [_ [phi0 [W0 [M0 [_ [_ [WMeqv0 [_ [_ [X _]]]]]]]]]]]]].
  generalize dependent WMeqv0; generalize dependent X; stSimpl; intros.
  sconjLift4 X; scm.
  assert (I'' : set_In x Gz).
  rewrite <- Geqv; assumption.
  hyr Hp0 (objHpElim I'' Hp0); scm.
  unfold hpredTrue; rewrite WMeqv; rewrite <- WMeqv0.
  rewrite Geqv; assumption.
  hpunfoldIn HH; lle HH.
  destruct H3 as [v0 [Hp0 Meq0]]; destruct H4 as [v1 [Hp1 Meq1]].
  assert (X1 := contContEq Hp0 HH); sconjLift2 HH; assert (X2 := contContEq Hp1 HH);
    clear HH Hp0 Hp1; subst x2 x3.
  unfold mapX in Meq0, Meq1; unfold map_unionnn in Meq0, Meq1;
    unfold map_union at 1 in Meq0, Meq1.
  lli.
  2:apply Hp.
  (* first field *)
  clear x I I' Hp; clear v1 Meq1 H8; rename H5 into okF.
  assert (IO := isOddOk v0); destruct (isOdd v0).
  injection Meq0; intro; subst x0; left; assumption.
  destruct okF as [okF | okF].
  elf; apply (not_even_and_odd v0); assumption.
  replace (rget (rset s rroot root) rtoStart) with (rget s rtoStart) in Meq0; [idtac | reflexivity].
  replace (rget (rset s rroot root) rtoEnd) with (rget s rtoEnd) in Meq0; [idtac | reflexivity].
  assert (X := set_memSpaceIdMap (rget s rtoStart) (rget s rtoEnd) v0).
  destruct (set_union_elim okF) as [I | I]; clear okF.
  rewrite set_mem_false in X.
  rewrite X in Meq0.
  destruct IsoU as [_ [IsoU _]]; hyr IsoU (IsoU _ _ Meq0).
  subst toObjs; rewrite <- (set_eqvUnionAssoc B0 G0 Bz''); right; sui1; assumption.
  intro I'; apply (set_disj_elim frToDisj v0).
  apply objRangeSetRangeSubset; rewrite <- WMeqv; assumption.
  assumption.
  rewrite set_mem_true in X.
  rewrite X in Meq0.
  injection Meq0; clear Meq0 X; intro; subst x0.
  right; subst toObjs; rewrite <- (set_eqvUnionAssoc B0 G0 Bz''); rewrite BG0eqv; set_InElim.
  sui1; sui1; sui1; assumption.
  sui1; sui1; sui2; assumption.
  sui2; subst Ba Bz''; set_InElim; apply rangeObjsIntro with (S k); omega.
  apply objRangeSetRangeSubset; rewrite toEq; set_InElim.
  sui1; sui1; assumption.
  sui1; sui2; assumption.
  sui2; sui2; assumption.
  (* field 2 *)
  clear x I I' Hp H; clear x0 v0 Meq0 H5; rename H8 into okF; 
    rename v1 into v0; rename Meq1 into Meq0; rename x1 into x0.
  assert (IO := isOddOk v0); destruct (isOdd v0).
  injection Meq0; intro; subst x0; left; assumption.
  destruct okF as [okF | okF].
  elf; apply (not_even_and_odd v0); assumption.
  replace (rget (rset s rroot root) rtoStart) with (rget s rtoStart) in Meq0; [idtac | reflexivity].
  replace (rget (rset s rroot root) rtoEnd) with (rget s rtoEnd) in Meq0; [idtac | reflexivity].
  assert (X := set_memSpaceIdMap (rget s rtoStart) (rget s rtoEnd) v0).
  destruct (set_union_elim okF) as [I | I]; clear okF.
  rewrite set_mem_false in X.
  rewrite X in Meq0.
  destruct IsoU as [_ [IsoU _]]; hyr IsoU (IsoU _ _ Meq0).
  subst toObjs; rewrite <- (set_eqvUnionAssoc B0 G0 Bz''); right; sui1; assumption.
  intro I'; apply (set_disj_elim frToDisj v0).
  apply objRangeSetRangeSubset; rewrite <- WMeqv; assumption.
  assumption.
  rewrite set_mem_true in X.
  rewrite X in Meq0.
  injection Meq0; clear Meq0 X; intro; subst x0.
  right; subst toObjs; rewrite <- (set_eqvUnionAssoc B0 G0 Bz''); rewrite BG0eqv; set_InElim.
  sui1; sui1; sui1; assumption.
  sui1; sui1; sui2; assumption.
  sui2; subst Ba Bz''; set_InElim; apply rangeObjsIntro with (S k); omega.
  apply objRangeSetRangeSubset; rewrite toEq; set_InElim.
  sui1; sui1; assumption.
  sui1; sui2; assumption.
  sui2; sui2; assumption.
  (* B'' *)
  assert (HH : s |||- (okObjAt (W |_| M) x0 ** top) ** top).
  assert (I' : set_In x0 W).
  rewrite W'eq; sui1; 
    destruct Iso' as [Iso' _]; destruct (Iso' x0) as [_ Q]; apply Q;
      rewrite H1; discriminate.
  destrBL bOk; sconjLift2 Hp0; scm;
  hyr Hp0 (objHpElim I' Hp0); scm.
  hpunfoldIn HH; lle HH.
  destruct H6 as [v0 [Hp0 Meq0]]; destruct H7 as [v1 [Hp1 Meq1]].
  assert (X1 := contContEq Hp0 HH); sconjLift2 HH; assert (X2 := contContEq Hp1 HH);
    clear HH Hp0 Hp1; subst x3 x4.
  lli.
  2:apply Hp.
  (* first field *)
  clear x0 x Hp H1 I; clear x2 v1 Meq1 H9; rename H4 into okF.
  assert (Meq : mapX (map_union (spaceIdMap (rget s rtoStart) (rget s rtoEnd)) 
    (map_union phi phi')) v0 = Some x1).
  unfold mapX in * ; unfold map_union in * .
  destruct (isOdd v0).
  assumption.
  generalize dependent Meq0; stSimpl; intro;
    destruct (spaceIdMap (rget s rtoStart) (rget s rtoEnd) v0); assumption.
  clear Meq0; unfold mapX in Meq; unfold map_union at 1 in Meq.
  assert (IO := isOddOk v0); destruct (isOdd v0).
  injection Meq; intro; subst x1; left; assumption.
  destruct okF as [okF | okF].
  elf; apply (not_even_and_odd v0); assumption.
  rewrite spaceIdMapNone in Meq.
  right; apply SubBGT; destruct IsoU as [_ [IsoU _]]; eapply IsoU; eauto.
  intros I; apply (set_disj_elim frToDisj v0).
  apply objRangeSetRangeSubset; rewrite <- WMeqv; assumption.
  assumption.
  (* field 2 *)
  clear x0 x Hp H1 I H; clear x1 v0 Meq0 H4; rename H9 into okF;
    rename x2 into x1; rename v1 into v0; rename Meq1 into Meq0.
  assert (Meq : mapX (map_union (spaceIdMap (rget s rtoStart) (rget s rtoEnd)) 
    (map_union phi phi')) v0 = Some x1).
  unfold mapX in * ; unfold map_union in * .
  destruct (isOdd v0).
  assumption.
  generalize dependent Meq0; stSimpl; intro;
    destruct (spaceIdMap (rget s rtoStart) (rget s rtoEnd) v0); assumption.
  clear Meq0; unfold mapX in Meq; unfold map_union at 1 in Meq.
  assert (IO := isOddOk v0); destruct (isOdd v0).
  injection Meq; intro; subst x1; left; assumption.
  destruct okF as [okF | okF].
  elf; apply (not_even_and_odd v0); assumption.
  rewrite spaceIdMapNone in Meq.
  right; apply SubBGT; destruct IsoU as [_ [IsoU _]]; eapply IsoU; eauto.
  intros I; apply (set_disj_elim frToDisj v0).
  apply objRangeSetRangeSubset; rewrite <- WMeqv; assumption.
  assumption.
  (* B and Ba *)
  assert (HpX : h0 ||- 
    objHp (B |_| Gz |_| rangeObjs (rget s' ralloc + 8) (rget s rtoEnd))
    (B |_| Bz'')).
  subst Bz''.
  unfold hpredTrue; rewrite (set_eqvUnionSymm B (rangeObjs (rget s' ralloc) (rget s rtoEnd))).
  rewrite v0Eq in Hp0.
  assert (AL : aligned8 (8 + rget s' ralloc) (rget s rtoEnd)).
  replace (8 + rget s' ralloc) with (rget s' ralloc + 8); [idtac | omega].
  rewrite <- allocEq; destruct Pre as [v1 [v2 [v3 [_ [phi0 [W0 [M0 [X [Y _]]]]]]]]]; 
    generalize dependent X; stSimpl; simpl; intuition.
  rewrite <- (rangeObjsStartAdd AL).
  replace (set_add (rget s' ralloc)
    (rangeObjs (8 + rget s' ralloc) (rget s rtoEnd)) |_| B) with 
  (set_add (rget s' ralloc)
    ((rangeObjs (8 + rget s' ralloc) (rget s rtoEnd)) |_| B)); [idtac | reflexivity].
  apply objHpIntro.
  scm.
  hpunfold; lli; scm; left; constructor; constructor.
  unfold hpredTrue;
    rewrite (set_eqvUnionSymm (rangeObjs (8 + rget s' ralloc) (rget s rtoEnd)) B);
      replace (8 + rget s' ralloc) with (rget s' ralloc + 8); [assumption | omega].
  clear Hp0; rename HpX into Hp0;
    rewrite <- allocEq in Hp0; apply (objHpWeaken Hp0).
  intros x I; subst toObjs; rewrite <- (set_eqvUnionAssoc B0 G0 Bz'');
    rewrite BG0eqv; set_InElim.
  sui1; sui1; sui1; assumption.
  sui1; sui1; sui2; 
    rewrite Geqv; assumption.
  sui2; subst Bz''; apply rangeObjsIntro with (S k); omega.
  (* G' and G'' *)
  unfold hpredTrue; rewrite GGeq; apply objHpMerge; 
    hpunfoldIn Hp0; unfold objHp; hpunfold; llMatch; lle Hp0.
  (* G' *)
  assert (HH : s |||- (okObjAt ((W |_| M) |_| B |_| G |_| Ba) x ** top) ** top).
  assert (I' : set_In x Gz).
  rewrite <- Geqv; rewrite G'eq; sui2; assumption.
  destruct Pre as [v1 [v2 [v3 [_ [phi0 [W0 [M0 [_ [_ [WMeqv0 [_ [_ [X _]]]]]]]]]]]]].
  generalize dependent WMeqv0; generalize dependent X; stSimpl; intros.
  clear H3 H4 Hp; sconjLift4 X; scm.
  hyr Hp (objHpElim I' Hp); scm.
  unfold hpredTrue; rewrite WMeqv; rewrite <- WMeqv0.
  rewrite Geqv; assumption.
  hpunfoldIn HH; lle HH.
  assert (X1 := contContEq H3 HH); sconjLift2 HH; assert (X2 := contContEq H4 HH);
    clear HH H3 H4; subst x2 x3.
  assert (Sub' : set_subset ((W |_| M) |_| B |_| G |_| Ba) ((W' |_| M |_| M') |_| toObjs)).
  intros y Iy; subst toObjs; rewrite <- (set_eqvUnionAssoc B0 G0 Bz''); rewrite BG0eqv.
  set_InElim.
  sui1; rewrite (set_eqvUnionSymm M M'); 
    rewrite <- (set_eqvUnionAssoc W' M' M); rewrite (set_eqvUnionSymm W' M'); 
      rewrite <- W'eq; sui1; assumption.
  sui1; sui2; sui1; assumption.
  sui2; sui1; sui1; 
    sui1; assumption.
  sui2; sui1; sui1; 
    sui2; assumption.
  sui2; sui2; 
    subst Ba Bz''; set_InElim; apply rangeObjsIntro with (S k); omega.
  lli; [apply (okFieldValWeaken H5); assumption | apply Hp0 | 
    apply (okFieldValWeaken H8); assumption].
  (* G'' *)
  assert (HH : s |||- (okObjAt (W |_| M) x0 ** top) ** top).
  assert (I' : set_In x0 W).
  rewrite W'eq; sui1; 
    destruct Iso' as [Iso' _]; destruct (Iso' x0) as [_ Q]; apply Q;
      rewrite H1; discriminate.
  clear Hp H6 H7; destrBL bOk; sconjLift2 Hp; scm; hyr Hp (objHpElim I' Hp); scm.
  hpunfoldIn HH; lle HH.
  assert (X1 := contContEq H6 HH); sconjLift2 HH; assert (X2 := contContEq H7 HH);
    clear HH H6 H7 H1 x0; subst x3 x4.
  assert (Sub' : set_subset (W |_| M) ((W' |_| M |_| M') |_| toObjs)).
  rewrite (set_eqvUnionSymm M M'); rewrite <- (set_eqvUnionAssoc W' M' M);
    rewrite (set_eqvUnionSymm W' M'); rewrite <- W'eq;
      intros y Iy; sui1; assumption.
  lli; [apply (okFieldValWeaken H4); assumption | apply Hp0 |
    apply (okFieldValWeaken H9); assumption].

  constructor; clear HpV; rename HpV' into HpV.

  (* show the heap is okay *)
  clear Hp h bOk; assert (Hp := sconjInsert Hp'' HpV); clear Hp'' HpV.
  sconjLift3 Hp; apply (sconjImplR Hp); clear Hp; intros h Hp.
  set (X := rget s rgcInfo
    |->, (rget s rfrStart :: rget s rfrEnd :: rget s rtoStart :: rget s rtoEnd :: nil)) in Hp.
  lle Hp.
  sconjLift2 Hp; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift2 Hp; apply (sconjImpl Hp); clear h Hp; intros h Hp.
  unfold hpredTrue; rewrite <- F'eq'; assumption.
  unfold gcHpPtrMem' in Hp; subst X; lle Hp; lli; exact Hp.

  (* show the virtual heap is okay *)
  clear Hp bOk; subst vh'; rename Hp'' into Hp'; rename HpV into Vhp.

  assert (Meven : forall x, set_In x (M |_| M') -> even x).
  intros y Iy; sconjLift3 Hp'; assert (HX : s' |||- (y |->_ ** top) ** top).
  scm.
  hyr Hp (mapHpElim1 Iy Hp); lleim.
  apply ptrPeven; lle HX; eapply contElim1L; eauto.

  assert (W'even : forall x, set_In x W' -> even x).
  intros y Iy; sconjLift2 Vhp; assert (HX : h' ||- (y |->_ ** top) ** top).
  scm.
  hyr Hp (objHpElim Iy Hp); unfold okObjAt in Hp; unfold okFieldAt in Hp; lleim.
  apply ptrPeven; lle HX; eapply contElim1L; eauto.

  assert (NIM : forall x, okFieldVal toObjs x -> ~set_In x (M |_| M')).
  intros y Iy Iy'.
  destruct Iy as [Iy | Iy].
  apply (not_even_and_odd y); auto.
  apply (set_disj_elim frToDisj y); apply objRangeSetRangeSubset.
  rewrite <- WMeqv; rewrite W'eq; set_InElim.
  sui2; assumption.
  sui1; sui1; assumption.
  rewrite toEq; assert (I' : set_In y (((B |_| G) |_| B'' |_| G'') |_| Bz'')).
  rewrite <- BG0eqv; rewrite (set_eqvUnionAssoc B0 G0 Bz''); assumption.
  clear Iy Iy'; rewrite F'eq; set_InElim.
  sui1; sui1; assumption.
  sui1; sui2; assumption.
  sui2; sui1; sui1; 
    sui1; assumption.
  sui2; sui1; sui1; 
    sui2; assumption.
  subst Ba Bz''; set_InElim.
  sui2.
  destruct k.
  sui1; sui2; 
    left; unfold POnat.POnat.A in * ; omega.
  sui2; apply rangeObjsIntro with k; omega.

  clear h; rename Iso into Iso1; rename Iso' into Iso2; rename IsoU into Iso.
  apply (sconjMapRangeImplS Vhp); clear Vhp; intros h Sub Hp.
  rename h into h0; apply (fasMapRangeImplIS Hp); clear Hp; intros h x Sub' _ Hp.
  assert (Sub'' : Heap.subHeap h (heapOf s')).
  eapply Heap.subHeapTrans; eauto.
  eapply Heap.subHeapTrans; eauto.
  eapply sconjEqSubHeap; eauto.

  clear h0 h' Sub Sub' Hp'.
  unfold okObjAt in Hp; unfold okFieldAt in Hp; lle Hp.
  assert (Hp' := subHeapImpl Hp Sub''); lle Hp'.
  unfold objCopiedH; unfold fieldCopiedH; lli.
  2:apply (sconjMapRangeImpl Hp); intros h' Hp''; apply (contMapRangeImpl Hp'').
  cbv beta; destruct Iso as [Iso _]; destruct (Iso x0) as [_ Iso0];
    destruct (map_union phi phi' x0).
  elf; apply (NIM x0 H3); apply Iso0; discriminate.
  scm.
  cbv beta; destruct Iso as [Iso _]; destruct (Iso x1) as [_ Iso0];
    destruct (map_union phi phi' x1).
  elf; apply (NIM x1 H4); apply Iso0; discriminate.
  sconjLift2 Hp'; scm.

  (* done dealing with the black objects.  now for the gray and white ones. *)
  sconjLift2of2 Hp; apply fasMerge.
  unfold objRespMapH; unfold fieldRespMapH;
    unfold objHp in Hp; rename h into h0; apply (sconjMapRangeImplS Hp); 
      clear Hp; intros h1 Sub' Hp;
	apply (fasMapRangeImplIS Hp); clear Hp; intros h x Sub'' _ Hp.

  assert (Sub3 : Heap.subHeap h (heapOf s')).
  eapply Heap.subHeapTrans; eauto.
  eapply Heap.subHeapTrans; eauto.
  eapply Heap.subHeapTrans; eauto.
  eapply sconjEqSubHeap; eauto.
  clear h0 Sub h1 Sub' Sub''.
  unfold okObjAt in Hp; unfold okFieldAt in Hp; lle Hp.
  assert (HpX := subHeapImpl Hp Sub3); lle HpX.
  lli.
  2:apply (sconjMapRangeImpl Hp); intros h'' Hp''; apply (contMapRangeImpl Hp'').
  (* first field *)
  exists x0; constructor.
  scm.
  clear h Sub3 Hp; clear x1 H4 HpX; rename H3 into okF.
  assert (DD : (odd x0 \/ set_In x0 toObjs) \/ set_In x0 W' \/ set_In x0 (M |_| M')).
  destruct okF as [okF | okF].
  left; left; assumption.
  destruct (set_union_elim okF) as [I | I]; [right | auto]; clear okF;
    destruct (set_union_elim I); auto.
  unfold mapX; unfold map_union at 1; clear okF; destruct DD as [DD | [DD | DD]].
  hyr NIM (NIM _ DD); clear DD.
  destruct Iso as [Iso _]; destruct (Iso x0) as [_ X];
    destruct (map_union phi phi' x0); clear Iso.
  elf; apply NIM; apply X; discriminate.
  destruct (isOdd x0); reflexivity.
  destruct Iso as [Iso _]; destruct (Iso x0) as [_ X];
    destruct (map_union phi phi' x0); clear Iso.
  rewrite isOddFalse; auto.
  destruct (isOdd x0); reflexivity.
  rewrite isOddFalse.
  destruct (map_union phi phi' x0); reflexivity.
  auto.
  (* second field *)
  exists x1; constructor.
  sconjLift2 HpX; scm.
  clear h Sub3 Hp; clear x0 H3 HpX; rename H4 into okF; rename x1 into x0.
  assert (DD : (odd x0 \/ set_In x0 toObjs) \/ set_In x0 W' \/ set_In x0 (M |_| M')).
  destruct okF as [okF | okF].
  left; left; assumption.
  destruct (set_union_elim okF) as [I | I]; [right | auto]; clear okF;
    destruct (set_union_elim I); auto.
  unfold mapX; unfold map_union at 1; clear okF; destruct DD as [DD | [DD | DD]].
  hyr NIM (NIM _ DD); clear DD.
  destruct Iso as [Iso _]; destruct (Iso x0) as [_ X];
    destruct (map_union phi phi' x0); clear Iso.
  elf; apply NIM; apply X; discriminate.
  destruct (isOdd x0); reflexivity.
  destruct Iso as [Iso _]; destruct (Iso x0) as [_ X];
    destruct (map_union phi phi' x0); clear Iso.
  rewrite isOddFalse; auto.
  destruct (isOdd x0); reflexivity.
  rewrite isOddFalse.
  destruct (map_union phi phi' x0); reflexivity.
  auto.

  (* now for the last set of objects *)
  assert (Sub3 : Heap.subHeap h (heapOf s')).
  eapply Heap.subHeapTrans; eauto.
  eapply Heap.subHeapTrans; eauto.
  eapply Heap.subHeapTrans; eauto.
  eapply sconjEqSubHeap; eauto.
  clear h0 Sub h1 Sub' Sub''.
  unfold okObjAt in Hp; unfold okFieldAt in Hp; lle Hp.
  assert (HpX := subHeapImpl Hp Sub3); lle HpX; clear Sub3.
  lli.
  2:apply (sconjMapRangeImpl Hp); intros h'' Hp''; apply (contMapRangeImpl Hp'').
  (* first field *)
  exists x0; constructor.
  scm.
  unfold mapX; unfold map_union at 1.
  destruct Iso as [Iso _]; destruct (Iso x0) as [_ Iso3].
  clear h x Hp HpX; clear x1 H4; destruct H3 as [okF | okF].
  rewrite (isOddTrue okF); destruct (map_union phi phi' x0).
  elf; apply (NIM x0).
  left; assumption.
  apply Iso3; discriminate.
  reflexivity.
  destruct (set_union_elim okF) as [I | I]; clear okF.
  rewrite (isOddFalse (W'even _ I)).
  destruct (map_union phi phi' x0); reflexivity.
  rewrite (isOddFalse (Meven _ I)).
  destruct (map_union phi phi' x0); reflexivity.
  (* second field *)
  exists x1; constructor.
  sconjLift2 HpX; scm.
  unfold mapX; unfold map_union at 1.
  destruct Iso as [Iso _]; destruct (Iso x1) as [_ Iso3].
  clear h x H Hp HpX; clear x0 H3; rename x1 into x0; destruct H4 as [okF | okF].
  rewrite (isOddTrue okF); destruct (map_union phi phi' x0).
  elf; apply (NIM x0).
  left; assumption.
  apply Iso3; discriminate.
  reflexivity.
  destruct (set_union_elim okF) as [I | I]; clear okF.
  rewrite (isOddFalse (W'even _ I)).
  destruct (map_union phi phi' x0); reflexivity.
  rewrite (isOddFalse (Meven _ I)).
  destruct (map_union phi phi' x0); reflexivity.

  (* show disjointedness.  again. *)
  unfold objHp in Hp; unfold okObjAt in Hp; unfold okFieldAt in Hp;
    eapply fasFasDisj'; eauto; cbv beta; illeim.

Qed.
(* a lot of the end of the proof was copied from somewhere else.  I think bakerAllocOk. *)





Lemma bakerLoop2Ok : iseqOkS Pbaker bakerLoop2Pre bakerLoop2Guar bakerLoop2.
Proof.
  unfold bakerLoop2.
  iseqWeakSolver PbakerSimpl.

  exact scanFieldGuarPresRa.

  intros s Pre.
  cbv beta.
  rewrite addiuStep.
  rewrite addiuStep; stSimpl.
  rewrite addiuStep; stSimpl.
  
  assert (E : exists s', s' = baker2Step s).
  econstructor; reflexivity.

  destruct E as [s' E].
  unfold WORDSIZE; unfold OBJSIZE.
  rewrite <- E.

  cut (scanFieldPre s' /\ 
    forall s'', scanFieldGuar s' s'' ->
      bakerLoopPre s'' /\
      forall s''', bakerLoopGuar s'' s''' ->
	bakerLoop2Guar s s''').
  intros [H1 F].
  constructor.
  constructor.
  assumption.
  intros s'' G; destruct (F _ G); assumption.
  intros s''' [s'' [G1 G2]]; destruct (F _ G1).
  eauto.

  destruct Pre as [Pre G1].
  subst s'.
  constructor.
  assumption.
  intros s'' G1'.
  hyr G1 (G1 _ G1').
  constructor.
  assumption.
  
  intros s''' G2.
  exists s''.
  constructor; assumption.
  
Qed.


Lemma bakerExitOk : iseqOkS Pbaker bakerExitPre bakerExitGuar bakerExit.
Proof.
  unfold bakerExit.
  iseqWeakSolver PbakerSimpl.
  unfold bakerExitPre; unfold gcHpPtrMem; unfold gcHpPtrMem'.
  intros s [alFA Hp].
  linLogElim Hp.
  replace (4 + (16 + rget s rgcInfo)) with (20 + rget s rgcInfo) in Hp; [idtac | omega].
  replace (4 + (20 + rget s rgcInfo)) with (24 + rget s rgcInfo) in Hp; [idtac | omega].
  set (free := rget s rfree) in * .
  set (alloc := rget s ralloc) in * .
  assert (Hp0 := Hp).
  stSimpl; unfold doBtest.

  compare free alloc.

  (* no space left: fail by going to infinite loop *)
  intro Eq; rewrite Eq; rewrite beq_true.
  intuition.

  (* space left, do the allocation *)
  intro NE; rewrite (beq_false NE).
  rewrite addiuStepZero; stSimpl;
    rewrite subuStep; stSimpl;
      rewrite addiuStepZero; stSimpl.
  assert (Hp' : s |||- (alloc - 8 |->__ **
    (foralls x in rangeObjs free (alloc - 8), x |->__)) ** 
  16 + rget s rgcInfo |-> x ** 20 + rget s rgcInfo |-> x0 ** 24 + rget s rgcInfo |-> x1 ** top).
  sconjLift4 Hp; scm.
  assert (I : set_In (alloc - 8) (rangeObjs free alloc)).
  destruct alFA as [k alFA]; apply rangeObjsIntro with (pred k); omega.
  hyr Hp (fasElim I Hp).
  scm; unfold hpredTrue; rewrite <- (rangeObjsEndRemove alFA NE); assumption.

  clear Hp; rename Hp' into Hp; lle Hp.
  rewrite (swStep Hp); try reflexivity.
  hyr Hp (contUpdateSt Hp 1).
  sconjLift2 Hp.
  rewrite (swStep Hp); try reflexivity.
  hyr Hp (contUpdateSt Hp 1).
  rewrite addiuStep0; stSimpl.
  sconjLift4 Hp.
  rewrite (swStep Hp); try reflexivity.
  hyr Hp (contUpdateSt Hp (rget s rscan)).
  sconjLift5 Hp.
  rewrite (swStep Hp); try reflexivity.
  hyr Hp (contUpdateSt Hp free).
  sconjLift6 Hp.
  rewrite (swStep Hp); try reflexivity.
  hyr Hp (contUpdateSt Hp alloc).
  rewrite addiuStep0; stSimpl.
  clear Hp.

  constructor.
  trivial.

  (* show the guar *)
  intros s' Eq; subst s'.
  unfold bakerExitGuar.
  constructor.

  unfold gcHpPtrMem; intros A Hp; unfold gcHpPtrMem' in Hp; lle Hp.
  assert (X := contContEq Hp0 Hp);
    sconjLift2 Hp0; sconjLift2 Hp; assert (X' := contContEq Hp0 Hp);
      sconjLift3 Hp0; sconjLift3 Hp; assert (X'' := contContEq Hp0 Hp);
	subst x4 x5 x6; clear Hp0.
  stSimpl; repeat (rewrite uhsetRsetEq); stSimpl.

  assert (Hp' : s |||- (alloc - 8 |->__ **
    (foralls x in rangeObjs free (alloc - 8), x |->__)) ** 
  24 + rget s rgcInfo |-> x1 ** 20 + rget s rgcInfo |-> x0 ** 16 + rget s rgcInfo |-> x ** A).
  sconjLift4 Hp; scm.
  assert (I : set_In (alloc - 8) (rangeObjs free alloc)).
  destruct alFA as [k alFA]; apply rangeObjsIntro with (pred k); omega.
  hyr Hp (fasElim I Hp).
  scm; unfold hpredTrue; rewrite <- (rangeObjsEndRemove alFA NE); assumption.

  clear Hp; rename Hp' into Hp.
  lle Hp.
  hyr Hp (contUpdateSt Hp 1).
  sconjLift2 Hp; hyr Hp (contUpdateSt Hp 1).
  sconjLift6 Hp; hyr Hp (contUpdateSt Hp (rget s rscan)).
  sconjLift6 Hp; hyr Hp (contUpdateSt Hp free).
  sconjLift6 Hp; hyr Hp (contUpdateSt Hp (alloc - 8)).
  unfold gcHpPtrMem'; lli.
  sconjLift4 Hp; sconjLift5 Hp; sconjLift3 Hp;
    sconjLift4 Hp; sconjLift5 Hp; exact Hp.

  (* show the registers are okay *)
  clear Hp0; stSimpl.
  constructor.
  intros r E; simpl in E; intuition.
  stSimpl; reflexivity.
  destruct alFA; subst alloc free.
  intuition.
  rewrite H.
  exists (pred x4); omega.

Qed.


Lemma loopExitOk : forall s,
  bakerLoopPre s ->
  bakerExitPre s /\
  (forall s', bakerExitGuar s s' -> bakerLoopGuar s s').
Proof.
  intros s Pre.
  constructor.

  (* show the precondition is okay *)
  destruct Pre as [phi [G [W [M [F [_ [_ [_ SF]]]]]]]]; destrSF SF.
  unfold bakerExitPre; stSimpl; constructor.
  assumption.
  sconjLift3 Hp; sconjLift5 Hp; scm; unfold hpredTrue; rewrite <- Feq; assumption.

  (* show the guar is okay *)
  clear Pre.
  unfold bakerExitGuar; stSimpl.
  intros s' [Guar [Req [allocEq [v0Eq [raEq alAl]]]]].
  constructor.
  
  intros phi G W M F A Geq [Le [Al SF]].
  destrSF SF.
  exists (empty_map nat nat).
  exists empty_set.
  exists empty_set.
  exists G.
  exists empty_set.
  exists W.
  exists empty_set.
  exists (set_remove (rget s' ralloc) F).

  constructor.
  simpl; apply set_eqvRefl.
  constructor.
  simpl.
  assert (Ial : set_In (rget s' ralloc) F).
  rewrite Feq; rewrite allocEq; rewrite allocEq in AlF; destruct AlF as [k AlF];
    destruct alAl as [k' X]; apply rangeObjsIntro with (pred k); omega.
  rewrite (addRemoveEqv Ial); apply set_eqvRefl.
  constructor.
  apply empty_mapIso.
  constructor.
  simpl; apply set_eqvRefl.
  rewrite <- Req; [stSimpl | regNotInListSolver].
  rewrite <- Req; [stSimpl | regNotInListSolver].
  constructor.
  constructor.
  rewrite <- (set_eqvEmptyRangeObjs (rget s rscan)); apply set_eqvRefl.
  auto.
  constructor.
  apply aligned8Refl.

  sconjLift2 Hp; destruct (sconjExtract Hp) as [hpw [Hp' HpW]]; clear Hp.
  sconjLift4 Hp'; destruct (sconjExtract Hp') as [hpg [Hp HpG]]; clear Hp'.
  assert (HX : s |||- gcHpPtrMem (rget s rgcInfo) **
    (foralls x in rangeObjs (rget s rfree) (rget s ralloc), x |->__) **
    Heap.heap_eqv hpg ** Heap.heap_eqv hpw ** mapHp M phi ** A).
  sconjLift4 Hp; sconjLift5 Hp; scm.
  unfold hpredTrue; rewrite <- Feq; assumption.
  assert (Hp' := Guar _ HX); clear HX Guar.
  constructor.

  intros x I.
  hyr HpW (objHpElim I HpW).
  sconjLift2 Hp; sconjLift5 Hp'.
  eapply hpHpObjCopied; eauto.
  instantiate (1 := top).
  unfold okObjAt in HpW; unfold okFieldAt in HpW; lle HpW; lli.
  instantiate (1 := x1).
  instantiate (1 := x0).
  scm.

  unfold bakerLoopStOk; unfold scanFieldStOk; unfold fromSpace; unfold toSpace.
  repeat (rewrite <- Req; [stSimpl | regNotInListSolver]); rewrite v0Eq.
  repeat (constructor; [assumption | ]).
  constructor.
  rewrite (set_eqvUnionSymm M empty_set); simpl; assumption.
  constructor.
  replace (rget s' ralloc) with (rget s ralloc - 8); [idtac | omega].
  rewrite Feq; apply rangeObjsEndRemove.
  assumption.
  destruct alAl; omega.
  constructor.
  unfold map_union; intros x y E.
  apply (NSP x y); clear Hp Hp'; destruct (phi x); assumption.
  constructor.
  destruct alAl; omega.
  constructor.
  omega.
  constructor.
  assumption.
  clear Req NSP Le1 Le2 WMdisj WMsub; rename Hp into Hp0.
  sconjLift2 Hp'; sconjLift3 Hp'; sconjLift6 Hp'; 
    sconjLift5 Hp'; sconjLift6 Hp'; sconjLift3 Hp'.
  apply (sconjImpl Hp'); clear Hp'; intros h Hp.
  eapply mapHpPhiEqv.
  instantiate (1 := phi).
  unfold hpredTrue; rewrite (set_eqvUnionSymm M empty_set); simpl; assumption.
  unfold map_union; intro x; destruct (phi x); reflexivity.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  unfold hpredTrue; rewrite (set_eqvUnionSymm M empty_set); simpl.
  rewrite <- hpredTrueEq; eapply heapEqvTrue.
  2:eauto.
  apply Heap.heap_eqvSymm; eapply eqvPredEqv; eauto.
  
  sconjLift2 Hp; apply (sconjImpl Hp); clear h Hp; intros h Hp.
  unfold hpredTrue; rewrite Feq.
  replace (rget s' ralloc) with (rget s ralloc - 8); [idtac | omega].
  assert (NEE : rget s rfree <> rget s ralloc).
  destruct alAl; omega.
  rewrite (rangeObjsEndRemove AlF NEE).
  replace (rget s' ralloc) with (rget s ralloc - 8) in Hp; [idtac | omega].
  assumption.
  lli.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  eapply heapEqvTrue.
  apply Heap.heap_eqvSymm; apply eqvPredEqv; eauto.
  unfold objHp in HpG; eapply fasImplIS; eauto.
  cbv beta; unfold okObjAt; unfold okFieldAt; intros z hh _ Sub HH; lle HH.
  assert (QQ : s |||- (z |-> x ** 4 + z |-> x0) ** top).
  eapply subHeapImpl.
  eauto.
  eapply Heap.subHeapTrans; eauto.
  eapply sconjEqSubHeap; eauto.
  lle QQ.
  unfold objCopied; unfold fieldCopied; lli.
  2:eauto.
  sconjMatch.
  sconjLift2 QQ; sconjMatch.

  lle Hp; assumption.

  (* deal with the registers *)
  constructor.
  intro r; simpl; intro E; decompose [and] E; clear E; 
    rewrite <- Req; [stSimpl; reflexivity | regNotInListSolver].

  rewrite v0Eq; rewrite raEq; rewrite <- allocEq; intuition.

Qed.


Lemma loopExit2Ok : forall s,
  bakerLoopPre s ->
  bakerExitPre (rset (rset s ast scan_per_gc) ast 0) /\
  (forall s',
    bakerExitGuar (rset (rset s ast scan_per_gc) ast 0) s' ->
    bakerLoopGuar s s').
Proof.
  intros s Pre.

  cut (bakerLoopPre (rset (rset s ast scan_per_gc) ast 0)).
  clear Pre; intro Pre; hyr Pre (loopExitOk Pre); destruct Pre as [Pre Guar]; 
    constructor; [assumption | clear Pre].

  intros s' G; hyr Guar (Guar s' G); clear G.
  generalize Guar; clear Guar; unfold bakerLoopGuar; stSimpl.
  intros; intuition; clear H H0 H2 H4.
  intros r Ir; hyr H1 (H1 r Ir).
  simpl in Ir; decompose [and] Ir; clear Ir.
  rewrite <- H1; stSimpl; reflexivity.

  generalize dependent Pre; unfold bakerLoopPre; stSimpl; auto.

Qed.


Lemma bakerInductiveGuarOk : forall s s' s'',
  rget s rscan <> rget s rfree ->
  (forall phi W M F A0,
    scanFieldStOk phi W M F
      (okFieldAt ((W |_| M) |_| toSpace s) (rget s rscan) ** A0) s ->
    exists M'', exists W', exists F', exists G, exists phi',
      set_eqv W (M'' |_| W') /\
      set_eqv F (G |_| F') /\
      iso_set phi' M'' G /\
      aligned8 (rget s rfree) (rget s' rfree) /\
      (forall x, set_In x W' ->
        s' |||- objCopied x s x ** top) /\
      scanFieldStOk (map_union phi phi') W' (M |_| M'') F'
      (fieldRespToMap (map_union phi phi') s (rget s rscan) **
        (foralls x in G, objCopiedPhi phi' s x) ** A0) s') ->
  regPres (regNotInList (rfree::a0::v0::ast::rtemp0::rtemp1::
    rtemp2::raSave0::nil))
    (rset (rset (rset (rset s ast scan_per_gc) ast 1) a0 (rget s rscan)) 
      ra BAKERLOOP2) s' ->
  bakerLoop2Guar s' s'' -> 
  bakerLoopGuar s s''.
Proof.
  intros s s' s3 NE0 Guar Req Guar'; constructor.
  destruct Guar' as [s'' [Guar' Guar'']].
  intros phi G W M F A Geqv BL.
  destrBL BL.

  assert (Hp2 : s |||- mapHp M phi ** objHp (W |_| M) W ** (foralls x in F, x |->__) **
    (okObjAt ((W |_| M) |_| toSpace s) (rget s rscan) **
    objHp ((W |_| M) |_| toSpace s) (set_remove (rget s rscan) G)) ** gcHpPtrMem (rget s rgcInfo) ** A).
  scm.

  assert (IG : set_In (rget s rscan) G).
  rewrite Geqv; eapply rangeObjsIntro.
  destruct Al1; omega.
  instantiate (1 := 0); auto.
  apply (objHpElim IG) in Hp; scm.

  unfold okObjAt in Hp2; unfold okFieldAt at 2 in Hp2; lle Hp2.
  sconjLift6 Hp2; destruct (sconjExtract Hp2) as [h' [Hp3 H'ok]]; clear Hp2 Hp.
  rename Hp3 into Hp.

  hyr Guar (Guar phi W M F
    (4 + rget s rscan |-> x ** Heap.heap_eqv h' ** gcHpPtrMem (rget s rgcInfo) ** A)).
  destruct Guar as [M' [W' [F' [G' [phi' X]]]]].
  unfold scanFieldStOk; intuition; simpl.
  sconjLift5 Hp; sconjLift5 Hp; sconjLift5 Hp; sconjLift5 Hp.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift3 Hp; sconjLift2 Hp; assumption.

  destruct X as [Weq [Feq2 [Iso' [Al' [W'copy SF']]]]].
  destruct Guar' as [Guar Req2].
  generalize dependent Guar; stSimpl; intro.
  hyr Guar (Guar (map_union phi phi') W' (M |_| M') F'
    (fieldRespToMap (map_union phi phi') s (rget s rscan) **
      (foralls x in G', objCopiedPhi phi' s x) **
      Heap.heap_eqv h' ** gcHpPtrMem (rget s rgcInfo) ** A)).
  destruct Guar as [M'' [W'' [F'' [G'' [phi'' X'']]]]].

  destrSF SF'; unfold scanFieldStOk; intuition; stSimpl.
  apply (sconjImplR Hp0); clear Hp0; intros h Hp0.
  apply (sconjImplR Hp0); clear h Hp0; intros h Hp0.
  apply (sconjImplR Hp0); clear h Hp0; intros h Hp0.
  sconjLift2 Hp0; sconjLift2 Hp0; sconjLift3 Hp0; scm.
  rewrite <- (Req rscan); [stSimpl | regNotInListSolver].
  unfold okFieldAt; lli; eauto.
  eapply okFieldValWeaken; eauto.
  unfold toSpace; stSimpl; rewrite <- Req; [idtac | regNotInListSolver].
  rewrite <- Req; [stSimpl | regNotInListSolver].
  rewrite Weq.
  rewrite (set_eqvUnionSymm M' W').
  rewrite (set_eqvUnionAssoc W' M' M).
  rewrite (set_eqvUnionSymm M' M).
  apply set_subsetRefl.

  destruct X'' as [W'eq [F'eq [Iso'' [Al'' [Wcopy'' SF'']]]]].
  destruct Guar'' as [Guar [Req'' Ra'']].
  hyr Guar (Guar (map_union (map_union phi phi') phi'')
    ((set_remove (rget s rscan) G) |_| G' |_| G'') W'' ((M |_| M') |_| M'') F''
    (objRespToMap (map_union (map_union phi phi') phi'') s (rget s rscan) ** A)).

  destruct Guar as [phi3 [B [B' [G3 [G4 [W3 [M3 [F3 X3]]]]]]]].
  rewrite <- Req2; [stSimpl | regNotInListSolver].
  rewrite <- Req; [stSimpl | regNotInListSolver].
  
  assert (HH1 : aligned8 (8 + rget s rscan) (rget s' rfree)).
  destruct Al1 as [k1 Al1]; destruct Al' as [k2 AlF'].
  destruct k1.
  elimtype False; omega.
  exists (k1 + k2); omega.

  assert (HH2 : rget s' rfree <= rget s'' rfree).
  destruct Al''; omega.

  unfold hpredTrue; rewrite <- (rangeObjsTransEqv HH1 HH2).
  clear HH1 HH2.

  assert (HH1 : aligned8 (8 + rget s rscan) (rget s rfree)).
  destruct Al1 as [k1 Al1]; destruct k1; [elimtype False | exists k1]; omega.
  assert (HH2 : rget s rfree <= rget s' rfree).
  destruct Al'; omega.

  rewrite <- (rangeObjsTransEqv HH1 HH2).
  clear HH1 HH2.
  setoid_replace (rangeObjs (8 + rget s rscan) (rget s rfree)) with 
    (set_remove (rget s rscan) G).
  2:rewrite <- (rangeRemoveEqv (rget s rscan) (rget s rfree)).
  2:rewrite <- Geqv; apply set_eqvRefl.

  assert (RA' : rget s' ralloc = rget s ralloc).
  rewrite <- Req; [stSimpl; reflexivity | regNotInListSolver].
  assert (RA'' : rget s'' ralloc = rget s ralloc).
  rewrite <- Req2; [stSimpl; assumption | regNotInListSolver]; reflexivity.

  assert (G'eqv : set_eqv G' (rangeObjs (rget s rfree) (rget s' rfree))).
  assert (X : set_eqv (rangeObjs (rget s rfree) (rget s ralloc)) (G' |_| F')).
  rewrite <- Feq; assumption.
  eapply rangeSplit.
  instantiate (1 := rget s ralloc).
  assert (U := SF'); destrSF U.
  rewrite RA' in Feq0; rewrite <- Feq0.
  apply set_disjSymm; eapply scanFieldStOkFGDisj.
  apply (scanFieldStOkImpl SF').
  intros hx HpX.
  sconjLift2 HpX; sconjMatch.
  simpl; intros h0 xy; unfold objCopiedPhi.
  unfold objCopied; unfold fieldCopied; intro HpZ; linLogElim HpZ; linLogIntro; sconjMatch.
  destruct Al''; destruct Al'.
  destrSF SF'.
  rewrite RA' in Le5.
  assumption.

  rewrite X.
  destrSF SF'.
  rewrite RA' in Feq0; rewrite <- Feq0; apply set_eqvRefl.
  assumption.

  rewrite <- G'eqv; clear G'eqv.
  assert (Geqv'' : set_eqv G'' (rangeObjs (rget s' rfree) (rget s'' rfree))).
  assert (X : set_eqv (rangeObjs (rget s' rfree) (rget s ralloc)) (G'' |_| F'')).
  destrSF SF'; rewrite RA' in Feq0; rewrite <- Feq0; assumption.
  eapply rangeSplit.
  instantiate (1 := rget s ralloc).
  assert (Q := SF'').
  destrSF SF''.
  rewrite RA'' in Feq0.
  rewrite <- Feq0.
  apply set_disjSymm; eapply scanFieldStOkFGDisj.
  apply (scanFieldStOkImpl Q).
  intros hx HpX.
  sconjLift2 HpX; sconjMatch.
  simpl; intros h0 xy; unfold objCopiedPhi.
  unfold objCopied; unfold fieldCopied; intro HpZ; linLogElim HpZ; linLogIntro; sconjMatch.
  destrSF SF''; rewrite RA'' in Le5; assumption.
  destrSF SF''; rewrite RA'' in Feq0; rewrite <- Feq0.
  apply set_eqvSymm; assumption.
  assumption.

  rewrite <- Geqv''.
  apply set_eqvSymm; apply set_eqvUnionAssoc.


  (* showing we can take the step, but showing s'' is nice *)

  unfold bakerLoopStOk; intuition.
  repeat (rewrite <- Req2; [stSimpl | regNotInListSolver]).
  repeat (rewrite <- Req; [stSimpl | regNotInListSolver]).
  omega.
  destrSF SF''.
  rewrite <- Req2; [stSimpl | regNotInListSolver].
  rewrite <- Req; [stSimpl | regNotInListSolver].
  eapply aligned8Trans.
  instantiate (1 := rget s rfree); assert (Y := Al1); destruct Y as [k Al1Y].
  destruct k.
  elimtype False; omega.
  exists k; omega.
  eapply aligned8Trans; eauto.
  apply (scanFieldStOkImpl SF'').
  rename Hp into Hp0; intros h Hp.

  assert (WMeqv : set_eqv (W'' |_| (M |_| M') |_| M'') (W |_| M)).
  setoid_rewrite (set_eqvUnionSymm (M |_| M') M'').
  setoid_rewrite <- (set_eqvUnionAssoc W'' M'' (M |_| M')).
  setoid_rewrite (set_eqvUnionSymm W'' M''); setoid_rewrite <- W'eq.
  setoid_rewrite (set_eqvUnionSymm M M').
  setoid_rewrite <- (set_eqvUnionAssoc W' M' M).
  setoid_rewrite (set_eqvUnionSymm W' M').
  setoid_rewrite <- Weq.
  apply set_eqvRefl.

  sconjLift3 Hp; hyr Hp (sconjAssocR Hp).
  sconjLift5 Hp; sconjLift5 Hp; sconjLift5 Hp; sconjLift5 Hp.
  hyr Hp (sconjAssocR (sconjAssocR Hp)).
  sconjMatch.
  unfold hpredTrue; rewrite WMeqv.
  setoid_replace (toSpace s'') with (toSpace s).
  2:unfold toSpace.
  2:repeat (rewrite <- Req2; [stSimpl | regNotInListSolver]).
  2:repeat (rewrite <- Req; [stSimpl | regNotInListSolver]).
  2:apply set_eqvRefl.
  sconjLift2of2 Hp; hyr Hp (sconjInsert Hp H'ok).
  apply objHpMerge; scm.
  apply objHpMerge; apply sconjComm; sconjMatch.
  hpunfoldIn Hp; generalize dependent Hp; stSimpl; intro.
  unfold objHp; apply (fasImpl Hp); clear h Hp.
  simpl; intros y h Hp; lle Hp.
  assert (Hp' : s' |||- (okObjAt ((W |_| M) |_| toSpace s) x0 ** top) ** top).
  clear Req Req2 Req''.
  assert (I : set_In x0 W').
  rewrite W'eq; sui1.
  destruct Iso'' as [Q _]; destruct (Q x0) as [_ YY].
  apply YY; rewrite H4; discriminate.
  destrSF SF'.
  clear H9 H10; sconjLift2 Hp1; scm.
  hyr Hp1 (objHpElim I Hp1); sconjMatch.
  eapply okObjAtWeaken; eauto.
  clear I; intros z I; sui1.
  destruct (set_union_elim I) as [ZZ | YY]; clear I.
  rewrite Weq; sui1; sui2; assumption.
  destruct (set_union_elim YY) as [ZZ | YYY]; clear YY.
  sui2; assumption.
  sui1; rewrite Weq; sui1; assumption.
  unfold okObjAt in Hp'; unfold okFieldAt in Hp'; linLogElim Hp'.
  rename Hp' into HpZ.
  assert (ZZ := contContEq H9 HpZ); sconjLift2 HpZ; assert (YY := contContEq H10 HpZ).
  subst x3 x4.
  unfold okObjAt; unfold okFieldAt; linLogIntro.
  2:eauto.
  assumption.
  assumption.

  unfold objHp; apply (fasImpl Hp); clear h Hp.
  hpunfold; simpl; intros y h Hp; linLogElim Hp.
  assert (Hp' : s |||- (okObjAt ((W |_| M) |_| toSpace s) x0 ** top) ** top).
  clear Req Req2 Req''.
  assert (I : set_In x0 W).
  rewrite Weq; sui1.
  destruct Iso' as [Q _]; destruct (Q x0) as [_ YY].
  apply YY; rewrite H4; discriminate.
  clear H9 H10; sconjLift3 Hp0; sconjMatch.
  hyr Hp0 (objHpElim I Hp0); sconjMatch.
  eapply okObjAtWeaken; eauto.
  clear I; intros z I; sui1; assumption.
  unfold okObjAt in Hp'; unfold okFieldAt in Hp'; linLogElim Hp'.
  rename Hp' into HpZ.
  assert (ZZ := contContEq H9 HpZ); sconjLift2 HpZ; assert (YY := contContEq H10 HpZ).
  subst x3 x4.
  unfold okObjAt; unfold okFieldAt; linLogIntro.
  2:eauto.
  assumption.
  assumption.
  rewrite <- Req2; [stSimpl | regNotInListSolver].
  rewrite <- Req; [stSimpl | regNotInListSolver].
  assumption.

  unfold objRespToMap; sconjMatch.
  apply fieldRespToMapAddRight; assumption.
  generalize dependent Hp; unfold fieldRespToMap; unfold fieldRespMap;
    unfold spaceIdMap; stSimpl; repeat (rewrite <- Req; [idtac | regNotInListSolver]);
      stSimpl; intro; linLogElim Hp; linLogIntro.
  2:apply Hp.
  destruct H4 as [v0 [HpV0 Meq0]].
  exists v0; constructor.
  2:assumption.
  clear h Meq0 Hp.
  assert (v0 = x).
  destrSF SF'.
  sconjLift6 Hp; exact (contContEq HpV0 Hp).
  subst v0; sconjLift6 Hp0; sconjMatch.

  (* putting it all together *)
  destruct X3 as [W''eq [F''eq [Iso3 [Geq3 [[BBeq scLE] [Al3 [Wcopy3 Bok3]]]]]]].
  exists (map_union (map_union phi' phi'') phi3).
  exists (set_add (rget s rscan) (set_inter B G)).
    (* objects that were just field forwarded *)
  exists ((set_inter B (G' |_| G'')) |_| B').
    (* objects that were copied then field fwded *)
  exists (set_inter G G3).  (* gray objs that weren't finished *)
  exists ((set_inter (G' |_| G'') G3) |_| G4).  (* new gray objs *)
  exists W3.  (* remaining white objs *)
  exists ((M' |_| M'') |_| M3).  (* newly copied objs *)
  exists F3.  (* objects still free *)

  assert (ScEq : rget s'' rscan = 8 + rget s rscan).
  rewrite <- Req2; [stSimpl | regNotInListSolver];
    rewrite <- Req; [stSimpl | regNotInListSolver]; reflexivity.
  rewrite ScEq in BBeq; rewrite ScEq in scLE; clear ScEq.

  assert (M'eq : set_eqv ((M' |_| M'') |_| M3) (M' |_| M'' |_| M3)).
  rewrite (set_eqvUnionAssoc M' M'' M3).
  apply set_eqvRefl.

  constructor.
  rewrite M'eq.
  rewrite (set_eqvUnionAssoc M' (M'' |_| M3) W3).
  rewrite (set_eqvUnionAssoc M'' M3 W3).
  rewrite Weq; rewrite W'eq; rewrite W''eq.
  apply set_eqvRefl.

  assert (Eq : set_eqv
    (((set_inter B (G' |_| G'')) |_| B') |_|
      ((set_inter (G' |_| G'') G3) |_| G4))
    ((G' |_| G'') |_| (B' |_| G4))).
  assert (X := set_subsetUnionIntro2 Geq3).
  setoid_rewrite (set_eqvUnionAssoc 
    (set_inter B (G' |_| G'')) B'
    (set_union (set_inter (G' |_| G'') G3) G4)).
  clear Req Req2 Req'' SF' Bok3 SF'' Wcopy'' W'copy Hp.
  setoid_rewrite <- (set_eqvUnionAssoc B'
    (set_inter (G' |_| G'') G3) G4).
  setoid_rewrite (set_eqvUnionSymm B' (set_inter (G' |_| G'') G3)).
  setoid_rewrite (set_eqvUnionAssoc
    (set_inter (G' |_| G'') G3) B' G4).
  setoid_rewrite <- (set_eqvUnionAssoc
    (set_inter B (G' |_| G''))
    (set_inter (G' |_| G'') G3)
    (set_union B' G4)).

  rewrite (set_interSymm B (G' |_| G'')).
  setoid_rewrite <- (set_unionInterDistrib2 (G' |_| G'') B G3).
  setoid_rewrite (subsetInterEq X).
  apply set_eqvRefl.

  constructor.
  rewrite Feq2; rewrite F'eq; rewrite F''eq.
  rewrite <- (set_eqvUnionAssoc G' G'' ((set_union B' G4) |_| set_add (rget s3 ralloc) F3)).
  setoid_rewrite Eq.
  setoid_rewrite <- (set_eqvUnionAssoc (G' |_| G'') (B' |_| G4) (set_add (rget s3 ralloc) F3)).
  apply set_eqvRefl.

  assert (GGdisj''' : set_disj G' G'').
  apply set_disjSymm.
  apply set_unionDisjElimL with F''.
  rewrite <- F'eq.
  eapply scanFieldStOkFGDisj.
  eapply scanFieldStOkImpl.
  apply SF'.
  intros h HH; sconjLift2 HH; sconjMatch.
  unfold objCopiedPhi; unfold objCopied; unfold fieldCopied;
    intros h x7 HH; lle HH; lli; sconjMatch.
  
  constructor.
  setoid_rewrite Eq.
  apply iso_union.
  apply iso_union.
  assumption.
  assumption.

  (* set_disj M' M'' *)
  apply set_unionDisjElimR with M.
  apply set_disjSymm.
  apply set_unionDisjElimL with W''.
  rewrite <- W'eq.
  apply (scanFieldStOkWMDisj SF').

  assumption.
  assumption.

  (* set_disj (M' |_| M'') M3 *)
  apply set_unionDisjElimR with M.
  rewrite <- (set_eqvUnionAssoc M M' M'').
  apply set_disjSymm.
  apply set_unionDisjElimL with W3.
  rewrite <- W''eq.
  apply (scanFieldStOkWMDisj SF'').

  (* set_disj (G' |_| G'') (B' |_| G4) *)
  apply set_disjSymm.
  apply set_unionDisjElimL with (set_add (rget s3 ralloc) F3).
  rewrite <- F''eq.
  eapply scanFieldStOkFGDisj.
  eapply scanFieldStOkImpl.
  apply SF''.
  instantiate (1 := top).
  instantiate (1 := fun x => x |->__).
  intros h HH; sconjLift2 HH; sconjLift4 HH; hyr HH (sconjAssocR HH); scm.
  apply fasMerge; scm.
  unfold objCopiedPhi in Hp0; unfold objCopied in Hp0; unfold fieldCopied in Hp0.
  eapply fasImpl.
  apply Hp0.
  cbv beta; intros x7 hh HH; lle HH; lli; sconjMatch.
  unfold objCopiedPhi in Hp0; unfold objCopied in Hp0; unfold fieldCopied in Hp0.
  eapply fasImpl.
  apply Hp0.
  cbv beta; intros x7 hh HH; lle HH; lli; sconjMatch.
  cbv beta; intros x7 hh HH; lle HH; lli; sconjMatch.

  (* set_eqv G (set_add (rget s rscan) (set_inter B G) |_| set_inter G G3) *)
  constructor.
  rewrite (set_addUnionDistrib (rget s rscan) (set_inter B G) (set_inter G G3)).
  rewrite (set_interSymm B G).
  setoid_rewrite <- (set_unionInterDistrib2 G B G3).
  setoid_rewrite (set_removeCancel (rget s rscan) (set_inter G (B |_| G3))).
  setoid_rewrite <- (set_removeInterDistrib (rget s rscan) G (B |_| G3)).
  setoid_rewrite (subsetInterEq (set_subsetUnionIntro1 Geq3)).
  assert (IscanG : set_In (rget s rscan) G).
  setoid_rewrite Geqv.
  destruct Al1.
  eapply rangeObjsIntro.
  omega.
  instantiate (1 := 0).
  omega.
  setoid_rewrite (addRemoveEqv IscanG).
  apply set_eqvRefl.

  constructor.
  clear Req Req2 Req''.
  constructor.
  assert (Beq : set_eqv B (set_inter B ((set_remove (rget s rscan) G) |_| G' |_| G''))).
  assert (D : set_disj B G3).
  destrBL Bok3; sconjLift5 Hp0; sconjLift5 Hp0.
  assert (Hp0' : s3 |||- fasd B ** fasd G3 ** top).
  llMatch; hpunfoldIn Hp0; lleim.
  apply (fasFasDisj Hp0'); auto.
  set (GG := set_remove (rget s rscan) G |_| G' |_| G'') in * .
  rewrite Geq3; rewrite (set_unionInterDistrib2 B B G3).
  rewrite (reflInterEq B); rewrite (set_disjInterEmpty D);
    rewrite (set_eqvUnionSymm B empty_set); apply set_eqvRefl.

  rewrite <- (set_eqvUnionAssoc (set_add (rget s rscan) (set_inter B G))
    (set_inter B (G' |_| G'')) B');
  rewrite (addRemoveInter (rget s rscan) B G).
  replace (set_add (rget s rscan) (set_inter B (set_remove (rget s rscan) G)) |_| 
      set_inter B (G' |_| G'')) with
    (set_add (rget s rscan) (set_inter B (set_remove (rget s rscan) G) |_| 
      set_inter B (G' |_| G''))).
  2:reflexivity.
  rewrite <- (set_unionInterDistrib2 B (set_remove (rget s rscan) G)
    (G' |_| G'')); rewrite <- Beq.
  replace (set_add (rget s rscan) B |_| B') with
    (set_add (rget s rscan) (B |_| B')); [idtac | reflexivity];
    rewrite BBeq.
  assert (AL : aligned8 (8 + rget s rscan) (rget s3 rscan)).
  assert (NE : rget s rscan <> rget s3 rscan).
  omega.
  assert (AL : aligned8 (rget s rscan) (rget s3 rscan)).
  apply aligned8LE_R with (rget s3 rfree).
  eapply aligned8Trans; eauto.
  eapply aligned8Trans; eauto.
  eapply aligned8Trans; eauto.
  destrBL Bok3; assumption.
  omega.
  destruct AL as [k AL]; exists (pred k); destruct k; omega.
  rewrite (rangeObjsStartAdd AL).
  apply set_eqvRefl.
  omega.

  constructor.
  eapply aligned8Trans; eauto.
  eapply aligned8Trans; eauto.

  assert (Wcopy' : forall x, set_In x W' -> s' |||- objCopied x s x ** top).
  generalize dependent W'copy; unfold objCopied; unfold fieldCopied.
  stSimpl.
  intros; apply W'copy.
  assumption.

  clear W'copy; rename Wcopy'' into Wcopy.

  assert (Wcopy'' : forall x, set_In x W'' -> s'' |||- objCopied x s' x ** top).
  generalize dependent Wcopy; unfold objCopied; unfold fieldCopied.
  stSimpl.
  intros; apply Wcopy.
  assumption.

  constructor.
  intros y Iy.
  eapply objCopiedTrans.
  2:eauto.
  eapply objCopiedTrans.
  2:apply Wcopy''.
  apply Wcopy'.
  setoid_rewrite W'eq; sui2; 
    setoid_rewrite W''eq; sui2; assumption.
  setoid_rewrite W''eq; sui2; assumption.

  rename Hp into Hp00; destrBL Bok3.

  assert (Meqv : set_eqv
    (set_union M (set_union (set_union M' M'') M3))
    (set_union (set_union (set_union M M') M'') M3)).
  setoid_rewrite <- (set_eqvUnionAssoc M (set_union M' M'') M3).
  setoid_rewrite <- (set_eqvUnionAssoc M M' M'').
  apply set_eqvRefl.

  unfold bakerLoopStOk.
  intuition.
  unfold scanFieldStOk.

  assert (PhiEqv : mapEqvnn (map_union phi (map_union (map_union phi' phi'') phi3))
    (map_union (map_union (map_union phi phi') phi'') phi3)).
  intros z.
  unfold map_union.
  destruct (phi z).
  reflexivity.
  destruct (phi' z).
  reflexivity.
  destruct (phi'' z); reflexivity.

  intuition.
  setoid_rewrite Meqv; assumption.
  eapply NSP0; eauto.
  instantiate (1 := x0).
  assert (X := PhiEqv x0).
  rewrite <- X.
  auto.

  apply (sconjImpl Hp); clear Hp; intros h Hp.

  unfold hpredTrue.
  setoid_rewrite Meqv; setoid_rewrite PhiEqv.
  auto.

  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  unfold hpredTrue.
  setoid_rewrite Meqv; assumption.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift5 Hp; sconjLift6 Hp.

  assert (Hp' : h ||- 
    gcHpPtrMem' (rget s'' rgcInfo) (rget s3 rscan) (rget s3 rfree) (rget s3 ralloc) **
    rget s3 v0 |->, (1::1::nil) **
    objRespToMap (map_union (map_union phi phi') phi'') s (rget s rscan) **
    (foralls x in B,
      objRespToMap (map_union (map_union (map_union phi phi') phi'') phi3) s'' x) **
    (foralls x in G3, objCopied x s'' x) **
    (foralls x in B', objIsToFwded (map_union (map_union phi phi') phi'') phi3 s'' x) **
    (foralls x in G4, objCopiedPhi phi3 s'' x) ** 
    A).
  sconjLift7 Hp; sconjLift3 Hp; sconjLift3 Hp; scm.

  clear Hp; rename Hp' into Hp.
  apply sconjGLift3'; apply sconjAssocL.
  sconjLift4 Hp; sconjLift4 Hp; hyr Hp (sconjAssocR Hp).

  assert (IscanB : ~ set_In (rget s rscan) B).
  intro IB.
  unfold objRespToMap in Hp; lle Hp.
  assert (Hp' : h ||- rget s rscan |->_ ** top **
    (rget s rscan |->_ ** top) ** top).
  scm.
  unfold fieldRespToMap in Hp; unfold fieldRespMap in Hp; llei Hp; eauto.
  hyr Hp (fasElim IB Hp).
  simpl in Hp; lle Hp; scm.
  unfold fieldRespToMap in Hp; unfold fieldRespMap in Hp; llei Hp; eauto.
  lle Hp'.

  assert (NIsc : ~ set_In (rget s rscan) (set_inter G B)).
  intro I; apply IscanB; destruct (set_inter_elim I); assumption.

  sconjLift5 Hp; hyr Hp (sconjAssocR Hp).
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  assert (Hp' : h ||-
    (foralls x in B',
      objIsToFwded (map_union (map_union phi phi') phi'') phi3 s'' x) **
    objRespToMap (map_union (map_union phi phi') phi'') s (rget s rscan) **
    (foralls x in set_inter (set_remove (rget s rscan) G) B,
      objRespToMap (map_union (map_union (map_union phi phi') phi'') phi3) s'' x) **
    (foralls x in set_inter (set_union G' G'') B,
      objRespToMap (map_union (map_union (map_union phi phi') phi'') phi3) s'' x)).
  sconjMatch.
  apply fasSplit.
  unfold hpredTrue.

  setoid_rewrite <- (set_unionInterDistrib1 (set_remove (rget s rscan) G)
    (set_union G' G'') B).
  setoid_rewrite Geq3.
  setoid_rewrite <- (selfInterEqv B G3).
  assumption.

  (* set_disj (set_inter (set_remove (rget s rscan) G) B)
       (set_inter (set_union G' G'') B) *)
  apply set_disj_intro.
  intros y I1 I2.
  assert (I1' := set_inter_elimL I1); clear I1.
  assert (I2' := set_inter_elimL I2); clear I2.
  destrSF SF''.

  assert (Hp' : s'' |||- top ** top ** top ** top ** 
    (foralls x in G'', x |->_ ** top) ** top ** (foralls x in G', x |->_ ** top) **
    (foralls x in set_remove (rget s rscan) G, x |->_ ** top) ** top).
  sconjMatch.
  eapply fasImpl; eauto.
  unfold objCopiedPhi; unfold objCopied; unfold fieldCopied; 
    intros yzz Z Hpp; linLogElim Hpp; linLogIntro; sconjMatch.
  eapply fasImpl; eauto.
  unfold objCopiedPhi; unfold objCopied; unfold fieldCopied; 
    intros yzz Z Hpp; linLogElim Hpp; linLogIntro; sconjMatch.
  eapply heapEqvTrue.
  apply Heap.heap_eqvSymm; apply eqvPredEqv; eauto.
  unfold objHp in H'ok; eapply fasImpl; eauto.
  cbv beta; unfold okObjAt; unfold okFieldAt; intros z hh HH; lle HH; lli; sconjMatch.

  lle Hp'; sconjLift3 Hp'.
  destruct (set_union_elim I2') as [I2 | I2]; clear I2'; eapply set_disj_elim.
  2:apply I2.
  2:apply I1'.
  sconjLift3 Hp'; apply (fasFasDisj Hp'); auto.
  2:apply I2.
  2:apply I1'.
  sconjLift2 Hp'; apply (fasFasDisj Hp'); auto.

  apply sconjComm.
  sconjLift3 Hp'; sconjLift3 Hp'; hyr Hp' (sconjAssocR Hp').

  (* on to the next sets *)
  clear Hp; scm.
  apply fasIntro.
  scm.
  replace (map_union phi (map_union (map_union phi' phi'') phi3)) with
    (map_unionnn phi (map_unionnn (map_unionnn phi' phi'') phi3)) by auto.
  unfold hpredTrue; setoid_rewrite PhiEqv.
  auto.
  rewrite <- hpredTrueEq.
  eapply objRespToSubmap.
  apply Hp.
  apply mapUnionSub.
  unfold hpredTrue.
  rewrite (set_interSymm B G).
  rewrite <- (set_removeEq NIsc).
  setoid_rewrite <- (set_removeInterDistrib (rget s rscan) G B).
  rewrite <- hpredTrueEq; eapply fasImplI.
  apply Hp.

  clear h Hp; cbv beta; intros y h Yi Hp.
  idtac.
  unfold hpredTrue.

  setoid_rewrite PhiEqv.
  apply (objCopiedRespToMap Hp).
  rewrite <- Req2; [stSimpl | regNotInListSolver].
  rewrite <- Req; [stSimpl | regNotInListSolver].
  reflexivity.
  rewrite <- Req2; [stSimpl | regNotInListSolver].
  rewrite <- Req; [stSimpl | regNotInListSolver].
  reflexivity.

  destrSF SF''.
  sconjLift8 Hp0.
  eapply hpHpObjCopied; eauto.
  instantiate (1 := top).
  destruct (set_inter_elim Yi) as [Yi' Yi'']; clear Yi.
  hyr H'ok (objHpElim Yi' H'ok); unfold okObjAt in H'ok;
    unfold okFieldAt in H'ok; lle H'ok; lli.
  instantiate (1 := x1).
  instantiate (1 := x0).
  scm.

  intro U; apply NIsc; rewrite (set_interSymm G B); assumption.

  apply fasMerge.
  apply sconjComm; sconjMatch.
  eapply fasImplI.
  eauto.
  clear h Hp; simpl; intros y h I Hp.

  assert (Hp' := objCopiedFwded Hp Iso3 (s := s)).
  assert (II : forall x : POnat.POnat.A, set_In x M3 -> s'' |||- objCopied x s x ** top).
  intros z Iz.
  assert (IW : set_In z W'').
  setoid_rewrite W''eq; sui1; assumption.
  assert (Q := Wcopy _ IW).
  eapply objCopiedTrans.
  instantiate (1 := z).
  instantiate (1 := s').
  apply Wcopy'.
  setoid_rewrite W'eq; sui2; assumption.
  assumption.

  assert (Hp'' := Hp' II); clear Hp Hp' II.

  assert (DMM : set_disj M'' M3).
  apply set_unionDisjElimR with (set_union M M').
  apply set_disjSymm; apply set_unionDisjElimL with W3.
  setoid_rewrite <- W''eq.
  eapply scanFieldStOkWMDisj; eauto.

  eapply objIsToFwdedShift.
  2:instantiate (1 := G' |_| G'').
  2:instantiate (1 := M' |_| M'').
  3:eauto.
  assert (ZZ : h ||- objIsToFwded (map_union (map_union phi phi') phi'') phi3 s y).
  apply Hp''.
  rewrite <- Req2; [stSimpl | regNotInListSolver].
  rewrite <- Req; [stSimpl | regNotInListSolver].
  reflexivity.
  rewrite <- Req2; [stSimpl | regNotInListSolver].
  rewrite <- Req; [stSimpl | regNotInListSolver].
  reflexivity.

  assert (PhiEqv' : mapEqv (map_union 
    (map_union (toSpaceIdMap s) (map_union (map_union phi phi') phi'')) phi3)
    (map_union (map_union (toSpaceIdMap s) (map_union phi (map_union phi' phi''))) phi3)).
  intros z.
  unfold map_union.
  destruct (phi z); reflexivity.

  unfold objIsToFwded in ZZ |- * ; unfold objIsFwded in ZZ |- *;
    unfold fieldIsFwded in ZZ |- * ; lle ZZ; lli; eauto.
  destruct H9 as [v0 [AA BB]]; exists v0; constructor.
  assumption.
  unfold mapX; rewrite <- PhiEqv'; assumption.
  destruct H10 as [v0 [AA BB]]; exists v0; constructor.
  assumption.
  unfold mapX; rewrite <- PhiEqv'; assumption.

  apply iso_union; auto.

  (* copied from above *)
  apply set_unionDisjElimR with M.
  apply set_disjSymm.
  apply set_unionDisjElimL with W''.
  rewrite <- W'eq.
  apply (scanFieldStOkWMDisj SF').

  (* copied from above *)
  apply set_unionDisjElimR with M.
  rewrite <- (set_eqvUnionAssoc M M' M'').
  apply set_disjSymm.
  apply set_unionDisjElimL with W3.
  rewrite <- W''eq.
  apply (scanFieldStOkWMDisj SF'').


  unfold hpredTrue; rewrite (set_interSymm B (G' |_| G'')); 
    rewrite <- hpredTrueEq.
  apply (fasImplI Hp).
  clear h Hp; intros y h Iy Hp.
  destruct (set_inter_elim Iy) as [Iy' _]; clear Iy.
  unfold objIsToFwded.
  destrSF SF''.
  destruct (set_union_elim Iy') as [Iy | Iy]; clear Iy'.

  sconjLift7 Hp0.
  eapply copyMapFwded.
  instantiate (2 := s'').
  instantiate (1 := top ** top).
  apply sconjAssocL.
  scm.
  hyr Hp0 (fasElim Iy Hp0).
  scm.

  assert (Hp0' : h0 ||- objCopiedPhi phi' s y).
  generalize dependent Hp0; unfold objCopiedPhi; 
    unfold objCopied; unfold fieldCopied; stSimpl; auto.

  apply (objCopiedPhiSubmap Hp0').
  intros a b Eqa; unfold map_union; rewrite Eqa; reflexivity.

  replace (map_union (A:=nat) (B:=nat)) with map_unionnn.
  replace (map_union (A:=nat) (B:=nat)) with map_unionnn.
  rreplace (map_union (A:=nat) (B:=nat)) with map_unionnn.
  rreplace (map_union (A:=nat) (B:=nat)) with map_unionnn.
  rreplace (map_unionnn
    (map_unionnn (spaceIdMap (rget s rtoStart) (rget s rtoEnd)) phi)
    (map_union (map_union phi' phi'') phi3)) with
  (map_unionnn
    (map_unionnn (spaceIdMap (rget s rtoStart) (rget s rtoEnd)) phi)
    (map_unionnn (map_unionnn phi' phi'') phi3)).
  unfold hpredTrue.
  assert (X := mapUnionAssoc (spaceIdMap (rget s rtoStart) (rget s rtoEnd)) phi
      (map_unionnn (map_unionnn phi' phi'') phi3)).
  idtac.
  idtac.
  rreplace (mapEqv (A := nat) (B := nat)) with mapEqvnn in X.
  setoid_rewrite X.
  clear X.

  unfold objRespToMap in Hp; unfold fieldRespToMap in Hp.
  replace (spaceIdMap (rget s'' rtoStart) (rget s'' rtoEnd)) with
    (spaceIdMap (rget s rtoStart) (rget s rtoEnd)) in Hp.
  unfold POnat.POnat.A in * .
  assert (X := mapUnionAssoc phi (map_unionnn phi' phi'') phi3);
    rreplace (mapEqv (A := nat) (B := nat)) with mapEqvnn in X; rewrite <- X; clear X.
  assert (X := mapUnionAssoc phi phi' phi'');
    rreplace (mapEqv (A := nat) (B := nat)) with mapEqvnn in X; rewrite <- X; clear X.
  apply Hp.
  rewrite <- Req2; [stSimpl | regNotInListSolver].
  rewrite <- Req2; [stSimpl | regNotInListSolver].
  rewrite <- Req; [stSimpl | regNotInListSolver].
  rewrite <- Req; [stSimpl | regNotInListSolver].
  reflexivity.
  reflexivity.
  reflexivity.

  (* variants of the above *)
  sconjLift5 Hp0.
  eapply copyMapFwded.
  instantiate (2 := s'').
  instantiate (1 := top ** top).
  apply sconjAssocL.
  sconjMatch.
  hyr Hp0 (fasElim Iy Hp0).
  scm.

  assert (Hp0' : h0 ||- objCopiedPhi phi'' s' y).
  generalize dependent Hp0; unfold objCopiedPhi; 
    unfold objCopied; unfold fieldCopied; stSimpl; auto.

  clear h Hp Eq Hp0.

  assert (Hp : h0 ||- objCopiedPhi phi'' s y).
  eapply objCopiedPhiCopied; eauto.
  intros z Iz.
  apply Wcopy'; setoid_rewrite W'eq.
  sui1; assumption.
  clear Hp0'.
  eapply objCopiedPhiSubmap.
  apply Hp.

  rreplace (map_union phi') with (map_unionnn phi').
  rreplace (map_union phi'') with (map_unionnn phi'').
  assert (Q : mapSubnn phi'' (map_unionnn phi' (map_unionnn phi'' phi3))).
  replace (mapSub phi'') with (mapSubnn phi''); [idtac | reflexivity].
  assert (X := mapUnionAssoc phi' phi'' phi3).
  rreplace (mapEqv (A := nat) (B := nat)) with mapEqvnn in X; rewrite <- X; clear X.
  eapply mapSubTrans.
  instantiate (1 := map_unionnn phi' phi'').
  2:unfold mapSubnn; unfold map_unionnn; apply mapUnionSub.
  
  eapply mapSub_isoDisj; eauto.
  apply set_disjSymm; apply set_unionDisjElimL with W''; apply set_disjSymm.
  apply set_unionDisjElimR with M.
  setoid_rewrite <- W'eq.
  apply set_disjSymm.
  eapply scanFieldStOkWMDisj; eauto.
  intros a b I.
  hyr Q (Q a b I).
  unfold map_unionnn in Q |- *; unfold map_union in Q |- * .
  destruct (phi' a).
  assumption.
  destruct (phi'' a); assumption.

  clear Hp0.
  unfold objRespToMap in Hp; unfold fieldRespToMap in Hp.
  replace (map_union (map_union (spaceIdMap (rget s rtoStart) (rget s rtoEnd)) phi)
    (map_union (map_union phi' phi'') phi3)) with
    (map_unionnn (map_unionnn (spaceIdMap (rget s rtoStart) (rget s rtoEnd)) phi)
      (map_unionnn (map_unionnn phi' phi'') phi3)).
  2:reflexivity.
  replace (spaceIdMap (rget s rtoStart) (rget s rtoEnd)) with
    (spaceIdMap (rget s'' rtoStart) (rget s'' rtoEnd)).
  unfold hpredTrue.
  assert (X := (mapUnionAssoc (spaceIdMap (rget s'' rtoStart) (rget s'' rtoEnd))
    phi (map_unionnn (map_unionnn phi' phi'') phi3))).
  rreplace (mapEqv (A := nat) (B := nat)) with mapEqvnn in X; rewrite X; clear X.
  assert (X := (mapUnionAssoc phi (map_unionnn phi' phi'') phi3)).
  rreplace (mapEqv (A := nat) (B := nat)) with mapEqvnn in X; rewrite <- X; clear X.
  assert (X := (mapUnionAssoc phi phi' phi'')).
  rreplace (mapEqv (A := nat) (B := nat)) with mapEqvnn in X; rewrite <- X; clear X.
  apply Hp.
  unfold spaceIdMap at 1.
  repeat (rewrite <- Req2; [stSimpl | regNotInListSolver]).
  repeat (rewrite <- Req; [stSimpl | regNotInListSolver]).
  reflexivity.

  rewrite (set_interSymm B (G' |_| G'')).
  hyr Hp (sconjAssocL (unitIR Hp)).
  apply set_disjSymm; apply (fasFasDisj Hp).
  unfold objIsToFwded; unfold objIsFwded; unfold fieldIsFwded.
  intros h0 x0 Hp0; lle Hp0; lli; sconjMatch.
  unfold objRespToMap; unfold fieldRespToMap; unfold fieldRespMap.
  intros h0 x0 Hp0; lle Hp0; lli; sconjMatch.
 
  sconjLift2 Hp; sconjLift4 Hp; sconjLift4 Hp; hyr Hp (sconjAssocR Hp); apply sconjAssocL.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.

  assert (Hp' : h ||- 
    ((foralls x in set_inter G G3, objCopied x s'' x) **
    (foralls x in set_inter (G' |_| G'') G3, objCopied x s'' x)) **
    (foralls x in G4, objCopiedPhi phi3 s'' x)).
  sconjMatch.
  assert (Hp' : h ||- 
    foralls x in (set_inter G G3) |_| (set_inter (G' |_| G'') G3),
      objCopied x s'' x).
  unfold hpredTrue.
  setoid_rewrite <- (set_unionInterDistrib1 G (G' |_| G'') G3).

  assert (G3eq : set_subset G3 (G |_| (G' |_| G''))).
  intros y I.
  destruct (Geq3 y) as [_ X].
  assert (U : set_In y (B |_| G3)).
  sui2; assumption.
  assert (Q := X U); clear I X U.
  destruct (set_union_elim Q) as [Q' | Q']; clear Q.
  destruct (set_remove_elim _ _ _ Q').
  sui1; assumption.
  sui2; assumption.

  setoid_rewrite (selfInterEqv2 G3eq).
  assumption.

  apply fasSplit.
  assumption.

  (* set_disj (set_inter G G3) (set_inter (G' |_| G'') G3) *)
  clear h Hp Hp' Wcopy Wcopy' Meqv PhiEqv NIsc.
  destrSF SF''.

  assert (HX : s'' |||- top ** top ** top ** top **
    (foralls x in G'', x |->_ ** top) ** rget s rscan |->_ **
    (foralls x in G', x |->_ ** top) ** 
    (foralls x in set_remove (rget s rscan) G, objCopied x s x) ** top).
  sconjMatch.
  eapply fasImpl.
  apply Hp.
  unfold objCopiedPhi; unfold objCopied; unfold fieldCopied; 
    intros yzz Z Hpp; linLogElim Hpp; linLogIntro; sconjMatch.
  unfold fieldRespToMap in Hp; unfold fieldRespMap in Hp.
  linLogElim Hp; linLogIntro; sconjMatch.
  eapply fasImpl.
  apply Hp.
  unfold objCopiedPhi; unfold objCopied; unfold fieldCopied; 
    intros yzz Z Hpp; linLogElim Hpp; linLogIntro; sconjMatch.
  clear Req Req2 Req'' NSP.
  eapply heapEqvTrue.
  apply Heap.heap_eqvSymm; apply eqvPredEqv; eauto.
  unfold objHp in H'ok; eapply fasImplIS; eauto.
  cbv beta; unfold okObjAt; unfold okFieldAt; intros z hh _ Sub HH; lle HH.
  assert (QQ : s |||- (z |-> x0 ** 4 + z |-> x1) ** top).
  eapply subHeapImpl.
  eauto.
  eapply Heap.subHeapTrans; eauto.
  eapply sconjEqSubHeap; eauto.
  lle QQ.
  unfold objCopied; unfold fieldCopied; lli.
  2:eauto.
  sconjMatch.
  sconjLift2 QQ; sconjMatch.

  linLogElim HX.
  apply set_disjInterIntro.
  clear Hp; rename HX into Hp.
  sconjLift4 Hp; sconjLift3 Hp; hyr Hp (sconjAssocR Hp).
  eapply fasFasDisj.
  instantiate (1 := top).
  instantiate (1 := (fun x => x |->_ ** top)).
  instantiate (1 := (fun x => x |->_ ** top)).
  instantiate (1 := heapOf s'').
  apply (sconjImpl Hp); clear Hp; intros h Hp.
  assert (IscanG : set_In (rget s rscan) G).
  setoid_rewrite Geqv.
  destruct Al1.
  eapply rangeObjsIntro.
  omega.
  instantiate (1 := 0).
  omega.
  unfold hpredTrue.
  setoid_rewrite <- (addRemoveEqv IscanG).
  apply fasIntro.
  lli; scm.

  apply topIR.
  eapply fasImpl.
  apply Hp.
  unfold objCopied; unfold fieldCopied; 
    intros yzz Z Hpp; llei Hpp; scm.
  intro II; destruct (set_remove_elim _ _ _ II) as [X _].
  apply X; reflexivity.

  hyr Hp (sconjAssocR Hp); sconjMatch.

  apply fasMerge'.
  apply sconjComm; assumption.
  auto.
  auto.
  auto.
  clear Hp; rename Hp' into Hp.
  linLogElim Hp.
  sconjMatch.
  eapply fasImplI.
  apply Hp.
  clear h Hp; simpl; intros y h Iy Hp.
  assert (NI : ~ set_In (rget s rscan) G3).
  intro Iy'.
  destruct (Geq3 (rget s rscan)) as [_ X].
  assert (Q : set_In (rget s rscan) (B |_| G3)).
  sui2; assumption.
  hyr X (X Q); clear Q.
  destruct (set_union_elim X) as [Y | Y]; clear X.
  destruct (set_remove_elim _ _ _ Y) as [Z _]; apply Z; reflexivity.
  destruct (set_union_elim Y) as [X | X]; clear Y.
  destrSF SF'.
  sconjLift5 Hp0; sconjLift5 Hp0.
  unfold fieldRespToMap in Hp0; unfold fieldRespMap in Hp0; linLogElim Hp0.
  assert (HX : s' |||- rget s rscan |->_ ** (rget s rscan |->_ ** top) ** top).
  sconjMatch.
  linLogIntro; eauto.
  hyr Hp0 (fasElim X Hp0).
  simpl in Hp0.
  unfold objCopiedPhi in Hp0; unfold objCopied in Hp0; linLogElim Hp0.
  sconjMatch.
  unfold fieldCopied in Hp0; linLogElim Hp0; linLogIntro; eauto.
  lle HX.

  destrSF SF''.
  sconjLift5 Hp0; sconjLift6 Hp0.
  unfold fieldRespToMap in Hp0; unfold fieldRespMap in Hp0; linLogElim Hp0.
  assert (HX : s'' |||- rget s rscan |->_ ** (rget s rscan |->_ ** top) ** top).
  sconjMatch.
  linLogIntro; eauto.
  hyr Hp0 (fasElim X Hp0).
  simpl in Hp0.
  unfold objCopiedPhi in Hp0; unfold objCopied in Hp0; linLogElim Hp0.
  sconjMatch.
  unfold fieldCopied in Hp0; linLogElim Hp0; linLogIntro; eauto.
  linLogElim HX.

  assert (NE : y <> rget s rscan).
  intro; subst y; apply NI; destruct (set_inter_elim Iy); assumption.
  destrSF SF''.
  sconjLift8 Hp0.
  assert (Hp' : s'' |||- (objCopied y s y ** top) ** top).
  sconjMatch.
  eapply heapEqvTrue.
  apply Heap.heap_eqvSymm; apply eqvPredEqv; eauto.
  assert (II : set_In y (set_remove (rget s rscan) G)).
  destruct (set_inter_elim Iy); apply set_remove_intro; assumption.
  hyr H'ok (objHpElim II H'ok); unfold okObjAt in H'ok; unfold okFieldAt in H'ok;
    lle H'ok.
  assert (QQ : s |||- (y |-> x0 ** 4 + y |-> x1 ** top) ** top).
  eapply subHeapImpl.
  instantiate (1 := h'); sconjMatch.
  eapply sconjEqSubHeap; eauto.
  lle QQ.
  unfold objCopied; unfold fieldCopied; lli.
  instantiate (1 := x0); scm.
  instantiate (1 := x1); scm.
  sconjLift2 QQ; sconjMatch.

  lle Hp'; eapply objCopiedTrans2; eauto.
  apply fasMerge'.
  sconjMatch.
  eapply fasImplI.
  apply Hp.
  clear h Hp.
  simpl; intros y h I Hp.

  eapply objCopiedPhiCopied'.
  apply Hp.
  clear h Hp.
  destrSF SF''.
  assert (Hp' : s'' |||- top ** top ** top **
    top ** (foralls x in G'', objCopiedPhi phi'' s' x) **
    top ** (foralls x in G', objCopiedPhi phi' s x) ** top).
  sconjMatch.
  linLogElim Hp'.
  assert (Hp'' : s'' |||- 
    (objCopiedPhi (map_union phi' (map_union phi'' phi3)) s y ** top) ** top).
  destruct (set_inter_elim I) as [I1 I2].
  destruct (set_union_elim I1) as [IA | IB]; clear I I1.
  sconjLift2 Hp'; sconjMatch.
  hyr Hp0 (fasElim IA Hp0).
  sconjMatch.
  eapply objCopiedPhiSubmap.
  apply Hp0.
  apply mapUnionSub.
  sconjMatch.
  hyr Hp0 (fasElim IB Hp0).
  sconjMatch.
  eapply objCopiedPhiSubmap.
  instantiate (1 := phi'').
  eapply objCopiedPhiCopied; eauto.
  intros z Iz.
  apply Wcopy'; setoid_rewrite W'eq.
  sui1; assumption.
  rreplace (map_union phi') with (map_unionnn phi').
  rreplace (map_union phi'') with (map_unionnn phi'').
  assert (Q : mapSubnn phi'' (map_unionnn phi' (map_unionnn phi'' phi3))).
  rreplace (mapSub phi'') with (mapSubnn phi'').
  assert (X := mapUnionAssoc phi' phi'' phi3).
  rreplace (mapEqv (A := nat) (B := nat)) with mapEqvnn in X;
    rewrite <- X; clear X.

  eapply mapSubTrans.
  instantiate (1 := map_unionnn phi' phi'').
  2:unfold mapSubnn; unfold map_unionnn; apply mapUnionSub.
  eapply mapSub_isoDisj; eauto.
  apply set_disjSymm; apply set_unionDisjElimL with W''; apply set_disjSymm.
  apply set_unionDisjElimR with M.
  setoid_rewrite <- W'eq.
  apply set_disjSymm.
  eapply scanFieldStOkWMDisj; eauto.
  apply Q.

  lle Hp''; scm.
  eapply objCopiedPhiSubmap; eauto.
  intros a b II.
  replace (map_union (map_union phi' phi'') phi3) with
    (map_unionnn (map_unionnn phi' phi'') phi3); [idtac | reflexivity].
  rewrite mapUnionAssoc; assumption.

  eapply fasImplI.
  apply Hp.
  simpl; clear h Hp; intros y h I Hp.
  eapply objCopiedPhiSubmap.
  instantiate (1 := phi3).
  eapply objCopiedPhiCopied; eauto.
  intros z Iz.
  eapply objCopiedTrans.
  2:apply Wcopy''.
  apply Wcopy'; setoid_rewrite W'eq; setoid_rewrite W''eq.
  sui2; sui1; assumption.
  setoid_rewrite W''eq; sui1; auto.
  
  replace (map_union phi') with (map_unionnn phi') by reflexivity.
  replace (map_union phi'') with (map_unionnn phi'') by reflexivity.
  assert (Q : mapSubnn phi3 (map_unionnn phi' (map_unionnn phi'' phi3))).
  assert (X := mapUnionAssoc phi' phi'' phi3).
  rreplace (mapEqv (A := nat) (B := nat)) with mapEqvnn in X; 
    rewrite <- X; clear X.
  eapply mapSub_isoDisj; eauto.
  instantiate (1 := set_union G' G'').
  instantiate (1 := set_union M' M'').
  apply iso_union; auto.

  apply set_unionDisjElimR with M.
  apply set_disjSymm; apply set_unionDisjElimL with W''.
  setoid_rewrite <- W'eq.
  eapply scanFieldStOkWMDisj; eauto.

  apply set_unionDisjElimR with M.
  rewrite <- (set_eqvUnionAssoc M M' M'').
  apply set_disjSymm.
  apply set_unionDisjElimL with W3.
  rewrite <- W''eq.
  apply (scanFieldStOkWMDisj SF'').

  intros a b EE.
  replace (map_union (map_unionnn phi' phi'') phi3 a) with
    (map_unionnn (map_unionnn phi' phi'') phi3 a) by reflexivity.
  rewrite mapUnionAssoc; eapply Q; eauto.

  unfold objCopiedPhi; unfold objCopied; unfold fieldCopied; 
    intros yzz Z Hpp; llei Hpp; scm.

  replace (rget s rgcInfo) with (rget s'' rgcInfo).
  assumption.
  rewrite <- Req2; [stSimpl | regNotInListSolver].
  rewrite <- Req; [stSimpl | regNotInListSolver].
  reflexivity.

  (* show the register file equivalences *) 
  clear Guar.
  destruct Guar' as [s'' [[_ Req'] Guar'']].
  destruct Guar'' as [_ [Req'' [raEq [allocEq v0Eq]]]].
  constructor.

  intros r I.
  rewrite <- Req''; [stSimpl | assumption].
  simpl in I; decompose [and] I; rewrite <- Req'; [idtac | regNotInListSolver].
  stSimpl; rewrite <- Req.
  stSimpl; reflexivity.
  regNotInListSolver.
  rewrite raEq; rewrite v0Eq; rewrite <- allocEq; clear allocEq raEq v0Eq.
  repeat (rewrite <- Req'; [stSimpl | regNotInListSolver]).
  repeat (rewrite <- Req; [stSimpl | regNotInListSolver]).
  intuition.

Qed.


Lemma bakerInductiveOk : forall s,
  bakerLoopPre s ->
  rget s rscan <> rget s rfree ->
  scanFieldPre
     (rset (rset (rset (rset s ast scan_per_gc) ast 1) a0 (rget s rscan)) ra
        BAKERLOOP2) /\
   (forall s' : state,
    scanFieldGuar
      (rset (rset (rset (rset s ast scan_per_gc) ast 1) a0 (rget s rscan)) ra
         BAKERLOOP2) s' ->
    bakerLoop2Pre s' /\
    (forall s'' : state, bakerLoop2Guar s' s'' -> bakerLoopGuar s s'')).
Proof.
  intros s Pre Ne.
  destruct Pre as [phi [G [W [M [F [Geq BLok]]]]]].
  destrBL BLok.

  assert (IG : set_In (rget s rscan) G).
  setoid_rewrite Geq.
  destruct Al1.
  eapply rangeObjsIntro.
  omega.
  instantiate (1 := 0).
  omega.

  constructor.

  (* showing it is safe to make first call to scanField *)
  unfold scanFieldPre; stSimpl.
  exists phi; exists W; exists M; exists F.
  unfold scanFieldStOk; stSimpl; intuition.
  unfold toSpace; stSimpl.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  replace (setRange (rget s rtoStart) (rget s rtoEnd)) with
    (toSpace s) by auto.

  assert (h ||- (okFieldAt ((W |_| M) |_| (toSpace s)) (rget s rscan) **
    top) ** top).
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  hyr Hp (objHpElim IG Hp).
  unfold okObjAt in Hp.
  lle Hp; scm.
  lli.
  lle H; auto.
  (* first precondition shown *)

  unfold scanFieldGuar; stSimpl.
  unfold toSpace; stSimpl.
  replace (setRange (rget s rtoStart) (rget s rtoEnd)) with (toSpace s) by auto.
  intros s' [G1 Req1].

  constructor.

  (* showing it is safe to jump to bakerLoop2 *)
  hyr G1 (G1 phi W M F (okFieldAt ((W |_| M) |_| (toSpace s)) 
    (4 + rget s rscan) ** 
    objHp ((W |_| M) |_| (toSpace s)) (set_remove (rget s rscan) G) ** 
    gcHpPtrMem (rget s rgcInfo) ** top)).
  destruct G1 as [M'' [W' [F' [G0 [phi' HH]]]]].
  unfold scanFieldStOk; stSimpl; intuition.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocL; apply sconjAssocL.
  sconjMatch.
  hyr Hp (objHpElim IG Hp).
  assumption.

  destruct HH as [Weq [Feq' [Iso [AlF' [Weq' SF]]]]].
  unfold bakerLoop2Pre.
  rewrite <- Req1; [stSimpl | regNotInListSolver].
  rewrite <- Req1; [stSimpl | regNotInListSolver].

  assert (WMeqv' : set_eqv (W |_| M) (set_union W' (set_union M M''))).
  setoid_rewrite Weq.
  setoid_rewrite (set_eqvUnionSymm M'' W').
  setoid_rewrite (set_eqvUnionAssoc W' M'' M).
  setoid_rewrite (set_eqvUnionSymm M'' M).
  apply set_eqvRefl.

  constructor.
  unfold scanFieldPre; stSimpl.
  unfold scanFieldStOk; stSimpl.
  clear Hp.
  destrSF SF.
  exists (map_union phi phi'); exists W'; exists (set_union M M''); exists F'.
  intuition.
  unfold toSpace; stSimpl.
  repeat (rewrite <- Req1; [stSimpl | regNotInListSolver]).
  replace (setRange (rget s rtoStart) (rget s rtoEnd)) with (toSpace s) by auto.
  apply (sconjImplR Hp); clear Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  sconjLift3 Hp.
  sconjMatch.
  unfold hpredTrue; setoid_rewrite <- WMeqv'; assumption.

  (* show it is okay to loop around again *)
  unfold scanFieldGuar; unfold toSpace; stSimpl.
  intros s'' [Guar Req2].
  destruct (Guar (map_union phi phi') W' (set_union M M'') F'
    (fieldRespToMap (map_union phi phi')
      (rset (rset (rset (rset s ast scan_per_gc) ast 1) a0 (rget s rscan)) ra BAKERLOOP2)
      (rget s rscan) **
    (foralls x in G0, objCopiedPhi phi'
      (rset (rset (rset (rset s ast scan_per_gc) ast 1) a0 (rget s rscan)) ra BAKERLOOP2) x) **
    objHp ((W |_| M) |_| toSpace s) (set_remove (rget s rscan) G) ** 
      gcHpPtrMem (rget s rgcInfo) ** top)) as [M3 [W3 [F3 [G3 [phi'' HH]]]]]; clear Guar.

  generalize dependent SF.
  unfold scanFieldStOk; unfold fromSpace; unfold toSpace; stSimpl.
  clear WMdisj WMsub Feq NSP Le1 Le2 Le3 AlF Hp.
  repeat (rewrite <- Req1; [stSimpl | regNotInListSolver]).
  intro; destrSF SF.
  intuition.
  sconjStepX.
  unfold hpredTrue; setoid_rewrite <- WMeqv'; assumption.

  assert (FeqX : set_eqv F' (rangeObjs (rget s' rfree) (rget s ralloc))).
  destrSF SF.
  rewrite <- (Req1 ralloc) in Feq0; [assumption | regNotInListSolver].

  rename SF into SF'.
  destruct HH as [Weq'' [Feq'' [Iso' [AlF'' [Weq''' SF]]]]].
  unfold bakerLoopPre.
  exists (map_union (map_union phi phi') phi'').
  exists (rangeObjs (rget s'' rscan) (rget s'' rfree)).
  exists W3.
  exists (set_union (set_union M M'') M3).
  exists F3.
  constructor.
  apply set_eqvRefl.

  unfold bakerLoopStOk.
  rewrite <- Req2; [stSimpl | regNotInListSolver].
  rewrite <- Req2; [stSimpl | regNotInListSolver].
  rewrite <- (Req2 rgcInfo); [stSimpl | regNotInListSolver].

  constructor.
  rewrite <- Req1; [stSimpl | regNotInListSolver].
  omega.
  constructor.
  destruct Al1 as [k1 R1]; destruct AlF' as [k2 R2]; destruct AlF'' as [k3 R3].
  destruct k1.
  elimtype False; omega.
  rewrite R3; rewrite R2; rewrite R1; exists (k1 + k2 + k3); omega.

  assert (WMeqv : set_eqv (set_union W3 (set_union (set_union M M'') M3)) (W |_| M)).
  setoid_rewrite (set_eqvUnionSymm (set_union M M'') M3).
  setoid_rewrite <- (set_eqvUnionAssoc W3 M3 (set_union M M'')).
  setoid_rewrite (set_eqvUnionSymm W3 M3).
  setoid_rewrite <- Weq''.
  setoid_rewrite (set_eqvUnionSymm M M'').
  setoid_rewrite <- (set_eqvUnionAssoc W' M'' M).
  setoid_rewrite (set_eqvUnionSymm W' M'').
  setoid_rewrite <- Weq.
  apply set_eqvRefl.

  eapply scanFieldStOkImpl.
  eauto.
  clear Weq''' Weq'; rename Hp into HpS.
  intros h Hp.
  rewrite <- Req1; [stSimpl | regNotInListSolver].
  sconjLift6 Hp; sconjLift3 Hp; sconjLift5 Hp; sconjLift6 Hp.
  hyr Hp (sconjAssocR Hp); hyr Hp (sconjAssocR Hp).
  sconjMatch.

  assert (Geqv' : set_eqv G0 (rangeObjs (rget s rfree) (rget s' rfree))).
  assert (X : set_eqv (rangeObjs (rget s rfree) (rget s ralloc)) (G0 |_| F')).
  rewrite <- Feq; assumption.
  eapply rangeSplit.
  instantiate (1 := rget s ralloc).
  rewrite <- FeqX.
  apply set_disjSymm; eapply scanFieldStOkFGDisj.
  apply (scanFieldStOkImpl SF').
  intros hx HpX.
  sconjLift2 HpX; sconjMatch.
  simpl; intros h0 x; unfold objCopiedPhi.
  unfold objCopied; unfold fieldCopied; intro HpZ; linLogElim HpZ; linLogIntro; sconjMatch.
  destruct AlF''; destruct AlF'.
  destrSF SF'.
  rewrite <- (Req1 ralloc) in Le5; [ | regNotInListSolver].
  generalize dependent Le5; stSimpl; auto.
  rewrite <- FeqX.
  rewrite <- Feq'; assumption.
  assumption.

  assert (Geqv'' : set_eqv G3 (rangeObjs (rget s' rfree) (rget s'' rfree))).
  assert (X : set_eqv (rangeObjs (rget s' rfree) (rget s ralloc)) (G3 |_| F3)).
  rewrite <- Feq''; apply set_eqvSymm; assumption.
  assert (RA : rget s'' ralloc = rget s ralloc).
  rewrite <- (Req2 ralloc); [stSimpl | regNotInListSolver].
  rewrite <- (Req1 ralloc); [stSimpl | regNotInListSolver].
  reflexivity.
  eapply rangeSplit.
  instantiate (1 := rget s ralloc).
  assert (Q := SF).
  destrSF SF.
  rewrite RA in Feq0.
  rewrite <- Feq0.
  apply set_disjSymm; eapply scanFieldStOkFGDisj.
  apply (scanFieldStOkImpl Q).
  intros hx HpX.
  sconjLift2 HpX; sconjMatch.
  simpl; intros h0 x; unfold objCopiedPhi.
  unfold objCopied; unfold fieldCopied; intro HpZ; linLogElim HpZ; linLogIntro; sconjMatch.
  destrSF SF.
  rewrite RA in Le5; assumption.
  destrSF SF.
  rewrite RA in Feq0; rewrite <- Feq0.
  apply set_eqvSymm; assumption.
  assumption.

  assert (HH1 : aligned8 (8 + rget s rscan) (rget s' rfree)).
  destruct Al1 as [k1 Al1]; destruct AlF' as [k2 AlF'].
  destruct k1.
  elimtype False; omega.
  exists (k1 + k2); omega.

  assert (HH2 : rget s' rfree <= rget s'' rfree).
  destrSF SF'; destruct AlF''; omega.

  unfold hpredTrue; rewrite <- (rangeObjsTransEqv HH1 HH2).
  rewrite <- Geqv''.
  clear HH1 HH2 Geqv''.

  assert (HH1 : aligned8 (8 + rget s rscan) (rget s rfree)).
  destruct Al1 as [k1 Al1]; destruct k1; [elimtype False | exists k1]; omega.
  assert (HH2 : rget s rfree <= rget s' rfree).
  destruct AlF'; omega.

  rewrite <- (rangeObjsTransEqv HH1 HH2).
  rewrite <- Geqv'.
  rewrite WMeqv.
  replace (toSpace s'') with (toSpace s).
  clear HH1 HH2 Geqv'.
  apply objHpMerge; sconjMatch.
  apply objHpMerge; sconjMatch.
  unfold hpredTrue; rewrite <- (rangeRemoveEqv (rget s rscan) (rget s rfree));
    rewrite <- Geq; assumption.

  eapply objHpIso.
  apply Hp.
  2:eauto.
  instantiate (1 := objHp ((W |_| M) |_| toSpace s) (set_diff W M'') ** top).
  stSimpl; apply sconjAssocL; sconjLift2 HpS; scm.
  apply objHpSplit.
  assert (Sub : set_subset M'' W).
  setoid_rewrite Weq.
  apply set_subsetUnionIntroL; apply set_subsetRefl.
  unfold hpredTrue; rewrite (set_eqvUnionSymm M'' (set_diff W M'')).
  rewrite (set_diffUnionEq Sub).
  apply (objHpWeaken Hp0).
  intros x Ix; sui1; auto.
  apply set_disj_intro; intros x Ix Ix'; destruct (set_diff_elim Ix'); auto.

  eapply objHpIso.
  apply Hp.
  2:eauto.
  stSimpl.
  instantiate (1 := objHp ((W |_| M) |_| toSpace s) (set_diff W' M3) ** top).
  destrSF SF'.
  stSimpl; apply sconjAssocL; sconjLift2 Hp0; scm.
  apply objHpSplit.
  assert (Sub : set_subset M3 W').
  setoid_rewrite Weq''.
  apply set_subsetUnionIntroL; apply set_subsetRefl.
  unfold hpredTrue; rewrite (set_eqvUnionSymm M3 (set_diff W' M3)).
  rewrite (set_diffUnionEq Sub).
  apply (objHpWeaken Hp0).
  setoid_replace (W' |_| M |_| M'') with (W |_| M).
  intros x Ix; sui1; auto.
  rewrite set_eqvUnionSymm. rewrite set_eqvUnionAssoc. rewrite <- Weq.
  rewrite set_eqvUnionSymm; apply set_eqvRefl.
  apply set_disj_intro; intros x Ix Ix'; destruct (set_diff_elim Ix'); auto.

  unfold toSpace.
  repeat (rewrite <- Req2; [stSimpl | regNotInListSolver]).
  repeat (rewrite <- Req1; [stSimpl | regNotInListSolver]).
  reflexivity.

  (* finally, show that adding on a run of loop2 is a run of the whole loop *)
  clear phi G W M F Geq Le0 Al1 WMdisj WMsub Feq NSP Le1 Le2 Le3 AlF Hp IG.
  intros; eapply bakerInductiveGuarOk; eauto.

Qed.


Lemma bakerLoopOk : iseqOkS Pbaker bakerLoopPre bakerLoopGuar bakerLoop.
Proof.
  unfold bakerLoop.
  iseqWeakSolver PbakerSimpl.

  exact scanFieldGuarPresRa.

  intros s Pre.
  cbv beta.
  unfold doBtest.
  compare (rget s rscan) (rget s rfree); intro E.

  rewrite E; rewrite beq_true.
  apply loopExitOk; assumption.

  (* first test passed *)
  rewrite (beq_false E).
  rewrite addiuStep.
  rewrite sltuStep; stSimpl.
  replace (scan_per_gc + 0) with scan_per_gc by omega.
  rewrite addiuStep; stSimpl.
  replace (0 + rget s rscan) with (rget s rscan) by omega.

  assert (D : rget s rcount < scan_per_gc \/ rget s rcount >= scan_per_gc).
  omega.

  destruct D as [Lt | Ge].
  rewrite (blt_true Lt).
  rewrite beq_false.
  2:auto.

  2:rewrite (blt_false Ge).
  2:rewrite beq_true.

  apply callChain.
  apply bakerInductiveOk; assumption.
  apply loopExit2Ok; assumption.

Qed.


(* same as for Cheney *)
Module BakerGCGuar <: BASIC_MGI_ARG.

  Definition gcGuar (s s':state) (rr:list rnum) := 
    forall objs h,
      s |||- (minObjHp (List.map (rget s) rr) objs && heap_eqv h) ** top ->
	exists h',
	  s' |||- heap_eqv h' ** top /\
	  isoState rr (h, rfileOf s) (h', rfileOf s').

End BakerGCGuar.


Module BMGI := BasicMGI BakerGCGuar.



Definition repr vs rf rr :=
  let g := RFile.get rf in
  let gcInfo  := g rgcInfo in
(*  let roots   := List.map g rr in  roots isn't actually used, oddly *)
    lexists frStart, lexists frEnd, lexists toStart, lexists toEnd, 
    lexists scan, lexists free, lexists alloc,
    let B  := rangeObjs toStart scan in
    let B' := rangeObjs alloc toEnd in
    let G  := rangeObjs scan free in
    let toObjs := B |_| G |_| B' in
      lexists phi, lexists W, lexists M, lexists h,
      !(~In rgcInfo rr /\
	aligned8s (toStart::scan::free::alloc::toEnd::nil) /\
	aligned8 frStart frEnd /\
	set_eqv (W |_| M) (rangeObjs frStart frEnd) /\
	iso_set phi M (B |_| G) /\
	(forall r, In r rr -> 
	  rget vs r = RFile.get rf r /\ okFieldVal toObjs (g r)) /\
	heap_eqv (heapOf vs) (Heap.HeapMap.map_mapRange 
	  (fun x => match phi x with Some y => y | None => x end) h)) **
      ((heap_eqv h &&
	objHp (W |_| M) W **
	objHp toObjs (B |_| B') ** 
	objHp ((W |_| M) |_| toObjs) G) ** 
      mapHp M phi ** 
      (foralls x in rangeObjs free alloc, x |->__) **
      gcInfo |->, (frStart::frEnd::toStart::toEnd::scan::free::alloc::nil)).



(*

before flipping, if
  size(fromSpace) < (1 + m)[size(tospace) - (T - B)]

holds then may run out of memory during collection.

"T is top of tospace"

B "points to bottom of free area"

m is the ratio of allocation to scanning.  which is 1/k.

I think you want to avoid the situation where you realize your current
space is too small in the middle of a cycle.  This would require
allocating things in a separate space, which probably messes up the
ability to easily tell what space a pointer is in.  You can end up in
a situation where you have spillover objects for both spaces, so you
end up having to do more checks.  But maybe the size of the spillover
space is bounded, so you can still do the checks in constant time.
But this is still pretty gross.  So let's assume you don't want to run
out of space during a collection.

Based on the other formula, to prevent an out of memory error, we need
this invariant:

  size(fromSpace) >= (1 + m)[size(tospace) - (T - B)]

T - B is the number of free objects.
size(toSpace) is presumably the number of objects toSpace can hold,
  or this doesn't make any sense at all.

So size(tospace) - (T - B) is the number of currently allocated objects.

If T == B, as in the example code, then we need m == 0 to ensure we
don't run out of space, if the spaces have the same size.  I think
this corresponds to scanning everything at every allocation.  This is
slightly funny, but I suppose the problem is that T isn't always B.
But once it is, we have to scan everything.  So really, in order to
avoid having to increase the size of the space every time, we should
stop before the heap is entirely full.  But how early?

Assume we never increase the size of the space.  It should be possible
to include that without interfering with other things.

The size of the space is x.  T-B is n.  k = 1/m. Assume x > n > 0.

x >= (1 + m) (x - n)
x >= x - n + m(x - n)

n >= m(x - n)

n / (x - n) >= m

1 / m >= (x - n) / n

k >= (x - n) / n

or, solving for n,

n >= mx - mn
n + mn >= mx
n(1 + m) >= mx
n >= mx/ (1 + m)
n >= x / (k + 1)

To make allocation time constant, we have to fix k, which requires
solving for n.  Though I guess that changes when the GC triggers, so I
don't know how all that works out.

But basically, it looks like we can set things up such that no check
is required during allocation, and that this is probably the way to
go, even though it may result in more float or whatever.


If we change scanField to use notFromSpacePtr instead of notToSpacePtr
I think we use scanField for both Cheney and Baker without alteration.
I got this from Baker's paper.

So the problem is that we still need to test for notToSpacePtr, to
distinguish fwded objects from unfwded ones cleanly.  Problem is, this
requires two more registers or two loads.

We can also be a little redundant in our tests:

  notToSpacePtr(x) ::= atom(x) \/ noFromSpacePtr(x)

... if x is atomic or fromspace, or tospace.


A problem occurs when:
  1) every object is reachable
  2) the from space is full
  3) at least one allocation has occured during the cycle

The collector goes along copying, but eventually exhausts the space.

The initial B = T check ensures that 1 more object can be allocated during
the GC increment, but up to 2k objects will need to be allocated, so that
doesn't help.  So you can get in a situation where the check in copyObj
will fail, which means we need the check.

BUT if we don't care if the GC trashes the object heap (not random memory,
just the object heap!), then the "B = T" check after the GC increment will
catch this case.  And we know that copying the objects from the from space
will never overrun the end of the to space.

The problem is that in a more realistic GC when you run out of space,
you will allocate more space, and just move everything over there.

Jones pp204: "As well as switching the roles of the semi-spaces, flip
also checks that the new Tospace is sufficiently large, expanding it
if necessary."

What does "sufficiently large" mean?  It can't mean "large enough to
hold all of the current objects, plus those that can be allocated
during the upcoming cycle", because the algorithm waits until the heap
is entirely full, so this would result in expanding the heap every
time.  I suppose I could trigger collection early enough to hold those
things, plus all of the objects, which would be a little pessimistic,
but would avoid the possibility of running out of space.

*)


(*
???
GC invariant:
  rget s free = rget s alloc ->
  rget s rscan = rget s rfree

In other words, when we run out of space, we're done scanning.

k * (rget s alloc - rget s rfree) = (rget s free - rget s rscan)

maybe?

note that this is something we care about in the function header, not
in the loop itself.


  (*scan_per_gc <= rget s rcount /\ *)


*)

