Require Import CheneyLemma.
Require Import ScanField.
Require Import Heap.
Require Import HeapDom.
Require Import VGC.
Require Import CheneyLoop.


Set Implicit Arguments.


(* code address constants *)
Definition CHENEYALLOC := 10.
Notation CHENEYREAD1 := 20.
Notation CHENEYREAD2 := 30.
Notation CHENEYWRITE1 := 40.
Notation CHENEYWRITE2 := 50.

Definition CHENEYPOSTGC := 15.
Definition CHENEYDOALLOC := 56.
Definition CHENEYENTER := 71.
Definition CHENEYLOOPHEADER := 84.

Definition INFLOOP := 89.


(* local register conventions *)
Definition rroot := t0.
Definition rgcInfo := s7.
Definition raSave2 := t9.


(* code blocks *)

Definition cheneyEnter :=
  (* swap spaces, init rtoStart and rtoEnd *)
  lw rtoStart 0 rgcInfo;
  lw rtemp 8 rgcInfo;
  sw rtemp 0 rgcInfo;
  sw rtoStart 8 rgcInfo;
  lw rtoEnd 4 rgcInfo;
  lw rtemp 12 rgcInfo;
  sw rtemp 4 rgcInfo;
  (* store final space limit later,
     so we can reuse this memory for scanField *)

  (* init rfree and rscan *)
  addiu rfree rtoStart 0;
  addiu rscan rtoStart 0;
  
  (* save ra *)
  addiu raSave ra 0;

  (* scan the root *)
  sw rroot 12 rgcInfo;
  addiu a0 rgcInfo 12;
  jal SCANFIELD CHENEYLOOPHEADER.


Definition cheneyLoopHeader :=
  (* restore updated root *)
  lw rroot 12 rgcInfo;

  (* finish space swap *)
  sw rtoEnd 12 rgcInfo;
  
  j CHENEYLOOP.


Definition cheneyAlloc :=
  addiu raSave2 ra 0;
  lw rfree 16 rgcInfo;
  lw rtemp 12 rgcInfo;
  bne rfree rtemp CHENEYDOALLOC;;
  jal CHENEYENTER CHENEYPOSTGC.


Definition cheneyDoAlloc :=
  addiu v0 zero 1;
  sw v0 0 rfree;
  sw v0 4 rfree;
  addiu v0 rfree 0;
  addiu rfree rfree 8;
  sw rfree 16 rgcInfo;
  addiu ra raSave2 0;
  jr ra.


Definition cheneyPostGC :=
  lw rtemp 12 rgcInfo;
  bne rfree rtemp CHENEYDOALLOC;;
  j INFLOOP.


Definition cheneyRead k :=
  lw v0 k rroot;
  jr ra.

Definition cheneyWrite k :=
  sw a0 k rroot;
  jr ra.

Definition infLoop :=
  j INFLOOP.


(* state specifications *)

Definition cheneyStOk fromSt fromEnd toSt toEnd free root gcInfo h :=
  let objs := rangeObjs toSt free in
    !(aligned8 toSt free /\
      aligned8 free toEnd /\
      okFieldVal objs root /\
      fromEnd - fromSt = toEnd - toSt /\
      fromSt <= fromEnd) **
    (objHp objs objs && heap_eqv h) ** buffer free toEnd **
    buffer fromSt fromEnd **
    gcInfo |->, (fromSt::fromEnd::toSt::toEnd::nil).

Definition repr vs rf rr :=
  lexists fromSt, lexists fromEnd, lexists toSt, lexists toEnd, lexists free,
  let objs := rangeObjs toSt free in
    !(~In rgcInfo rr /\
      aligned8 toSt free /\
      aligned8 free toEnd /\
      (forall r, In r rr -> rget vs r = RFile.get rf r /\ okFieldVal objs (rget vs r)) /\
      fromEnd - fromSt = toEnd - toSt /\
      fromSt <= fromEnd) **
    (objHp objs objs && heap_eqv (heapOf vs)) ** 
    buffer free toEnd ** buffer fromSt fromEnd **
    RFile.get rf rgcInfo |->, (fromSt::fromEnd::toSt::toEnd::free::nil).


(* specs *)

Definition cheneyLoopHeaderPre s :=
  s |||- 12 + rget s rgcInfo |->_ ** (fun h => cheneyLoopPre (h, rfileOf s)).

Definition cheneyLoopHeaderGuar s s'' :=
  (forall x A,
    s |||- 12 + rget s rgcInfo |-> x ** A ->
      exists s',
	s' |||- 12 + rget s rgcInfo |-> rget s rtoEnd ** A /\
	  rfileEqv s' (rset s rroot x) /\
	  cheneyLoopGuar s' s'') /\
  regPres (regNotInList (rroot::ra::rscan::rtemp::temp0::temp1::t2::t4::t7::a0::rfree::v0::nil)) s s'' /\
  rget s'' ra = rget s raSave.


Definition cheneyEnterPre s :=
  exists fromSt, exists fromEnd, exists toSt, exists toEnd, exists h,
    s |||- cheneyStOk fromSt fromEnd toSt toEnd
      (rget s rfree) (rget s rroot) (rget s rgcInfo) h ** top.

Definition cheneyEnterGuar s s' :=
  (forall toSt toEnd fromSt fromEnd A h,
    s |||- cheneyStOk toSt toEnd fromSt fromEnd
      (rget s rfree) (rget s rroot) (rget s rgcInfo) h ** A ->
    exists phi, exists objs, exists h',
      s' |||- cheneyStOk fromSt fromEnd toSt toEnd
        (rget s' rfree) (rget s' rroot) (rget s' rgcInfo) h' ** A /\
      h ||- minObjHp (rget s rroot::nil) objs ** top /\
      iso_set phi objs (rangeObjs toSt (rget s' rfree)) /\
      mapX phi (rget s rroot) = Some (rget s' rroot) /\
      h' ||- (foralls x in rangeObjs toSt (rget s' rfree), objIsFwdedH phi h x)) /\
  calleeSavedOk s s' /\
  rget s raSave2 = rget s' raSave2.


Definition cheneyDoAllocPre s := 
  s |||- rget s rfree |->__ ** 16 + rget s rgcInfo |->_ ** top.

Definition cheneyDoAllocGuar s s' :=
  (forall A,
    s |||- rget s rfree |->__ ** 16 + rget s rgcInfo |->_ ** A ->
      s' |||- rget s' v0 |->, (1::1::nil) ** 
	16 + rget s rgcInfo |-> 8 + rget s rfree ** A) /\
  rfileEqv s' (rset (rset (rset s v0 (rget s rfree)) rfree (8 + rget s rfree)) ra (rget s raSave2)).


Definition cheneyPostGCPre s :=
  exists toEnd,
    s |||- 12 + rget s rgcInfo |-> toEnd ** top /\
    (rget s rfree <> toEnd -> cheneyDoAllocPre s).

Definition cheneyPostGCGuar s s' := 
  forall toEnd,
    s |||- 12 + rget s rgcInfo |-> toEnd ** top ->
      rget s rfree <> toEnd /\
      cheneyDoAllocGuar (rset s rtemp toEnd) s'.


Definition cheneyAllocPre : state -> Prop := VGC.allocPre repr rroot.

Definition cheneyAllocGuar : state -> state -> Prop :=
  VGC.allocGuar repr rroot CopyMGI.gcGuar CopyMGI.presReg 1.


Definition cheneyReadPre : nat -> state -> Prop := VGC.readPre repr rroot.
Definition cheneyReadGuar : nat -> state -> state -> Prop := 
  VGC.readGuar repr rroot CopyMGI.gcGuar CopyMGI.presReg.
Notation cheneyReadSpec := (fun k => bspeci (cheneyReadPre k) (cheneyReadGuar k)).

Definition cheneyWritePre : nat -> state -> Prop := VGC.writePre repr rroot.
Definition cheneyWriteGuar : nat -> state -> state -> Prop := 
  VGC.writeGuar repr rroot CopyMGI.gcGuar CopyMGI.presReg.
Notation cheneyWriteSpec := (fun k => bspeci (cheneyWritePre k) (cheneyWriteGuar k)).

(* verifying the implementations *)

Definition P : chspec := fun l => 
  match Ploop l with
    | SomeT s => SomeT s
    | NoneT => 
      if beq_nat l CHENEYLOOPHEADER then
	SomeT (bspeci cheneyLoopHeaderPre cheneyLoopHeaderGuar)
      else if beq_nat l CHENEYENTER then
	SomeT (bspeci cheneyEnterPre cheneyEnterGuar)
      else if beq_nat l CHENEYDOALLOC then
	SomeT (bspeci cheneyDoAllocPre cheneyDoAllocGuar)
      else if beq_nat l CHENEYALLOC then
	SomeT (bspeci cheneyAllocPre cheneyAllocGuar)
      else if beq_nat l CHENEYPOSTGC then
	SomeT (bspeci cheneyPostGCPre cheneyPostGCGuar)
      else if beq_nat l INFLOOP then
	SomeT (bspeci (fun s => True) (fun s s' => False))
      else if beq_nat l CHENEYREAD1 then
        SomeT (cheneyReadSpec 0)
      else if beq_nat l CHENEYREAD2 then
	SomeT (cheneyReadSpec 4)
      else if beq_nat l CHENEYWRITE1 then
        SomeT (cheneyWriteSpec 0)
      else if beq_nat l CHENEYWRITE2 then
	SomeT (cheneyWriteSpec 4)
      else
	NoneT _
  end.

Ltac Psimpl := 
  unfold P; unfold Ploop;
  rewrite beq_true; 
  repeat (rewrite beq_false; [idtac | discriminate]); 
  reflexivity.


(* showing the blocks match their specs *)

Lemma cheneyEnterGuarPresRA : forall s s' : state,
   cheneyEnterGuar s s' -> rget s ra = rget s' ra.
Proof.
  intros s s' [_ [CSO _]].
  apply CSO.
  unfold calleeSaved; simpl; auto.
Qed.


Lemma gcGuarRefl : forall vs rr,
  CopyMGI.gcGuar vs vs rr.
Proof.
  intros vs rr objs h Vhp.
  exists h; constructor.
  hypReplace Vhp (sconjDistrib2 Vhp).
  destruct (aconjElim Vhp) as [_ X]; assumption.

  destruct (sconjElim Vhp) as [ha [hb [_ [HA _]]]].
  assert (h ||- minObjHp (List.map (rget vs) rr) objs).
  eapply heapEqvImpl; eauto.
  clear Vhp HA hb ha.
  unfold minObjHp in H; destruct (aconjElim H) as [_ Hp]; clear H.
  unfold okObjHp in Hp; linLogElim Hp.
  eapply isoStateRefl.
  2:instantiate (1 := objs).
  assumption.
  intros r I.
  hypReplace H1 (H1 (rget vs r)).
  apply H1.
  clear H1 Hp h objs.
  induction rr; simpl; simpl in I.
  assumption.
  destruct I as [I | I]; [left | right].
  subst a; reflexivity.
  apply IHrr; assumption.

Qed.


Ltac sllImpl :=
  let h := fresh in
  let x := fresh in
  let Hp := fresh in
    (unfold objRespMap; unfold fieldRespMap;
     unfold objIsFwded; unfold fieldIsFwded;
     cbv beta; intros h x Hp; linLogElim Hp; 
     linLogIntro; sconjMatch).

Lemma infLoopOk : iseqOkS P (fun _ : state => True) (fun _ _ : state => False) infLoop.
Proof.
  unfold infLoop.
  iseqWeakSolver Psimpl.
  auto.
Qed.


Lemma reprSconj : forall h v rf rr,
  h ||- repr v rf rr -> forall A,
  v |||- A ->
  h ||- A ** top.
Proof.
  unfold repr.
  intros h v rf rr Hp A Vhp.
  lle Hp.
  scm.
  apply unitER.
  apply sconjInsert with (heapOf v); auto.
  lli; destruct (aconjElim Hp); assumption.
Qed.


Lemma mvRgetEq : forall h vs s rr,
  h ||- repr vs (rfileOf s) rr -> forall r,
    In r rr ->
    rget s r = rget vs r.
Proof.
  unfold repr; intros h vs s rr Hp r Ir; lle Hp.
  intuition.
  destruct (H2 _ Ir); auto.
Qed.


Lemma readReprUpdate : forall k,
  k = 0 \/ k = 4 -> forall s r x,
    s |||- k + rget s r |-> x ** top -> forall vs rr A,
      s |||- repr vs (rfileOf s) rr ** A ->
	In r rr ->
	s |||- repr (rset vs v0 x) (rfileOf (rset s v0 x)) (v0 :: rr) ** A.
Proof.
  unfold repr.
  intros k K s r x HpX vs rr A Hp Irr.
  lle Hp.

  assert (okX : okFieldVal (rangeObjs x2 x4) x).
  assert (okR : okFieldVal (rangeObjs x2 x4) (rget s r)).
  destruct H3 as [_ [_ [_ [X _]]]]; clear Hp.
  destruct (X _ Irr) as [E1 Ok2].
  rewrite E1 in Ok2; assumption.
  destruct okR as [okR | okR].
  elf; apply (not_even_and_odd (rget s r)).
  apply ptrPeven.
  assert (X := contElim1L HpX).
  destruct K; subst k; [assumption | inversion X; assumption].
  assumption.
  assert (X' : s |||- (okObjAt (rangeObjs x2 x4) (rget s r) ** top) ** top).
  clear HpX H3; scm; destruct (aconjElim Hp) as [Hp' _]; clear Hp.
  hyr Hp' (objHpElim okR Hp'); scm.
  lle X'; clear H3 Hp.
  unfold okObjAt in X'; unfold okFieldAt in X'; lle X'.
  destruct K; subst k.
  rewrite (contContEq HpX X'); assumption.
  sconjLift2 X'; rewrite (contContEq HpX X'); assumption.

  (* now back to the main event *)
  sconjLift9of9 Hp; apply sconjComm; scm.
  existsl x0; existsl x1; existsl x2; existsl x3; existsl x4.
  destruct H3 as [NIrr [Al1 [Al2 [RR [Xeq LE]]]]].
  lli.
  constructor.
  simpl; intros [E | E]; [discriminate E | auto].
  constructor; [auto | constructor; [auto | idtac]].
  constructor.
  intros r' Irr'; simpl in Irr'.
  destruct (regCompare v0 r') as [E | E].
  subst r'; stSimpl; constructor; auto.
  stSimpl; destruct Irr'; [elf; auto | idtac].
  apply (RR _ H).
  intuition.

  assumption.

Qed.


Lemma virtualReadOk : forall k,   
  k = 0 \/ k = 4 -> forall s r x,
  s |||- k + rget s r |-> x ** top -> forall vs rr A,
  s |||- repr vs (rfileOf s) rr ** A ->
  In r rr ->
  vs |||- k + rget vs r |-> x ** top.
Proof.
  intros k K s r x Hpx vs rr A Hp Irr.

  assert (E : rget vs r = rget s r).
  unfold repr in Hp; lle Hp; clear K; intuition.
  destruct (H2 _ Irr) as [E _]; rewrite E; reflexivity.

  rewrite E.
  assert (VHp : vs |||- (k + rget s r |->_ ** top) ** top).
  assert (ptrX := contElim1L Hpx).
  clear Hpx x.
  unfold repr in Hp; lle Hp.
  destruct H3 as [_ [_ [_ [rrOk _]]]].
  assert (Hp' := heapEqvImplL Hp); clear Hp.
  destruct (rrOk _ Irr) as [E' okr]; clear rrOk Irr.
  destruct okr as [okR | okR].
  elf; apply (not_even_and_odd (rget s r)).
  apply ptrPeven; destruct K; subst k.
  assumption.
  inversion ptrX; assumption.
  rewrite E in okR; assumption.

  hyr Hp' (objHpElim okR Hp'); scm.
  rewrite E' in Hp.
  hpunfoldIn Hp; lle Hp; destruct K; subst k; [idtac | sconjLift2of2 Hp];
    lli; scm.
  lle VHp.
  destruct (sconjExtract Hp) as [h [Hp' Hp'']].
  assert (HpQ := reprSconj Hp'' _ VHp).
  lle HpQ; hyr Hp' (sconjInsert Hp' HpQ); lle Hp'.
  rewrite (contContEq Hpx Hp'); assumption.
Qed.


Lemma cheneyReadOk : forall k,
  k = 0 \/ k = 4 ->
  iseqOkS P (cheneyReadPre k) (cheneyReadGuar k) (cheneyRead k).
Proof.
  intros k K; unfold cheneyRead; iseqWeakSolver fail.

  intros s [vs [Hp Vhp]]; lle Vhp.

  destruct (sconjExtract Hp) as [h [Hp'' Hp']]; 
    clear Hp; rename Hp'' into Hp.
  rewrite <- (mvRgetEq _ Hp') in Vhp; [idtac | left; reflexivity].
  assert (Hpx := reprSconj Hp' _ Vhp); lle Hpx;
    hyr Hpx (sconjInsert Hp Hpx); lle Hpx.
  msteps Hpx.

  constructor; auto.
  clear vs Vhp Hp' h Hp.
  intros s' E; subst s'.
  unfold cheneyReadGuar; unfold readGuar; constructor.

  intros vs A rr Hp Irr Irr'.
  exists (rset vs v0 x); stSimpl.
  constructor.

  (* show the guar holds *)
  intros objs h Vhp'.

  assert (X : h ||- minObjHp (List.map (rget vs) rr) objs).
  eapply heapEqvImplL; eauto.

  exists h; stSimpl; constructor.
  scm; destruct (aconjElim Hp0); auto.
  eapply isoStateRefl'.
  instantiate (1 := objs).
  unfold minObjHp in X; destruct (aconjElim X).
  unfold okObjHp in H0; lle H0; assumption.

  intros r I'; destruct (regCompare r v0).
  subst r; elf; assert (Y := Irr' _ I'); calleeSavedSolver Y.
  unfold repr in Hp; lle Hp; destruct H4 as [_ [_ [_ [Req _]]]].
  destruct (Req _ I') as [X' Y].
  stSimpl.
  constructor.
  2:reflexivity.
  unfold minObjHp in X; destruct (aconjElim X) as [_ HOK].
  unfold okObjHp in HOK; lle HOK.
  apply H2.
  apply rrlistMapIntro; assumption.

  constructor.
  eapply readReprUpdate; eauto.
  eapply virtualReadOk; eauto.

  intros r I; destruct I; stSimpl; auto.

Qed.


Lemma cheneyWriteOk : forall k,
  k = 0 \/ k = 4 ->
  iseqOkS P (cheneyWritePre k) (cheneyWriteGuar k) (cheneyWrite k).
Proof.
  intros k K.
  unfold cheneyWrite.
  iseqWeakSolver fail.
  intros s [vs [Hp Vhp]].
  unfold repr in Hp.
  lle Hp; lle Vhp.
  destruct H3 as [_ [_ [_ [Req _]]]].
  
  assert (Req' : rget vs rroot = rget s rroot /\
    okFieldVal (rangeObjs x1 x3) (rget vs rroot)).
  generalize dependent Req; stSimpl; intro.
  apply Req; left; reflexivity.

  clear Req; destruct Req' as [Req Rok].
  rewrite Req in Vhp.
  destruct (aconjElim (sconjDistrib2 Hp)) as [_ Hp'].
  hyr Hp' (sconjInsert Hp' Vhp); lle Hp'.
  rewrite (swStep Hp'); try reflexivity.
  constructor.
  trivial.

  intros s' E; subst s'.
  unfold cheneyWriteGuar; unfold writeGuar.
  constructor.

  rename vs into vs0; rename Hp into Hp0.
  clear Hp' Req Rok.
  unfold repr.
  intros vs A rr Hp Incl.

  lle Hp; destruct H3 as [NI' [al1 [al2 [Req Eok]]]].

  assert (X : rget vs rroot = rget s rroot /\ okFieldVal (rangeObjs x7 x9) (rget s rroot)).
  assert (II : In rroot rr).
  apply Incl; left; reflexivity.
  destruct (Req _ II) as [ReqX Q]; constructor.
  assumption.
  rewrite ReqX in Q; exact Q.

  assert (Y : rget vs a0 = rget s a0 /\ okFieldVal (rangeObjs x7 x9) (rget s a0)).
  assert (II : In a0 rr).
  apply Incl; right; left; reflexivity.
  destruct (Req _ II) as [ReqX Q]; constructor.
  assumption.
  rewrite ReqX in Q; exact Q.

  destruct X as [ReqR OkR].
  destruct Y as [ReqA OkA].
  exists (uhset vs (k + rget s rroot) (rget s a0)); constructor.

  rewrite ReqR; rewrite ReqA; apply gcGuarRefl.

  lli.
  split.
  auto.
  split.
  apply al1.
  split.
  apply al2.
  split.
  intros r Irr; exact (Req _ Irr).
  eauto.

  assert (HpX := heapEqvImplL Hp).
  destruct (aconjElim (sconjDistrib2 Hp)) as [_ Hp']; clear Hp.

  destruct OkR as [OkR | OkR].

  (* case: root is atomic.  proof by contradiction. *)
  elimtype False; eapply not_even_and_odd.
  2:apply OkR.
  apply ptrPeven.
  assert (X := contElim1L Vhp).
  destruct K; subst k; simpl in X; [idtac | inversion X]; assumption.

  (* case: root is not atomic *)
  hyr HpX (objHpElim OkR HpX).
  unfold okObjAt in HpX; unfold okFieldAt in HpX; lle HpX.
  rename x10 into v0; rename x11 into v1; rename H3 into v0Ok; rename H6 into v1Ok.

  assert (HE : Heap.set (heapOf vs) (k + rget s rroot) (rget s a0) = 
    Some (heapOf (uhset vs (k + rget s rroot) (rget s a0)))).
  destruct vs; simpl.
  destruct K; subst k; [idtac | sconjLift2 HpX]; apply (contSetEqL HpX).

  hyr Hp' (sconjHpSet Hp' _ _ HE).
  sconjMatch.
  apply aconjIntro.
  2:assumption.
  eapply heapEqvTrue.
  apply heap_eqvSymm; eapply heap_eqvTrans.
  apply Hp.
  apply heap_eqvSymm; apply HeapMap.map_canonEqv.  
  assert (HpX' : uhset vs (k + rget s rroot) (rget s a0)
    |||- objHp (rangeObjs x7 x9) (set_add (rget s rroot) 
      (set_remove (rget s rroot) (rangeObjs x7 x9)))).
  destruct K; subst k; [idtac | sconjLift2 HpX]; 
    hypReplace HpX (contUpdateSt HpX (rget s a0));
    [idtac | sconjLift2 HpX];
    hypReplace HpX (sconjAssocR HpX);
    apply objHpIntro; sconjMatch; unfold okObjAt; unfold okFieldAt; lli.
  2:sconjMatch.
  4:sconjMatch.
  assumption.
  assumption.
  assumption.
  assumption.
  unfold hpredTrue in HpX'; rewrite (addRemoveEqv OkR) in HpX'.
  trivial.

  intros r I; stSimpl; reflexivity.

Qed.


Lemma cheneyPostGCOk : iseqOkS P cheneyPostGCPre cheneyPostGCGuar cheneyPostGC.
Proof.
  unfold cheneyPostGC.
  iseqWeakSolver Psimpl.
  
  intros s [toEnd [Hp Pre]].
  rewrite (lwStep Hp); try reflexivity; stSimpl.
  assert (BT := doBtestSound btne (rget s rfree) toEnd).
  destruct (doBtest btne (rget s rfree) toEnd); simpl in BT.

  (* case: GC freed up more space *)
  hypReplace Pre (Pre BT).
  constructor.
  clear Hp.
  unfold cheneyDoAllocPre in * |- * ; stSimpl; assumption.
  intros s' Guar toEnd' Hp'.
  assert (X := contContEq Hp Hp'); clear Hp'; subst toEnd'; auto.

  (* case: still no free space: abort *)
  constructor; auto; intros; elf; assumption.

Qed.


Lemma cheneyEnterOk : iseqOkS P cheneyEnterPre cheneyEnterGuar cheneyEnter.
Proof.
  unfold cheneyEnter.
  iseqWeakSolver Psimpl.
  apply scanFieldGuarPresRa.
  
  (* first, step through the block *)
  unfold cheneyEnterPre.
  intros s [toSt [toEnd [fromSt [fromEnd [hx Hp]]]]].
  unfold cheneyStOk in Hp; lle Hp.

  assert (Hp' : s |||- (objHp (rangeObjs fromSt (rget s rfree)) (rangeObjs fromSt (rget s rfree)) && 
    heap_eqv hx) ** buffer (rget s rfree) fromEnd ** buffer toSt toEnd **
    rget s rgcInfo |-> toSt ** 4 + rget s rgcInfo |-> toEnd **
    8 + rget s rgcInfo |-> fromSt ** 12 + rget s rgcInfo |-> fromEnd ** top);
    [scm | clear Hp; rename Hp' into Hp].

  destruct H1 as [alSt [alFr [OkR [sizeOk frLe]]]].
  assert (Hp0 := Hp).
  sconjLift4 Hp0; msteps Hp0.
  sconjLift6 Hp0; msteps Hp0.
  sconjLift2 Hp0; msteps Hp0; hyr Hp0 (contUpdateSt Hp0 fromSt).
  sconjLift2 Hp0; msteps Hp0; hyr Hp0 (contUpdateSt Hp0 toSt).
  sconjLift6 Hp0; msteps Hp0; stSimpl.
  sconjLift7 Hp0; msteps Hp0.
  sconjLift2 Hp0; msteps Hp0; hyr Hp0 (contUpdateSt Hp0 fromEnd).
  sconjLift2 Hp0; msteps Hp0; hyr Hp0 (contUpdateSt Hp0 (rget s rroot)).

  apply callChain.

  assert (HH : exists s', s' =
    (rset (rset (uhset (rset (rset (rset (uhset (rset (rset (uhset
      (uhset (rset (rset s rtoStart toSt) rtemp fromSt) (rget s rgcInfo) fromSt)
      (8 + rget s rgcInfo) toSt) rtoEnd toEnd) rtemp fromEnd) (4 + rget s rgcInfo) fromEnd)
    rfree toSt) rscan toSt) raSave (rget s ra)) (12 + rget s rgcInfo) 
    (rget s rroot)) a0 (12 + rget s rgcInfo)) ra CHENEYLOOPHEADER)).
  econstructor; reflexivity.

  destruct HH as [s1 S1eq]; rewrite <- S1eq.

  assert (alFrSE := aligned8Trans alSt alFr).

  assert (alToSE : aligned8 toSt toEnd).
  destruct alFrSE as [k E].
  exists k.
  omega.

  assert (LEff : rget s rfree <= fromEnd).
  destruct alFr as [y E].
  omega.

  set (objs := rangeObjs fromSt (rget s rfree)).

  assert (Oeq : set_eqv (set_union objs empty_set) objs).
  rewrite (set_eqvUnionSymm objs empty_set).
  rewrite <- (set_eqvUnionEmptyUnit objs).
  apply set_eqvRefl.

  assert (Hp0Ok : 
    scanFieldStOk (fun n:nat => None (A:=nat))
    objs empty_set (rangeObjs toSt toEnd)
    (okFieldAt (set_union objs empty_set) (rget s1 a0) ** 
      rget s rgcInfo |-> fromSt ** 4 + rget s rgcInfo |-> fromEnd **
	8 + rget s rgcInfo |-> toSt ** buffer (rget s rfree) fromEnd ** top)
    s1).

  rewrite S1eq.
  unfold scanFieldStOk; stSimpl.
  constructor.
  subst objs.
  assert (HH : 8 * set_card (rangeObjs fromSt (rget s rfree)) <=
    8 * set_card (rangeObjs toSt toEnd)).
  rewrite (rangeObjsCard alSt).
  rewrite (rangeObjsCard alToSE).
  rewrite sizeOk.
  omega.
  omega.

  constructor.
  rewrite Oeq.

  assert (Hp' : s |||- buffer fromSt (rget s rfree) ** buffer toSt toEnd ** top).
  sconjLift3 Hp; sconjLift2 Hp.
  scm.
  eapply objHpIsBuffer.
  destruct (aconjElim Hp); eauto.
  assumption.

  clear Hp s1 S1eq.

  assert (PT : ptrP toSt \/ toSt = toEnd).
  compare toSt toEnd; intro R.
  right; assumption.
  left.
  assert (s |||- top ** ((!(ptrP toSt) ** top) ** top) ** top).
  scm.
  assert (Al4 : aligned4 toSt toEnd).
  destruct alToSE as [q E]; exists (2 * q); omega.
  destruct Al4 as [k4 Al4].
  unfold buffer in Hp.
  assert (I : set_In toSt (addrRange toSt toEnd)).
  apply addrRangeIntro with 0.
  omega.
  auto.
  hyr Hp (fasElim I Hp).
  scm; llei Hp.
  exact (contElim1 Hp).
  lle H; auto.

  assert (PF : ptrP fromSt \/ fromSt = rget s rfree).
  compare fromSt (rget s rfree); intro R.
  right; assumption.
  left.
  assert (s |||- ((!(ptrP fromSt) ** top) ** top) ** top).
  scm.
  assert (Al4 : aligned4 fromSt (rget s rfree)).
  destruct alSt as [q E]; exists (2 * q); omega.
  destruct Al4 as [k4 Al4].
  unfold buffer in Hp.
  assert (I : set_In fromSt (addrRange fromSt (rget s rfree))).
  apply addrRangeIntro with 0; omega.
  hyr Hp (fasElim I Hp); scm.
  llei Hp; exact (contElim1 Hp).
  lle H; auto.

  destruct PT as [PT | Teq].
  destruct PF as [PF | Feq].

  assert (D : set_disj (addrRange fromSt (rget s rfree))
    (addrRange toSt toEnd)).
  refine (fasFasDisj Hp' _ _);
    intros h x Hp; llei Hp; assert (Q := unitIR Hp); scm.

  subst objs.
  refine (set_disjSub _ (objRangeAddrSetSubset fromSt (rget s rfree))).
  refine (addrAddrSetDisj _ _ _ _ D); auto.

  destruct (ptrP_elim PF) as [k E].
  destruct alSt as [k' E'].
  rewrite E'; rewrite E.
  apply (ptrP_intro (k + 2 * k')).
  omega.

  destruct (ptrP_elim PT) as [k E].
  destruct alToSE as [k' E'].
  rewrite E'; rewrite E.
  apply (ptrP_intro (k + 2 * k')).
  omega.

  subst objs.
  rewrite <- Feq.
  rewrite <- (set_eqvEmptyRangeObjs fromSt).
  apply set_disjSymm; apply set_disjEmptyUnit.

  rewrite Teq.
  rewrite <- (set_eqvEmptySetRange toEnd).
  apply set_disjEmptyUnit.


  constructor.
  simpl; apply set_eqvRefl.
  constructor.
  discriminate 2.
  constructor.
  omega.

  hyr Hp0 (unitIL Hp0).
  permConj Hp0 (1::6::8::2::5::3::4::7::9::nil).
  scm.
  apply mapHpIntroE; assumption.
  unfold hpredTrue; setoid_rewrite Oeq; destruct (aconjElim Hp0); assumption.
  apply bufferIsObjs; assumption.
  unfold okFieldAt; existsl (rget s rroot); lli.
  rewrite Oeq; assumption.
  assumption.

  sconjLift4 Hp; sconjLift2 Hp; sconjLift3 Hp.

  (* scanFieldPre *)
  constructor.
  unfold scanFieldPre; stSimpl.
  exists (fun n:nat => None (A:=nat));
    exists objs; exists empty_set; exists (rangeObjs toSt toEnd).
  rewrite S1eq in Hp0Ok.
  eapply scanFieldStOkImpl.
  rewrite S1eq; apply Hp0Ok.
  stSimpl; intros h H; scm; subst s1; stSimpl; auto.


  (* post condition *)
  intros s' [Guar1 Rp'].

  subst.
  assert (X := Guar1 _ _ _ _ _ Hp0Ok).
  destruct X as [phi' [W' [M' [F' [G' X]]]]].
  generalize dependent X.
  stSimpl.
  unfold fieldRespMap; stSimpl.
  unfold cheneyLoopHeaderPre; unfold cheneyLoopPre.
  unfold cheneyLoopStOk; unfold scanFieldStOk.
  rewrite (regPresEq Rp' rtoStart); [stSimpl | regNotInListSolver].
  rewrite (regPresEq Rp' rtoEnd); [stSimpl | regNotInListSolver].
  intro X.
  destruct X as [objsEq' [GFeq' [iso' [reach' [al8' [IWok SCok]]]]]].

  rewrite (regPresEq Rp'); [stSimpl | regNotInListSolver].
  destruct SCok as [WFlt' [WMdisj' [Feq' [phiOk' [stLt' Hp']]]]].
  sconjLift4 Hp'.

  assert (HH : s' |||- (foralls x in objs, x |->__) **
    rget s rgcInfo |-> fromSt **
      4 + rget s rgcInfo |-> fromEnd ** 8 + rget s rgcInfo |-> toSt ** 
	12 + rget s rgcInfo |->_ ** top).
  assert (HH : s' |||- 12 + rget s rgcInfo |->_ **
    mapHp (set_union empty_set M')
    (map_union (fun _ : nat => None) phi') **
    objHp (set_union W' (set_union empty_set M')) W' **
    top ** top ** rget s rgcInfo |-> fromSt **
      4 + rget s rgcInfo |-> fromEnd ** 8 + rget s rgcInfo |-> toSt ** 
	top).
  scm.
  llei Hp1; eauto.
  permConj HH (2::3::6::7::8::1::9::4::5::nil); hyr HH (sconjAssocR HH); scm.
  unfold hpredTrue; setoid_rewrite objsEq'.
  rewrite <- hpredTrueEq; apply fasMerge.
  scm.
  unfold mapHp in Hp1; linLogElim Hp1.
  unfold hpredTrue; setoid_rewrite (set_eqvUnionEmptyUnit M');
    rewrite <- hpredTrueEq.
  llMatch; llei Hp1; scm.
  apply (objHpIsObjs Hp1).

  (* set_disj M' W' *)
  unfold mapHp in Hp'; unfold objHp in Hp'.
  lle Hp'.
  eapply fasFasDisj.
  instantiate (1 := unit).
  instantiate (3 := h).
  lli; scm; [apply (mapHpIsObjs Hp1) | apply (objHpIsObjs Hp1)].
  cbv beta; intros h' x' Hp''; llei Hp''; scm.
  cbv beta; intros h' x' Hp''; llei Hp''; scm.

  assert (infoNotInObjs :
    forall x, set_In x objs -> forall y, 
      y = 0 \/ y = 4 \/ y = 8 \/ y = 12 ->
      x <> (y + rget s rgcInfo) /\ 4 + x <> (y + rget s rgcInfo)).
  intros x IO y Yeq.
  assert (HH' : s' |||- (x |->_ ** 4 + x |->_ ** top) ** 
    y + rget s rgcInfo |->_ ** top).
  apply (sconjImpl HH); clear HH Hp'; intros h Hp'.
  hyr Hp' (fasElim IO Hp'); simpl in Hp'; hyr Hp' (sconjAssocL Hp'); scm.
  destruct Yeq as [R | [R | [R | R]]]; rewrite R.
  lli; scm.
  sconjLift2 Hp'; lli; scm.
  sconjLift3 Hp'; lli; scm.
  linLogElim Hp'; sconjLift4 Hp'; lli; scm.

  clear Yeq Hp' HH IWok Rp' Hp Hp0.
  assert (QQ : ~(x = y + rget s rgcInfo \/ 4 + x = y + rget s rgcInfo)).
  intros R.
  assert (QQ : s' |||- (y + rget s rgcInfo |->_ ** top) ** y + rget s rgcInfo |->_ ** top).
  apply (sconjImplL HH'); clear HH'; intros h HH'.
  destruct R as [R | R]; rewrite R in HH'; [idtac | sconjLift2 HH']; scm.
  lle QQ.
  intuition.

  assert (Pgc0 : ptrP (rget s rgcInfo)).
  assert (Pgc : s |||- (!ptrP (rget s rgcInfo) ** top) ** top).
  sconjLift3 Hp; scm; lli; exact (contElim1 Hp).  
  lle Pgc; auto.

  assert (Pgc4 : ptrP (4 + rget s rgcInfo)).
  assert (Pgc : s |||- (!ptrP (4 + rget s rgcInfo) ** top) ** top).
  sconjLift5 Hp; scm; lli; exact (contElim1 Hp).  
  lle Pgc; assumption.

  assert (Pgc8 : ptrP (8 + rget s rgcInfo)).
  assert (Pgc : s |||- (!ptrP (8 + rget s rgcInfo) ** top) ** top).
  sconjLift6 Hp; scm; lli; exact (contElim1 Hp).  
  lle Pgc; assumption.

  assert (Pgc12 : ptrP (12 + rget s rgcInfo)).
  assert (Pgc : s |||- (!ptrP (12 + rget s rgcInfo) ** top) ** top).
  sconjLift7 Hp; scm; lli; exact (contElim1 Hp).
  lle Pgc; assumption.

  (* show it is okay to enter the loop *)
  constructor.
  apply (sconjImpl Hp'); clear Hp'; intros h Hp'.
  lle Hp'; exists x; assumption.

  unfold hpredTrue.
  exists (map_union (fun _ : nat => None) phi').
  exists G'; exists W'; exists (set_union empty_set M'); exists F'.
  stSimpl.
  rewrite (regPresEq Rp' rtoStart); [stSimpl | regNotInListSolver].
  rewrite (regPresEq Rp' rtoEnd); [stSimpl | regNotInListSolver].
  rewrite (regPresEq Rp' rscan); [stSimpl | regNotInListSolver].

  constructor.
  apply rangeSplit with toEnd.
  setoid_rewrite <- Feq'.
  sconjLift3 Hp'; sconjLift4 Hp'.
  eapply fasFasDisj.
  apply Hp'.
  unfold objCopiedPhi; unfold objCopied; unfold fieldCopied; stSimpl; simpl.
  intros h' x H; linLogElim H; linLogIntro; sconjMatch.
  simpl; intros; apply (sconjImplR H); intros; linLogIntro.
  omega.
  setoid_rewrite <- Feq'; apply set_eqvSymm; assumption.
  assumption.

  constructor.
  omega.
  repeat (constructor; [assumption | idtac]).

  simpl.
  unfold heap_canon; rewrite (HeapMap.map_canonPEq (HeapMap.map_canonCanonP h));
    rewrite <- hpredTrueEq.
  scm.
  clear Rp' Hp0 IWok.
  unfold objHp; llMatch.
  hpunfoldIn Hp0.
  lle Hp0; hpunfold.

  assert (IO : set_In x0 objs).
  setoid_rewrite objsEq'; apply set_union_intro1.
  destruct iso' as [I1 _].
  destruct (I1 x0) as [_ X].
  apply X; rewrite H1; discriminate.

  hyr infoNotInObjs (infoNotInObjs x0 IO).

  assert (Hp2 : s |||- (okObjAt objs x0 ** top) ** top).
  sconjLift2 Hp; scm; destruct (aconjElim Hp) as [Hp' _]; hyr Hp (objHpElim IO Hp'); scm.

  hpunfoldIn Hp2; lle Hp2.

  clear IO HH; rename H7 into HH.
  unfold POnat.POnat.A in * |- * .

  assert (Eq1 : x3 = x1).
  assert (H' : 
    uset (uset (uset 
      (uset (heapOf s) (rget s rgcInfo) fromSt) (8 + rget s rgcInfo) toSt)
    (4 + rget s rgcInfo) fromEnd) (12 + rget s rgcInfo) (rget s rroot) ||- 
    x0 |-> x3 ** top).
  refine (contUpdateNeqTop _ _ _ Pgc12).
  refine (contUpdateNeqTop _ _ _ Pgc4).
  refine (contUpdateNeqTop _ _ _ Pgc8).
  refine (contUpdateNeqTop _ _ _ Pgc0).
  scm.
  destruct (infoNotInObjs 0); auto.
  destruct (infoNotInObjs 8); [auto | omega].
  destruct (infoNotInObjs 4); [auto | omega].
  destruct (infoNotInObjs 12); [auto | omega].
  exact (contContEq H' H6).

  assert (Eq2 : x4 = x2).
  assert (H' : uset
    (uset (uset (uset (heapOf s) (rget s rgcInfo) fromSt)
      (8 + rget s rgcInfo) toSt) (4 + rget s rgcInfo) fromEnd)
    (12 + rget s rgcInfo) (rget s rroot) ||- 4 + x0 |-> x4 ** top).
  refine (contUpdateNeqTop _ _ _ Pgc12).
  refine (contUpdateNeqTop _ _ _ Pgc4).
  refine (contUpdateNeqTop _ _ _ Pgc8).
  refine (contUpdateNeqTop _ _ _ Pgc0).
  sconjLift2 Hp2; scm.
  destruct (infoNotInObjs 0); auto.
  destruct (infoNotInObjs 8); [auto | omega].
  destruct (infoNotInObjs 4); [auto | omega].
  destruct (infoNotInObjs 12); [auto | omega].
  exact (contContEq H' HH).

  subst.
  lli; eauto.
  setoid_rewrite (set_eqvUnionSymm W' M').
  setoid_rewrite <- objsEq'; assumption.
  setoid_rewrite (set_eqvUnionSymm W' M').
  setoid_rewrite <- objsEq'; assumption.

  (* showing guar for entire function holds *)
  intros s3 Guar2.
  constructor.

  unfold cheneyStOk; intros toSt0 toEnd0 fromSt0 fromEnd0 A h0 Hp00.
  lle Hp00.
  clear H1.
  assert (toSt = toSt0).
  sconjLift3 Hp; sconjLift4 Hp00; exact (contContEq Hp Hp00).
  assert (toEnd = toEnd0).
  sconjLift5 Hp; sconjLift5 Hp00; exact (contContEq Hp Hp00).
  assert (fromSt = fromSt0).
  sconjLift6 Hp; sconjLift6 Hp00; exact (contContEq Hp Hp00).
  assert (fromEnd = fromEnd0).
  sconjLift7 Hp; sconjLift7 Hp00; exact (contContEq Hp Hp00).
  subst toSt0 toEnd0 fromSt0 fromEnd0.
  clear Hp; rename Hp00 into Hp.
  clear Hp0 Hp0Ok M' W' F' G' phi' objsEq' GFeq' iso' IWok WFlt' reach'
    WMdisj' Feq' phiOk' Hp'.

  destruct (Guar1 (fun n:nat => None (A:=nat)) objs empty_set (rangeObjs toSt toEnd)
    (rget s rgcInfo |-> fromSt ** 4 + rget s rgcInfo |-> fromEnd **
	8 + rget s rgcInfo |-> toSt ** buffer (rget s rfree) fromEnd ** A)) as 
    [phi [W [M [F [G X]]]]]; clear Guar1.
  (* a lot of this is just cut and paste from before *)
  unfold scanFieldStOk; stSimpl.
  constructor.
  subst objs.
  assert (HH' : 8 * set_card (rangeObjs fromSt (rget s rfree)) <=
    8 * set_card (rangeObjs toSt toEnd)).
  rewrite (rangeObjsCard alSt).
  rewrite (rangeObjsCard alToSE).
  rewrite sizeOk.
  omega.
  omega.

  constructor.
  setoid_rewrite Oeq.
  sconjLift3 Hp; sconjLift2 Hp.

  assert (Hp' : s |||- buffer fromSt (rget s rfree) ** buffer toSt toEnd ** top).
  scm.
  eapply objHpIsBuffer.
  destruct (aconjElim Hp); eauto.
  assumption.

  (* set_disj objs (setRange toSt toEnd) *)
  clear Hp.
  assert (PT : ptrP toSt \/ toSt = toEnd).
  compare toSt toEnd; intro R.
  right; assumption.
  left.
  sconjLift2 Hp'.
  assert (s |||- ((!(ptrP toSt) ** top) ** top) ** top).
  scm.
  assert (Al4 : aligned4 toSt toEnd).
  destruct alToSE as [q E]; exists (2 * q); omega.
  destruct Al4 as [k4 Al4].
  unfold buffer in Hp.
  assert (I : set_In toSt (addrRange toSt toEnd)).
  eapply addrRangeIntro.
  omega.
  instantiate (1 := 0); auto.
  hyr Hp (fasElim I Hp).
  scm; llei Hp; exact (contElim1 Hp).
  lle H; assumption.

  assert (PF : ptrP fromSt \/ fromSt = rget s rfree).
  compare fromSt (rget s rfree); intro R.
  right; assumption.
  left.
  assert (s |||- ((!(ptrP fromSt) ** top) ** top) ** top).
  scm.
  assert (Al4 : aligned4 fromSt (rget s rfree)).
  destruct alSt as [q E]; exists (2 * q); omega.
  destruct Al4 as [k4 Al4].
  unfold buffer in Hp.
  assert (I : set_In fromSt (addrRange fromSt (rget s rfree))).
  eapply addrRangeIntro.
  omega.
  instantiate (1 := 0); auto.
  hypReplace Hp (fasElim I Hp).
  scm; llei Hp; exact (contElim1 Hp).
  lle H; assumption.

  destruct PT as [PT | Teq].
  destruct PF as [PF | Feq].

  assert (D : set_disj (addrRange fromSt (rget s rfree))
    (addrRange toSt toEnd)).
  refine (fasFasDisj Hp' _ _); illeim; assert (Q := unitIR H); scm.

  subst objs.
  refine (set_disjSub _ (objRangeAddrSetSubset fromSt (rget s rfree))).
  refine (addrAddrSetDisj _ _ _ _ D).

  assumption.
  assumption.

  destruct (ptrP_elim PF) as [k E].
  destruct alSt as [k' E'].
  rewrite E'; rewrite E.
  apply (ptrP_intro (k + 2 * k')).
  omega.

  destruct (ptrP_elim PT) as [k E].
  destruct alToSE as [k' E'].
  rewrite E'; rewrite E.
  apply (ptrP_intro (k + 2 * k')).
  omega.

  subst objs.
  rewrite <- Feq.
  setoid_rewrite <- (set_eqvEmptyRangeObjs fromSt).
  apply set_disjSymm; apply set_disjEmptyUnit.

  rewrite Teq.
  setoid_rewrite <- (set_eqvEmptySetRange toEnd).
  apply set_disjEmptyUnit.
  (* done showing objs and to-space are disjoint *)

  constructor.
  simpl; apply set_eqvRefl.
  constructor.
  discriminate 2.
  constructor.
  omega.

  sconjLift4 Hp; hypReplace Hp (contUpdateSt Hp fromSt).
  sconjLift6 Hp; hypReplace Hp (contUpdateSt Hp toSt).
  sconjLift6 Hp; hypReplace Hp (contUpdateSt Hp fromEnd).
  sconjLift7 Hp; hypReplace Hp (contUpdateSt Hp (rget s rroot)).
  hypReplace Hp (unitIL Hp).

  permConj Hp (1::6::8::2::5::3::4::7::9::nil).
  scm.
  apply mapHpIntroE; assumption.
  unfold hpredTrue; setoid_rewrite Oeq; destruct (aconjElim Hp); assumption.
  apply bufferIsObjs; assumption.
  unfold okFieldAt; existsl (rget s rroot); lli.
  rewrite Oeq; assumption.
  assumption.

  (* done showing what Guar1 does *)
  generalize dependent X; unfold fieldRespMap; stSimpl; intro.
  destruct X as [objsEq [toEq [Iso [reach [_ [WcopyOk X]]]]]].
  destruct X as [leWF [WMdisj [Feq [NSP [_ Hp']]]]].
  destruct Guar2 as [Guar2 [Req2 RaEq2]].

  sconjLift4 Hp'; linLogElim Hp'.
  assert (GIeq : rget s' rgcInfo = rget s rgcInfo).
  rewrite <- Rp'; stSimpl; [reflexivity | regNotInListSolver].
  rewrite GIeq in Guar2.
  sconjLift3 Hp'; destruct (sconjExtract Hp') as [hobjs [Hp'2 hobjsOk]]; clear Hp'.
  rename Hp'2 into Hp'; sconjLift3 Hp'; sconjLift3 Hp'.
  destruct (Guar2 _ _ Hp') as [s'' [Hp'' [Req'' Guar3]]]; clear Guar2; rename Hp' into Hp0'Y.
  destruct Guar3 as [Guar3 [Req3 _]].

  clear Req2; rename Req3 into Req'''; rename Rp' into Req'.
  destruct (Guar3 phi G W M F 
    (rget s rgcInfo |->, (fromSt::fromEnd::toSt::toEnd::nil) **
      buffer (rget s rfree) fromEnd ** A)) as [phi' [W' [M' [F' [G' X]]]]]; clear Guar3.
  unfold cheneyLoopStOk; unfold scanFieldStOk.
  repeat (rewrite Req''); stSimpl.
  rewrite <- (Req' rscan); [stSimpl | regNotInListSolver].
  rewrite <- (Req' rtoStart); [stSimpl | regNotInListSolver].
  rewrite <- (Req' rtoEnd); [stSimpl | regNotInListSolver].

  assert (GFdisj : set_disj G F).
  sconjLift4 Hp''; sconjLift5 Hp''; eapply fasFasDisj; eauto.
  hpunfold; stSimpl; illeim.
  simpl; illeim.

  assert (ToE : rget s' rtoEnd = toEnd).
  rewrite <- (Req' rtoEnd); [stSimpl; reflexivity | regNotInListSolver].

  assert (ToS : rget s' rtoStart = toSt).
  rewrite <- (Req' rtoStart); [stSimpl; reflexivity | regNotInListSolver].

  rewrite ToE in WMdisj; rewrite ToE in Feq; rewrite ToE in NSP;
    rewrite ToE in Hp''.
  rewrite ToS in WMdisj; rewrite ToS in NSP.

  assert (Geqv : set_eqv G (rangeObjs toSt (rget s' rfree))).
  eapply rangeSplit.
  instantiate (1 := toEnd).
  rewrite <- Feq; assumption.
  destruct stLt'; assumption.
  apply set_eqvSymm; rewrite <- Feq; assumption.
  assumption.

  repeat (constructor; [auto | idtac]).
  lli.
  permConj Hp'' (2::3::4::5::6::7::8::1::9::10::nil).
  scm.
  hyr Hp0 (sconjInsert (unitIR Hp0) hobjsOk); lle Hp0; assumption.
  generalize dependent Hp0; hpunfold; stSimpl; intro; 
    unfold objHp; unfold okObjAt; unfold okFieldAt; llMatch; lle Hp0.

  assert (I' : set_In x1 objs).
  rewrite objsEq; apply set_union_intro1.
  destruct Iso as [Iso _]; destruct (Iso x1) as [_ X]; apply X.
  rewrite H1; discriminate.
  sconjLift4 Hp; hypReplace Hp (contUpdateSt Hp fromSt).
  sconjLift6 Hp; hypReplace Hp (contUpdateSt Hp toSt).
  sconjLift6 Hp; hypReplace Hp (contUpdateSt Hp fromEnd).
  sconjLift7 Hp; hypReplace Hp (contUpdateSt Hp (rget s rroot)).
  sconjLift5 Hp; destruct (aconjElim (sconjDistrib2 Hp)) as [HpH _]; clear Hp.

  assert (Q : uset
          (uset
             (uset (uset (heapOf s) (rget s rgcInfo) fromSt)
                (4 + (4 + rget s rgcInfo)) toSt) (4 + rget s rgcInfo) fromEnd)
          (4 + (4 + (4 + rget s rgcInfo))) (rget s rroot) 
	  ||- (okObjAt (rangeObjs fromSt (rget s rfree)) x1 ** top) ** top).
  clear H7 H8; scm.
  hypReplace Hp (objHpElim I' Hp); scm.
  unfold okObjAt in Q; unfold okFieldAt in Q; lle Q.
  assert (X := contContEq H7 Q).
  sconjLift2 Q; assert (Y := contContEq H8 Q).
  subst x3 x4; clear Q HpH.
  unfold okObjAt; unfold okFieldAt; lli; eauto.
  rewrite (set_eqvUnionSymm W M); rewrite <- objsEq; assumption.
  rewrite (set_eqvUnionSymm W M); rewrite <- objsEq; assumption.

  (* have the final state *)
  destruct X as [Weqv [Feqv [Iso' [reach' [Al' X]]]]].
  destruct X as [scFrEq [scLt [al3 SF]]].
  unfold scanFieldStOk in SF.
  assert (E1 : rget s3 rtoStart = toSt).
  rewrite <- Req'''; [stSimpl | regNotInListSolver].
  rewrite Req''; stSimpl.
  rewrite <- Req'; [stSimpl; reflexivity | regNotInListSolver].
  assert (E2 : rget s3 rtoEnd = toEnd).
  rewrite <- Req'''; [stSimpl | regNotInListSolver].
  rewrite Req''; stSimpl.
  rewrite <- Req'; [stSimpl; reflexivity | regNotInListSolver].
  rewrite E1 in SF; rewrite E2 in SF.
  destruct SF as [WFle' [WMdisj' [F'eq [NSP' [Le' Hp30]]]]].

  assert (reach0 : forall x, set_In x (G |_| G') -> 
    rootReachable (heapOf s3) (rget s3 rroot::nil) x).
  rewrite <- Req'''; [stSimpl | regNotInListSolver].
  rewrite Req''; stSimpl.
  assert (reachG : forall z, set_In z G -> reachable (heapOf s3) x z).
  intros z I.
  hypReplace reach (reach _ I).
  assert (X := contGetEqL Hp0'Y).
  rewrite reach in X; injection X; intro; subst z; constructor.
  intros z I0.
  exists x; constructor.
  left; reflexivity.
  destruct (set_union_elim I0) as [I | I]; clear I0.
  apply reachG; assumption.
  destruct (reach' _ I) as [z' [I' R']].
  hypReplace reachG (reachG _ I').
  eapply reachableTrans; eauto.

  (* done proving the reachability we want *)
  clear reach reach'; rename reach0 into reach.
  sconjLift5 Hp30; sconjLift5 Hp30; hypReplace Hp30 (sconjAssocR Hp30).
  destruct (sconjExtract Hp30) as [h3 [Hp3 H3ok]]; clear Hp30.

  assert (WMdisj2 : set_disj W M).
  assert (HpX : s'' |||- (foralls x in W, x |->__) ** (foralls x in M, x |->__) ** top).
  sconjLift2 Hp''; sconjLift3 Hp''; sconjMatch.
  eapply objHpIsObjs; eapply heapEqvTrue.
  apply heap_eqvSymm; apply eqvPredEqv; eauto.
  eauto.
  simpl in Hp0; hypReplace Hp0 (mapHpIsObjs Hp0); assumption.
  eapply fasFasDisj; eauto.
  sllImpl.
  sllImpl.

  assert (MM'disj : set_disj M M').
  apply set_disjSymm; eapply set_unionDisjElimL.
  instantiate (1 := W').
  rewrite <- Weqv.
  assumption.

  assert (Iso'' : iso_set (map_union phi phi') (M |_| M') (G |_| G')).
  apply iso_union.
  assumption.
  assumption.
  assumption.
  eapply fasFasDisj.
  instantiate (1 := unit).
  linLogIntro; eauto.
  sllImpl.
  sllImpl.

  assert (GGdisj : set_disj G G').
  eapply fasFasDisj.
  instantiate (1 := unit).
  instantiate (3 := h3).
  linLogIntro; eauto.
  sllImpl.
  sllImpl.

  assert (MW'disj : set_disj M' W').
  assert (X : s3 |||- ((foralls x in M, x |->__) ** (foralls x in M', x |->__)) ** 
    (foralls x in W', x |->__) ** top).
  sconjLift3 Hp3; sconjLift3 Hp3; sconjMatch.
  apply fasSplit.
  eapply mapHpIsObjs; eauto.
  assumption.
  eapply objHpIsObjs; eauto.
  linLogElim X; sconjLift3 X; sconjLift3 X; eapply fasFasDisj; eauto.
  sllImpl.
  sllImpl.

  set (s1 :=
    rset (rset (uhset
      (rset (rset
	(rset (uhset
	  (rset (rset
	    (uhset (uhset
	      (rset (rset s rtoStart toSt) rtemp fromSt) (rget s rgcInfo) fromSt)
	    (8 + rget s rgcInfo) toSt) rtoEnd toEnd) rtemp fromEnd)
	  (4 + rget s rgcInfo) fromEnd) rfree
	toSt) rscan toSt) raSave 
      (rget s ra))
      (12 + rget s rgcInfo)
      (rget s rroot)) a0 (12 + rget s rgcInfo)) ra CHENEYLOOPHEADER).

  assert (Hp3F : h3 ||- foralls g in (G |_| G'), 
    objIsFwded (fun x:nat => None) (map_union phi phi') s1 g).
  apply fasMerge.
  2:assumption.
  sconjMatch.
  (* show that objects in G have been forwarded *)
  eapply fasImplI.
  apply Hp0.
  cbv beta; intros y hh I HpH.
  eapply copyMapFwded.
  2:unfold map_union; eauto.
  instantiate (1 := top ** top); apply sconjAssocL; clear HpH.
  sconjLift5 Hp''; sconjMatch.
  hypReplace Hp1 (fasElim I Hp1); sconjMatch.
  generalize dependent Hp1; unfold objCopiedPhi; unfold objCopied; unfold fieldCopied.
  intro Q; linLogElim Q; linLogIntro.
  3:apply Q.
  instantiate (1 := x0); unfold map_union; rewrite H1; reflexivity.
  apply H7.
  apply H8.
  apply objIsFwdShift.
  (* show that objects in G' have been forwarded *)
  eapply objCopiedFwded.
  eauto.
  2:eauto.
  rename Hp0'Y into Hp'; sconjLift3 Hp'; sconjLift3 Hp''.
  assert (HX := objHpIsObjs hobjsOk).
  assert (Q := eqvObjCopied Hp' Hp'' _ HX).

  assert (HpX : s'' |||- ((foralls x in M', objCopied x s' x) ** 
    (foralls x in W', objCopied x s' x)) ** top).
  sconjMatch.
  apply fasSplit.
  unfold hpredTrue; rewrite <- Weqv; assumption.
  assumption.
  lle HpX; scm.
  eapply fasImplI.
  eauto.
  cbv beta; intros x0 h4 I HpX.

  eapply objCopiedTrans2.
  2:eauto.
  apply WcopyOk.
  rewrite Weqv; apply set_union_intro1; assumption.
  intros x0 y0 I.
  destruct Iso as [Iso _]; destruct (Iso x0) as [_ Iso1]; clear Iso.
  destruct (phi x0).
  2:reflexivity.
  elimtype False; apply (set_disj_elim MM'disj x0).
  apply Iso1; discriminate.
  destruct Iso' as [Iso' _]; destruct (Iso' x0) as [_ Q]; apply Q.
  rewrite I; discriminate.
  (* done showing forwarding of G |_| G' *)

  clear NSP NSP'.
  unfold map_union in H3.

  assert (PhiR : mapX phi (rget s rroot) = Some x).
  destruct H3 as [v0 [HQ Xeq]].
  unfold mapX in Xeq.
  assert (v0 = rget s rroot).
  sconjLift4 Hp; hypReplace Hp (contUpdateSt Hp fromSt).
  sconjLift6 Hp; hypReplace Hp (contUpdateSt Hp toSt).
  sconjLift6 Hp; hypReplace Hp (contUpdateSt Hp fromEnd).
  sconjLift7 Hp; hypReplace Hp (contUpdateSt Hp (rget s rroot)).
  sconjLift5 Hp; destruct (aconjElim (sconjDistrib2 Hp)) as [HpH _]; clear Hp.
  sconjLift2 HpH; exact (contContEq HQ HpH).
  subst v0; assumption.

  clear H3.

  assert (GF'disj : set_disj G' F').
  hypReplace Hp3 (sconjInsert Hp3 H3ok); linLogElim Hp3.
  sconjLift5 Hp3; sconjLift3 Hp3.
  eapply fasFasDisj; eauto.
  cbv beta; unfold objIsFwded; unfold fieldIsFwded; intros z h Z; lleim.
  cbv beta; illeim.

  assert (RE : rget s' rtoEnd = toEnd).
  rewrite <- Req'; [stSimpl; reflexivity | regNotInListSolver].

  rewrite RE in Feq.

  assert (G'eq : set_eqv G' (rangeObjs (rget s' rfree) (rget s3 rfree))).
  eapply rangeSplit.
  instantiate (1 := toEnd).
  rewrite <- F'eq.
  assumption.
  destruct Le'; assumption.
  apply set_eqvSymm.
  rewrite <- Feq; rewrite <- F'eq; assumption.
  replace (rget s' rfree) with (rget s'' rfree).
  assumption.
  rewrite Req''; stSimpl; reflexivity.

  assert (GFdisj : set_disj G F).
  sconjLift4 Hp''; sconjLift5 Hp''; eapply fasFasDisj; eauto.
  unfold objCopiedPhi; unfold objCopied; unfold fieldCopied; stSimpl;
    intros h x0 HpX; linLogElim HpX; linLogIntro; sconjMatch.
  simpl; intros h x0 HpX; linLogElim HpX; linLogIntro; sconjMatch.

  assert (Geqv : set_eqv G (rangeObjs toSt (rget s' rfree))).
  eapply rangeSplit.
  instantiate (1 := toEnd).
  rewrite <- Feq.
  assumption.
  destruct stLt'; assumption.
  apply set_eqvSymm; rewrite <- Feq; assumption.
  assumption.

  assert (GGeqv : set_eqv (G |_| G') (rangeObjs toSt (rget s3 rfree))).
  rewrite Geqv; rewrite G'eq; apply rangeObjsTransEqv.
  assumption.
  destruct Al' as [k E]; rewrite E; rewrite Req''; stSimpl; omega.


 (* show the final heap is well-formed *)
  exists (map_union phi phi'); exists (M |_| M'); exists h3.
  constructor.
  linLogIntro.

  assert (Al3 : aligned8 toSt (rget s3 rfree)).
  simpl in al8'; eapply aligned8Trans.
  apply al8'.
  eapply aligned8Trans.
  2:apply Al'.
  rewrite Req''; stSimpl; apply aligned8Refl.

  constructor.
  assumption.
  constructor.
  destruct alToSE as [k R]; rewrite R.
  destruct Al3 as [k' R']; rewrite R'.
  exists (k - k'); omega.
  constructor.
  rewrite <- (Req''' rroot); [stSimpl | regNotInListSolver].
  rewrite Req''; stSimpl.
  rewrite <- GGeqv.
  assert (X := isOddOk (rget s rroot)).
  unfold mapX in PhiR; destruct (isOdd (rget s rroot)); [left | right].
  injection PhiR; intro; subst x; assumption.
  apply set_union_intro1; destruct Iso as [_ [Iso _]]; eapply Iso; eauto.
  constructor; [auto | destruct alFrSE; omega].
  
  (* match up the concrete heap *)
  lle Hp3; sconjLift3 Hp3; sconjLift3 Hp3; hyr Hp3 (sconjAssocR Hp3).
  rewrite <- (Req''' rgcInfo); [idtac | regNotInListSolver];
    rewrite (Req'' rgcInfo); stSimpl;
      rewrite <- (Req' rgcInfo); [stSimpl | regNotInListSolver].
  sconjLift8 Hp3; hyr Hp3 (sconjAssocR Hp3).
  permConj Hp3 (2::3::1::4::5::6::7::8::nil).
  scm.

  (* showing the objs heap is okay *)
  apply aconjIntro.
  2:assumption.
  eapply heapEqvTrue.
  apply heap_eqvSymm; apply (eqvPredEqv Hp0).
  unfold hpredTrue; rewrite <- GGeqv.
  rewrite <- hpredTrueEq; eapply objHpIsFwded; eauto.

  apply objsIsBuffer.
  unfold hpredTrue; rewrite <- F'eq; assumption.

  (* aligned8 (rget s3 rfree) toEnd *)
  destruct Al' as [k E]; rewrite E; rewrite Req''; stSimpl.
  destruct al8' as [k' E']; rewrite E'; destruct alToSE as [k'' E'']; rewrite E''.
  exists (k'' - (k + k')).
  cut (k + k' <= k'').
  intro.
  omega.
  destruct Le' as [_ Le']; 
  replace (rget s3 rfree) with (toSt + 8 * k' + 8 * k) in Le'.
  rewrite E'' in Le'.
  omega.
  rewrite <- E'.
  rewrite E; rewrite Req''; stSimpl; reflexivity.

  apply objsIsBuffer.
  2:assumption.
  unfold hpredTrue; rewrite <- (rangeObjsTransEqv alSt LEff).
  apply fasMerge.
  apply sconjComm; scm.
  apply bufferIsObjs; assumption.
  subst objs; unfold hpredTrue; rewrite objsEq; rewrite Weqv.
  rewrite <- (set_eqvUnionAssoc M M' W').
  apply fasMerge; scm.
  eapply mapHpIsObjs; eauto.
  eapply objHpIsObjs; eauto.
  (* set_disj (M |_| M') W' *)
  apply set_unionDisjIntro.
  apply set_disjSymm; eapply set_unionDisjElimR; instantiate (1 := M').
  rewrite <- Weqv; assumption.
  assumption.

  apply rangeSplitDisj.

  rewrite <- (Req''' rroot); [stSimpl | regNotInListSolver].
  rewrite (Req'' rroot); stSimpl.

  constructor.

  (* (h0 ||- minObjHp (rget s rroot :: nil) (M |_| M') ** top *)
  unfold minObjHp.
  assert (HpL : h0 ||- objHp objs (M |_| M') ** top).
  assert (HpL : h0 ||- objHp objs (M |_| M') ** objHp objs W').
  apply objHpSplit.
  unfold hpredTrue; rewrite (set_eqvUnionAssoc M M' W'); subst objs.
  rewrite <- hpredTrueEq; eapply heapEqvImplL; eauto.
  instantiate (2 := heapOf s); sconjMatch.
  destruct (aconjElim Hp) as [Hp1 Hp2]; apply aconjIntro.
  unfold hpredTrue; rewrite <- Weqv; rewrite <- objsEq; assumption.
  assumption.

  apply set_unionDisjIntro.
  apply set_disjSymm; eapply set_unionDisjElimR; instantiate (1 := M').
  rewrite <- Weqv; assumption.  

  assumption.

  scm.

  apply (sconjImplS HpL).
  2:intros; lli.
  intros h S Hp'.
  assert (h ||- objHp (M |_| M') (M |_| M')).

  eapply objHpUnforward; eauto.
  eapply subHeapTrans; eauto.
  destruct (sconjDistrib2 Hp) as [_ Hp1].
  sconjLift4 Hp1; hypReplace Hp1 (contUpdateSt Hp1 fromSt).
  sconjLift6 Hp1; hypReplace Hp1 (contUpdateSt Hp1 toSt).
  sconjLift6 Hp1; hypReplace Hp1 (contUpdateSt Hp1 fromEnd).
  sconjLift7 Hp1; hypReplace Hp1 (contUpdateSt Hp1 (rget s rroot)).
  sconjLift5 Hp1; eapply sconjEqSubHeap; apply Hp1.

  assert (OK : okFieldVal (M |_| M') (rget s rroot)).
  assert (X := isOddOk (rget s rroot)).
  unfold mapX in PhiR; destruct (isOdd (rget s rroot)); [left | right].
  assumption.
  apply set_union_intro1; destruct Iso as [Iso _]; destruct (Iso (rget s rroot)) as [_ Q];
    apply Q; rewrite PhiR; discriminate.

  apply aconjIntro.
  intros z I.

  assert (RR' : rootReachable (heapOf s1) (rget s rroot::nil) z).
  assert (OO := isOddOk (rget s rroot)).
  unfold mapX in PhiR; destruct (isOdd (rget s rroot)).
  
  (* case: root is atomic *)
  elf; destruct Iso'' as [Iso'' ISO]; destruct (Iso'' z) as [X _].
  hypReplace X (X I); clear I Iso''.
  destruct ISO as [ISO _]; hypReplace ISO (ISO z).
  destruct (map_union phi phi' z).
  assert (II : set_In a (G |_| G')).
  eauto.
  hypReplace reach (reach _ II).
  replace (rget s3 rroot) with x in reach.
  injection PhiR; intro U; subst x; clear PhiR X.
  destruct reach as [u [V X]].
  simpl in V; destruct V.
  subst u.
  apply (not_even_and_odd (rget s rroot)).
  2:assumption.
  apply ptrPeven.
  inversion X.
  subst x a.
  hypReplace Hp3F (fasElim II Hp3F).
  cbv beta in Hp3F; unfold objIsFwded in Hp3F; unfold fieldIsFwded in Hp3F; linLogElim Hp3F.
  apply (contElim1L Hp3F).
  unfold Heap.get in H0; assert (Q := isPtrOk (rget s rroot)).
  destruct (isPtr (rget s rroot)); [assumption | discriminate H0].
  unfold Heap.get in H0; assert (Q := isPtrOk (4 + rget s rroot)).
  destruct (isPtr (4 + rget s rroot)); [idtac | discriminate H0].
  inversion Q; assumption.
  assumption.
  rewrite <- Req'''; [stSimpl | regNotInListSolver].
  rewrite Req''; stSimpl; reflexivity.
  elimtype False; apply X; reflexivity.

  (* case: root is a pointer *)
  eapply rootReachForward.
  apply Iso''.
  assert (X := sconjInsert Hp3 Hp3F).
  instantiate (1 := s3); sconjMatch.
  apply reach.
  rewrite <- Req'''; [stSimpl | regNotInListSolver].
  rewrite Req''; stSimpl.
  unfold map_union; rewrite PhiR; reflexivity.
  assumption.

  eapply rootReachableSubheap.
  eapply heapEqvTrue.
  2:apply H.
  apply Heap.heap_eqvSymm; apply Heap.HeapMap.map_canonEqv.
  3:apply RR'.
  2:assumption.
  eapply Heap.subHeapTrans.
  instantiate (1 := h).
  intros q r EE; rewrite Heap.heap_canonEqv; assumption.
  eapply Heap.subHeapTrans.
  eauto.
  destruct (aconjElim (sconjDistrib2 Hp)) as [_ Hp1].
  sconjLift4 Hp1; hypReplace Hp1 (contUpdateSt Hp1 fromSt).
  sconjLift6 Hp1; hypReplace Hp1 (contUpdateSt Hp1 toSt).
  sconjLift6 Hp1; hypReplace Hp1 (contUpdateSt Hp1 fromEnd).
  sconjLift7 Hp1; hypReplace Hp1 (contUpdateSt Hp1 (rget s rroot)).
  sconjLift5 Hp1; eapply sconjEqSubHeap; eauto.

  (* show that M |_| M' is a closed object heap *)
  unfold okObjHp; linLogIntro.
  simpl; intros v I; destruct I as [I | I].
  subst v; assumption.
  elimtype False; assumption.
  assumption.

  constructor.
  rewrite <- GGeqv; assumption.
  constructor.
  unfold mapX in PhiR |- *; destruct (isOdd (rget s rroot)).
  assumption.
  unfold map_union; rewrite PhiR; reflexivity.

  (* show that the objects are forwarded properly *)
  assert (Hp1 := Hp).
  sconjLift4 Hp1; hypReplace Hp1 (contUpdateSt Hp1 fromSt).
  sconjLift6 Hp1; hypReplace Hp1 (contUpdateSt Hp1 toSt).
  sconjLift6 Hp1; hypReplace Hp1 (contUpdateSt Hp1 fromEnd).
  sconjLift7 Hp1; hypReplace Hp1 (contUpdateSt Hp1 (rget s rroot)).
  sconjLift5 Hp1.

  unfold hpredTrue; rewrite <- GGeqv; rewrite <- hpredTrueEq; eapply fasImpl.
  apply Hp3F.
  cbv beta; unfold objIsFwded; unfold fieldIsFwded; unfold objIsFwdedH;
    intros z h HpZ; lle HpZ.
  lli.
  2:simpl; eauto.
  destruct H6 as [v0 [Hpv0 M0]]; destruct H7 as [v1 [Hpv1 M1]].
  exists x0; exists v0; exists v1; constructor.
  assumption.
  constructor.
  simpl.
  assert (Sub : Heap.subHeap h0 (heapOf s1)).
  destruct (aconjElim (sconjDistrib2 Hp1)) as [Hp1a Hp1b].
  eapply sconjEqSubHeap; eauto.
  linLogIntro.
  assert (Hp0 := heapEqvImplL Hp1).
  assert (I : set_In x0 (rangeObjs fromSt (rget s rfree))).
  subst objs; rewrite objsEq; rewrite Weqv.
  rewrite <- (set_eqvUnionAssoc M M' W').
  apply set_union_intro1.
  destruct Iso'' as [Iso'' _]; destruct (Iso'' x0) as [_ MM]; apply MM;
    rewrite H1; discriminate.
  hypReplace Hp0 (fasElim I Hp0);
  cbv beta in Hp0; unfold okObjAt in Hp0; unfold okFieldAt in Hp0; linLogElim Hp0.
  assert (E11 := subHeapContEq Sub Hpv0 Hp0); subst x3.
  sconjLift2 Hp0; assert (E12 := subHeapContEq Sub Hpv1 Hp0); subst x4.
  sconjLift2 Hp0; sconjMatch.

  (* showing the mapping matches up *)
  unfold map_union in M0, M1 |- .
  unfold mapX in M0, M1 |- * .
  constructor.
  destruct (isOdd v0); assumption.
  destruct (isOdd v1); assumption.


  (* show the registers are okay *)
  destruct Guar2 as [_ [Req'' RaEq]].
  rewrite <- Req''; [stSimpl | regNotInListSolver].
  rewrite <- Rp'; [stSimpl | regNotInListSolver].
  constructor.
  2:reflexivity.
  intros r X; unfold calleeSaved in X; simpl in X.
  destruct X as [X | [X | [X | [X | [X | [X | [X | [X | [X | X]]]]]]]]];
    try (subst r; 
      first [rewrite RaEq |
	rewrite <- Req''; [stSimpl | regNotInListSolver]];
      rewrite <- Rp'; [stSimpl | regNotInListSolver]; reflexivity).
  elimtype False; assumption.

Qed.


Lemma cheneyIsoOk' : forall s s' objs objs' phi h h' rr,
  okFieldVal objs (rget s rroot) ->
  iso_set phi objs objs' ->
  mapX phi (rget s rroot) = Some (rget s' rroot) ->

  h ||- minObjHp (List.map (rget s) (rroot :: rr)) objs ->

  h' ||- objHp objs' objs' ->
  h' ||- (foralls x in objs', objIsFwdedH phi h x) ->

  (forall r' : rnum, In r' (rroot :: rr) -> r' = rroot) ->

  (h' ||- Heap.heap_eqv h' ** top) /\
  isoState (rroot :: rr) (h, rfileOf s) (h', rfileOf s').

Proof.
  intros s s' objs objs' phi h h' rr rootOk Iso rootMap.
  intros Hp Hp' HpFwd RRok.

  constructor.
  assert (UU : h' ||- Heap.heap_eqv h').
  apply eqvPredEqv'.
  apply Heap.heap_eqvRefl.
  hypReplace UU (unitIR UU); scm.

  exists objs; exists objs'; exists phi; stSimpl.
  constructor.
  destruct (aconjElim Hp) as [_ Hp2]; unfold okObjHp in Hp2; lle Hp2; auto.
  repeat (constructor; try assumption).
  intros r I; hyr RRok (RRok _ I); subst r; auto.
Qed.
  

Lemma cheneyIsoOk : forall s s' vs objs objs' phi h h'  rr,
  rget vs rroot = rget s rroot ->
  okFieldVal objs (rget vs rroot) ->
  mapX phi (rget s rroot) = Some (rget s' rroot) ->
  vs |||- (minObjHp (List.map (rget vs) (rroot :: rr)) objs && Heap.heap_eqv h) ** top ->
  (forall r' : rnum, In r' (rroot :: rr) -> r' = rroot) ->
  iso_set phi objs objs' ->
  h' ||- (foralls x in objs', objIsFwdedH phi (heapOf vs) x) ->
  okFieldVal objs' (rget s' rroot) ->
  s |||- Heap.heap_eqv (heapOf vs) ** top ->
  s' |||- (objHp objs' objs' && Heap.heap_eqv h') ** top ->
    (h' ||- Heap.heap_eqv h' ** top) /\ isoState (rroot :: rr) (h, rfileOf vs) (h', rfileOf s').
Proof.
  intros s s' vs objs objs' phi h h' rr rootEq rootOk rootMap Hp rphi iso 
    Hp' rootOk' Hp0 Hp0'.
  eapply cheneyIsoOk'.
  2:eauto.
  assumption.
  rewrite rootEq; assumption.
  eapply heapEqvImplL; eauto.
  eapply heapEqvImplL; eauto.
  2:assumption.

  clear rootEq rootOk rootOk' rootMap rphi Hp0 Hp0'.
  apply (fasImplI Hp'); clear Hp'.
  unfold objIsFwdedH; intros x h0 I Hp0; linLogElim Hp0.
  destruct H1 as [x' [v0 [v1 [Peq [Hp' [Meq0 Meq1]]]]]]; lli.
  2:eauto.
  exists x'; exists v0; exists v1; intuition.
  assert (S : Heap.subHeap h (heapOf vs)).
  destruct (aconjElim (sconjDistrib2 Hp)) as [_ Hp1b].
  eapply sconjEqSubHeap; eauto.

  hyr Hp (heapEqvImplL Hp).
  destruct (aconjElim Hp) as [_ Hp1]; clear Hp; rename Hp1 into Hp.
  unfold okObjHp in Hp; lle Hp; clear H1.

  assert (I' : set_In x' objs).
  destruct iso.
  destruct (H x') as [_ Q]; apply Q; rewrite Peq; discriminate.

  hypReplace Hp (objHpElim I' Hp); unfold okObjAt, okFieldAt in Hp; lle Hp.
  llei Hp'.
  
  assert (E1 := subHeapContEq S Hp' Hp).
  sconjLift2 Hp'; sconjLift2 Hp.
  assert (E2 := subHeapContEq S Hp' Hp).
  subst x2 x3.
  sconjLift2 Hp; scm.

Qed.


Lemma cheneyAllocFreeSpaceOk : forall frSt frEnd toSt toEnd free s vs,
  aligned8 frSt free ->
  aligned8 free frEnd ->
  toEnd - toSt = frEnd - frSt ->
  toSt <= toEnd ->
  rget vs rroot = rget s rroot ->
  okFieldVal (rangeObjs frSt free) (rget vs rroot) ->
  s |||- (objHp (rangeObjs frSt free) (rangeObjs frSt free) && heap_eqv (heapOf vs)) **
    buffer free frEnd ** buffer toSt toEnd **
    rget s rgcInfo |-> toSt ** 4 + rget s rgcInfo |-> toEnd **
      8 + rget s rgcInfo |-> frSt ** 12 + rget s rgcInfo |-> frEnd **
        16 + rget s rgcInfo |-> free ** top ->
  free <> frEnd ->
  cheneyDoAllocPre
  (rset (rset (rset s raSave2 (rget s ra)) rfree free) rtemp frEnd) /\
  (forall s' : state,
    cheneyDoAllocGuar
    (rset (rset (rset s raSave2 (rget s ra)) rfree free) rtemp frEnd) s' ->
    cheneyAllocGuar s s').
Proof.
  unfold cheneyDoAllocPre, cheneyDoAllocGuar.
  intros until vs; intros al1 al2 sizeOk Le rootEq rootOk Hp BT; stSimpl.

  assert (I : set_In free (rangeObjs free frEnd)).
  destruct al2 as [k E]; subst frEnd; destruct k.
  elf; auto.
  apply rangeObjsIntro with 0; omega.

  constructor.
  
  sconjLift8 Hp; sconjLift3 Hp.
  assert (Hp' : s |||- (free |->__ ** top) ** 16 + rget s rgcInfo |->_ ** top).
  scm.
  hypReplace Hp (fasElim I (bufferIsObjs _ al2 Hp)); scm.
  lli; eauto.
  llei Hp'; eauto.

  intros s' [G Req]; constructor.

  rename Hp into Hp0; rename vs into vs0; clear rootEq rootOk.
  unfold repr; intros vs A Hp; lle Hp.
  assert (x = toSt).
  sconjLift4 Hp0; sconjLift4 Hp; exact (contContEq Hp Hp0).
  assert (x0 = toEnd).
  sconjLift5 Hp0; sconjLift5 Hp; exact (contContEq Hp Hp0).
  assert (x1 = frSt).
  sconjLift6 Hp0; sconjLift6 Hp; exact (contContEq Hp Hp0).
  assert (x2 = frEnd).
  sconjLift7 Hp0; sconjLift7 Hp; exact (contContEq Hp Hp0).
  assert (x3 = free).
  sconjLift8 Hp0; sconjLift8 Hp; exact (contContEq Hp Hp0).
  subst x x0 x1 x2 x3.
  destruct H3 as [_ [_ [_ [rootOk0 _]]]].

  assert (rootOk : rget vs rroot = RFile.get (rfileOf s) rroot /\
    okFieldVal (rangeObjs frSt free) (rget vs rroot)).
  apply rootOk0; simpl; auto.

  clear rootOk0 Hp0 vs0.
  destruct rootOk as [rootEq rootOk].
  exists vs; exists free.
  constructor.
  apply gcGuarRefl.
  
  unfold repr.
  assert (Hp' : s |||- (free |->__ ** buffer (8 + free) frEnd) ** 
    (objHp (rangeObjs frSt free) (rangeObjs frSt free) && Heap.heap_eqv (heapOf vs)) **
    buffer toSt toEnd **
    RFile.get (rfileOf s) rgcInfo |->, (toSt::toEnd::frSt::frEnd::nil) ** 
    16 + rget s rgcInfo |->_ ** A).
  aspConj Hp ([|2, [|1, [|3, [|[|4, [|5, [|6, 7|]|]|], [|8, 9|] |]|]|]|]); scm.
  hyr Hp (fasElim I (bufferIsObjs _ al2 Hp)); scm.
  apply objsIsBuffer.
  unfold hpredTrue; rewrite <- (rangeRemoveEqv free frEnd); assumption.
  destruct al2 as [k E]; subst frEnd; destruct k.
  elimtype False; auto.
  exists k; omega.
  lli; eauto.

  clear Hp; rename Hp' into Hp.
  constructor.
  hyr Hp (sconjAssocL Hp); sconjLift6 Hp; sconjLift2 Hp.
  hypReplace Hp (G _ Hp); clear G.
  replace (RFile.get (rfileOf s) rgcInfo) with (rget s rgcInfo) in Hp by auto.
  replace (rget s rgcInfo) with (rget s' rgcInfo) in Hp by
    (rewrite (Req rgcInfo); stSimpl; reflexivity).
  apply sconjComm; sconjLift7of7 Hp; scm.
  existsl toSt; existsl toEnd; existsl frSt; existsl frEnd; existsl (8 + free).
  lli.

  constructor.
  simpl; intros [D | [D | D]]; auto; discriminate D.
  constructor.
  destruct al1 as [k E]; exists (S k); omega.
  constructor.
  destruct al2 as [k E]; destruct k.
  elf; omega.
  exists k; omega.
  constructor.
  2:intuition.
  intros r Ir.
  replace (RFile.get (rfileOf s') r) with (rget s' r) by reflexivity.
  simpl in Ir; destruct Ir as [D | [D | D]].
  subst r.
  rewrite Req; stSimpl.
  constructor.
  reflexivity.
  right.
  
  destruct (rangeObjsElim I) as [X _].
  destruct al1; apply rangeObjsIntro with x; auto.

  subst r; stSimpl.
  rewrite Req; stSimpl.
  constructor.
  assumption.
  destruct rootOk; [left | right].
  assumption.
  2:elf; auto.

  destruct (rangeObjsElim H) as [X [k Y]].
  apply rangeObjsIntro with k; omega.

  (* showing the heap is okay *)
  stSimpl.
  replace (rget s' v0) with free in Hp by (rewrite Req; stSimpl; auto).

  assert (SE : set_eqv (rangeObjs frSt (8 + free)) (set_add free (rangeObjs frSt free))).
  apply set_eqvSymm; apply rangeObjsEndAdd; assumption.
  permConj Hp (4::1::3::5::6::2::nil); hyr Hp (sconjAssocR Hp); scm.

  destruct (aconjElim (sconjDistrib2 Hp)) as [Hp1 Hp2]; clear Hp Req.
  apply aconjIntro; [clear Hp2 | clear Hp1].
  unfold hpredTrue; rewrite SE.
  apply objHpIntro.
  apply sconjComm; scm.
  eapply objHpWeaken; eauto.
  intros x IX; apply set_add_intro2; assumption.
  simpl in Hp; unfold okObjAt; unfold okFieldAt; lli; eauto;
    left; constructor; constructor.

  rename Hp2 into Hp; simpl in Hp; sconjLift2 Hp.
  simpl; apply unitER.
  apply heapEqvAdd.
  apply sconjGLift2'.
  apply heapEqvAdd.
  linLogIntro; assumption.
  permConj Hp (1::3::4::5::2::nil).
  auto.

  (* showing that the object is fresh *)
  intros B Vhp; stSimpl.

  clear G Req.
  simpl; replace (S (S (S (S free)))) with (4 + free) by reflexivity.
  linLogElim Hp.

  assert (X : s |||- (!(ptrP free) ** top) ** top).
  scm; lli; eapply contElim1; eauto.

  lle X; clear X; rename H1 into freeP.
  permConjG (3::2::1::nil).
  sconjLift4 Hp; hypReplace Hp (sconjDistrib2 Hp); destruct (aconjElim Hp) as [_ Hp'];
    clear Hp; rename Hp' into Hp.
  apply heapAddOk.
  apply heapAddOk; auto.
  intro Vhp'; lle Vhp'.
  hyr Hp (sconjInsert Hp Vhp'); lle Hp.
  destruct (ptrP_elim freeP) as [k E]; apply (ptrP_intro (S k)); omega.
  intro Vhp'; lle Vhp'.
  sconjLift2 Hp; hyr Hp (contUpdateSt Hp 1).
  hypReplace Hp (heapEqvAdd Hp).
  hypReplace Hp (sconjInsert Hp Vhp'); lle Hp.

  clear Hp G.
  unfold CopyMGI.presReg; unfold calleeSaved.
  intros r Z.
  simpl in Z.
  rewrite Req.
  intuition; subst; stSimpl; reflexivity.

Qed.


Lemma cheneyAllocCallGCOk : forall frSt frEnd toSt toEnd free s vs,
  aligned8 frSt free ->
  aligned8 free frEnd ->
  toEnd - toSt = frEnd - frSt ->
  toSt <= toEnd ->
  rget vs rroot = rget s rroot ->
  okFieldVal (rangeObjs frSt free) (rget vs rroot) ->
  s |||- (objHp (rangeObjs frSt free) (rangeObjs frSt free) && heap_eqv (heapOf vs)) **
    buffer free frEnd ** buffer toSt toEnd ** rget s rgcInfo |-> toSt **
      4 + rget s rgcInfo |-> toEnd ** 8 + rget s rgcInfo |-> frSt **
        12 + rget s rgcInfo |-> frEnd ** 16 + rget s rgcInfo |-> free ** top ->
  free = frEnd ->
  let s' := rset (rset (rset (rset s raSave2 (rget s ra)) rfree free) rtemp frEnd)
    ra CHENEYPOSTGC in
    (cheneyEnterPre s' /\
      (forall s'', cheneyEnterGuar s' s'' -> cheneyPostGCPre s'')) /\
    (forall s''', (exists s'', cheneyEnterGuar s' s'' /\ cheneyPostGCGuar s'' s''') ->
      cheneyAllocGuar s s''').
Proof.
  intros until vs; intros al1 al2 sizeOk Le rootEq rootOk Hp BT.
  subst free.
  apply callChain.
  constructor.

  (* it is safe to call the collector *)
  unfold cheneyEnterPre; stSimpl.
  exists toSt; exists toEnd; exists frSt; exists frEnd; exists (heapOf vs).
  unfold cheneyStOk; lli.
  rewrite <- rootEq; auto.
  scm.

  (* simplify things a bit *)
  unfold cheneyEnterGuar; stSimpl.
  intros s' [Guar [Req RaEq]].
  constructor.

  (* show it is safe to call postGC afterwards *)
  destruct (Guar toSt toEnd frSt frEnd (16 + rget s rgcInfo |-> frEnd ** top) 
    (heapOf vs)) as [phi [objs [h' X]]]; clear Guar.
  unfold cheneyStOk; lli; [rewrite <- rootEq; auto | scm].
  clear Hp; destruct X as [Hp _].
  exists toEnd.
  constructor.
  unfold cheneyStOk in Hp; lle Hp; sconjLift7 Hp; scm.
  intros Ne; unfold cheneyDoAllocPre.
  rewrite <- (Req rgcInfo); [stSimpl | idtac].
  2:unfold calleeSaved; unfold rgcInfo; right; right; right; right; right; right;
    right; right; left; reflexivity.
  unfold cheneyStOk in Hp; lle Hp.
  assert (HH : s' |||- (rget s' rfree |->__ ** top) ** 
    16 + rget s rgcInfo |->_ ** top).
  sconjLift8 Hp; sconjLift3 Hp; scm.
  2:lli; eauto.
  destruct H1 as [_ [X _]].
  hypReplace Hp (bufferObjElim _ Hp X Ne); scm.
  llei HH; eauto.


  (* show the guarantee holds together *)
  intros s'' Guar2; constructor.

  rename vs into vs0; clear rootOk rootEq.
  destruct (aconjElim (sconjDistrib2 Hp)) as [Hp0 _]; clear Hp vs0.
  unfold repr; intros vs A Hp; lle Hp.
  replace (RFile.get (rfileOf s) rgcInfo) with (rget s rgcInfo) in Hp by auto.
  destruct H3 as [_ [_ [_ [rootOk _]]]].

  assert (x = toSt).
  sconjLift4 Hp0; sconjLift4 Hp; exact (contContEq Hp Hp0).
  assert (x0 = toEnd).
  sconjLift5 Hp0; sconjLift5 Hp; exact (contContEq Hp Hp0).
  assert (x1 = frSt).
  sconjLift6 Hp0; sconjLift6 Hp; exact (contContEq Hp Hp0).
  assert (x2 = frEnd).
  sconjLift7 Hp0; sconjLift7 Hp; exact (contContEq Hp Hp0).
  assert (x3 = frEnd).
  sconjLift8 Hp0; sconjLift8 Hp; exact (contContEq Hp Hp0).
  subst x x0 x1 x2 x3; clear Hp0.

  hyr rootOk (rootOk rroot).
  assert (Z : rget vs rroot = rget s rroot /\
    okFieldVal (rangeObjs frSt frEnd) (rget vs rroot)).
  clear Guar Hp.
  replace (RFile.get (rfileOf s) rroot) with (rget s rroot) in rootOk by auto.
  apply rootOk; simpl; auto.

  clear rootOk; destruct Z as [rootEq rootOk].
  destruct (Guar toSt toEnd frSt frEnd (16 + rget s rgcInfo |-> frEnd ** A) 
    (heapOf vs)) as [phi [objs [h' X]]]; clear Guar.
  unfold cheneyStOk; lli; [rewrite <- rootEq; auto | scm].

  destruct X as [Hp' [MHp [Iso [rootIso Vhp]]]].
  unfold cheneyPostGCGuar in Guar2.
  unfold cheneyStOk in Hp'; lle Hp'.
  destruct H1 as [Al1' [Al2' [rootOk' _]]].

  assert (Hp'2 : s' |||- 12 + rget s' rgcInfo |-> toEnd ** top).
  sconjLift7 Hp'; scm.
  destruct (Guar2 _ Hp'2) as [NE [Guar' Req']]; clear Guar2 Hp'2.
  assert (GIeq : rget s' rgcInfo = rget s rgcInfo).
  rewrite <- Req.
  stSimpl; reflexivity.
  unfold calleeSaved; unfold rgcInfo;
    right; right; right; right; right; right; right; right; left; reflexivity.
  generalize dependent Guar'; stSimpl; intro.
  rewrite GIeq in Guar'; rewrite GIeq in Hp'.
  sconjLift8 Hp'; sconjLift3 Hp'.
  assert (Hp'2 : s' |||- (rget s' rfree |->__ ** buffer (rget s' rfree + 8) toEnd) **
    16 + rget s rgcInfo |->_ ** (objHp (rangeObjs toSt (rget s' rfree))
      (rangeObjs toSt (rget s' rfree)) && Heap.heap_eqv h') **
    buffer frSt frEnd ** rget s rgcInfo |-> frSt ** 4 + rget s rgcInfo |-> frEnd **
      8 + rget s rgcInfo |-> toSt ** 12 + rget s rgcInfo |-> toEnd ** A).
  scm; clear Guar' Hp Req Req'.
  hypReplace Hp0 (bufferObjElim _ Hp0 Al2' NE); auto.
  lli; eauto.

  hypReplace Hp'2 (sconjAssocL Hp'2); sconjLift3 Hp'2; sconjLift2 Hp'2.
  assert (Hp'' := Guar' _ Hp'2); clear Hp' Hp'2 Guar'.

  sconjLift4 Hp''.
  generalize dependent Hp''; rewrite (Req' v0); stSimpl; intro.
  replace (rget s' rfree |->, (1 :: 1 :: nil)) with
    (rget s' rfree |-> 1 ** 4 + rget s' rfree |-> 1) in Hp'' by auto.
  exists (h', rfileOf s''); exists (rget s' rfree).
  stSimpl.

  assert (Hp''2 : s'' |||- (objHp (rangeObjs toSt (rget s' rfree))
                 (rangeObjs toSt (rget s' rfree)) && 
               Heap.heap_eqv h') **
              rget s' rfree |-> 1 ** 4 + rget s' rfree |-> 1 **
              buffer (rget s' rfree + 8) toEnd **
              buffer frSt frEnd **
              rget s rgcInfo |->, (frSt::frEnd::toSt::toEnd::8+rget s' rfree::nil) ** 
	      A).
  clear Hp Req sizeOk Req Req' RaEq Vhp; lle Hp''.
  aspConj Hp'' ([|1,[|2,[|3,[|5,[|6,[|[|7,[|8,[|9,[|10,5|]|]|]|],11|]|]|]|]|]|]).
  exact Hp''.

  clear Hp''; rename Hp''2 into Hp''.
  constructor.

  (* show that the collector respects the GC guar *)
  unfold CopyMGI.gcGuar, CopyGCGuar.gcGuar.
  intros objs0 h Vhp'.

  (* first a little cleanup, including dumping s' *)
  stSimpl.
  generalize dependent (rget s' rfree); intro free'; intros.

  assert (RootEq : rget s' rroot = rget s'' rroot).
  rewrite Req'; stSimpl; reflexivity.

  rewrite RootEq in rootIso; rewrite RootEq in rootOk'.
  clear s' Req GIeq Req' RootEq RaEq.
  rename s'' into s'.

  assert (OE : set_eqv objs objs0).
  eapply minObjHpEqv; eauto.
  destruct (aconjElim (sconjDistrib2 Vhp')); eauto.
  constructor; intro I.
  destruct I.
  left; rewrite rootEq; assumption.
  elf; assumption.
  rewrite <- rootEq.
  replace (rget vs rroot :: nil) with (set_add (rget vs rroot) empty_set).
  assumption.
  reflexivity.

  exists h'.
  eapply cheneyIsoOk.
  6:apply Iso.
  eauto.
  assert (vs |||- (!(okFieldVal objs (rget vs rroot)) ** top) ** top).
  clear Vhp'; scm.
  lli.
  eapply minObjHpRootOk; eauto.
  rewrite rootEq; left; reflexivity.
  lle H; assumption.
  assumption.
  scm.
  destruct (aconjElim Hp0) as [HpA HpB]; apply aconjIntro.
  unfold hpredTrue; rewrite OE; assumption.
  assumption.
  simpl; intros r' [E | E']. auto. elf; auto.
  assumption.
  assumption.
  scm; destruct (aconjElim Hp); assumption.
  scm.

  (* back to the heap *)
  constructor.

  (* show that the final heap is well-formed *)
  sconjLift7of7 Hp''; apply sconjComm; scm.
  existsl frSt; existsl frEnd; existsl toSt; existsl toEnd; existsl (8 + rget s' rfree); lli.

  (* side conditions for the heap *)
  constructor.
  simpl; intros [X | [X | X]]; auto; discriminate X.
  constructor.
  destruct Al1' as [k Al1']; rewrite Al1'; exists (S k); omega.
  constructor.
  destruct Al2' as [k Al2']; rewrite Al2'.
  destruct k; [elf | exists k]; omega.
  constructor.

  intros r I; simpl in I; destruct I as [I | [I | I]].
  3:elf; assumption.
  (* case: r = v0 *)
  subst r; stSimpl.
  replace (RFile.get (rfileOf s'') v0) with (rget s'' v0); [rewrite Req'; stSimpl | reflexivity].
  constructor.
  reflexivity.
  destruct Al1' as [k Al1'].
  right; eapply rangeObjsIntro.
  auto.
  rewrite Al1'; reflexivity.
  (* case: r = rroot *)
  subst r; stSimpl; constructor.
  reflexivity.
  replace (rget (h', rfileOf s'') rroot) with (rget s'' rroot) by reflexivity.
  rewrite Req' ; stSimpl.
  destruct rootOk' as [H | H]; [left | right].
  assumption.
  destruct (rangeObjsElim H) as [B [k C]].
  rewrite C; eapply rangeObjsIntro.
  2:reflexivity.
  omega.
  constructor.
  auto.
  destruct al1; omega.
  
  (* heap itself *)
  clear Hp; rename Hp0 into Hp.
  hyr Hp (sconjAssocR (sconjAssocR Hp)); scm.
  lle Hp.
  destruct (aconjElim (sconjDistrib2 Hp)) as [Hp1 Hp2]; apply aconjIntro; clear Hp;
    [clear Hp2 | clear Hp1].
  unfold hpredTrue; rewrite <- (rangeObjsEndAdd Al1').
  apply objHpIntro.
  apply sconjComm; scm.
  apply (objHpWeaken Hp).
  intros x I; apply set_add_intro2; assumption.
  unfold okObjAt; unfold okFieldAt; linLogIntro.
  2:eauto.
  left; constructor; constructor.
  left; constructor; constructor.
  sconjLift2 Hp2; hypReplace Hp2 (heapEqvAdd Hp2).
  sconjLift2of2 Hp2; apply unitER.
  replace (heapOf (uhset (uhset (h', rfileOf s'') (rget s' rfree) 1) (4 + rget s' rfree) 1)) with
    (Heap.uset (Heap.uset h' (rget s' rfree) 1) (4 + rget s' rfree) 1) by reflexivity.
  apply heapEqvAdd.
  lli; assumption.
  replace (8 + rget s' rfree) with (rget s' rfree + 8); [assumption | omega].
  replace (RFile.get (rfileOf s'') rgcInfo) with (rget s'' rgcInfo) by reflexivity.
  rewrite Req'; stSimpl; rewrite <- Req; stSimpl.
  assumption.
  right; right; right; right; right; right; right; right; left; reflexivity.

  assert (X : s'' |||- (!(ptrP (rget s' rfree)) ** top) ** top).
  sconjLift2 Hp''; scm; lli.
  eapply contElim1; eauto.
  lle X; clear X; rename H1 into frP.

  assert (Hp' : s'' |||- Heap.heap_eqv h' ** rget s' rfree |-> 1 **
              4 + rget s' rfree |-> 1 ** top).
  scm; destruct (aconjElim Hp0); assumption.

  clear Hp Hp''; intros B Hp.
  apply sconjComm; lli; apply sconjComm; lli.
  replace (heapOf (uhset (uhset (h', rfileOf s'') (rget s' rfree) 1) (4 + rget s' rfree) 1)) with
    (Heap.uset (Heap.uset h' (rget s' rfree) 1) (4 + rget s' rfree) 1) by reflexivity.

  apply heapAddOk.
  apply sconjComm; apply heapAddOk; try assumption.
  intro Vhp'; lle Vhp'; hyr Hp (sconjInsert Hp' Vhp'); lle Hp.
  destruct (ptrP_elim frP) as [k E]; apply (ptrP_intro (S k)); omega.
  intro Vhp'; lle Vhp'; sconjLift2 Hp'; hyr Hp' (contUpdateSt Hp' 1).
  hyr Hp' (heapEqvAdd Hp'); hyr Hp' (sconjInsert Hp' Vhp'); lle Hp'.

  (* show the registers are preserved *)
  unfold regPres; unfold CopyMGI.presReg.
  intros r C.
  destruct (Guar toSt toEnd frSt frEnd top (heapOf vs)) as [phi [h' X]].
  unfold cheneyStOk.
  lli.
  rewrite <- rootEq; auto.
  scm.

  clear Hp Guar.
  destruct X as [h'' [X _]].
  assert (Hp' : s' |||- 12 + rget s' rgcInfo |-> toEnd ** top).
  unfold cheneyStOk in X; lle X.
  sconjLift7 X; scm.

  clear X; destruct (Guar2 _ Hp') as [NE X]; clear Guar2 Hp'.
  destruct X as [_ X]; rewrite X; clear X sizeOk Le rootEq al1 al2 rootOk.
  assert (C' := C).
  unfold calleeSaved in C'; simpl in C'.
  destruct C' as [X | [X | [X | [X | [X | [X | [X | [X | [X | X]]]]]]]]]; 
    try (elf; assumption); subst r; stSimpl; try (rewrite <- Req; stSimpl; auto; fail).

  simpl in RaEq; rewrite RaEq; reflexivity.

Qed.


Lemma cheneyAllocOk : iseqOkS P cheneyAllocPre cheneyAllocGuar cheneyAlloc.
Proof.
  unfold cheneyAlloc.
  iseqWeakSolver Psimpl.
  
  intros; apply cheneyEnterGuarPresRA; assumption.

  unfold cheneyAllocPre; intros s [vs Hp]; unfold repr in Hp; lle Hp.
  rename x into toSt; rename x0 into toEnd; 
    rename x1 into frSt; rename x2 into frEnd; rename x3 into free.
  destruct H3 as [_ [al1 [al2 [rootOk0 [sizeOk Le]]]]].
  assert (rootOk : rget vs rroot = RFile.get (rfileOf s) rroot /\ 
    okFieldVal (rangeObjs frSt free) (rget vs rroot)).
  apply rootOk0; simpl; auto.
  clear rootOk0; destruct rootOk as [rootEq rootOk].
  assert (HpX : s |||- (objHp (rangeObjs frSt free) (rangeObjs frSt free) && 
    heap_eqv (heapOf vs)) ** buffer free frEnd ** buffer toSt toEnd **
    rget s rgcInfo |-> toSt ** 4 + rget s rgcInfo |-> toEnd **
      8 + rget s rgcInfo |-> frSt ** 12 + rget s rgcInfo |-> frEnd **
	16 + rget s rgcInfo |-> free ** top).
  scm.
  clear Hp; rename HpX into Hp.

  assert (Hp' := Hp).
  rewrite addiuStep0; stSimpl.
  sconjLift8 Hp'; rewrite (lwStep Hp'); try reflexivity.
  sconjLift8 Hp'; rewrite (lwStep Hp'); try reflexivity; stSimpl.
  assert (BT := doBtestSound btne free frEnd).
  clear Hp'.
  destruct (doBtest btne free frEnd); simpl in BT.
  eapply cheneyAllocFreeSpaceOk; eauto.
  eapply cheneyAllocCallGCOk; eauto.

Qed.


Lemma cheneyDoAllocOk : iseqOkS P cheneyDoAllocPre cheneyDoAllocGuar cheneyDoAlloc.
Proof.
  unfold cheneyDoAlloc.
  iseqWeakSolver Psimpl.
  unfold cheneyDoAllocPre.
  intros s Hp; lle Hp.
  rewrite addiuStepZero; stSimpl.
  rewrite (swStep Hp); try reflexivity.
  hyr Hp (contUpdateSt Hp 1).
  sconjLift2 Hp; rewrite (swStep Hp); try reflexivity.
  hyr Hp (contUpdateSt Hp 1).
  rewrite addiuStep0; rewrite addiuStep; stSimpl.
  sconjLift3 Hp; rewrite (swStep Hp); try reflexivity; stSimpl.
  rewrite addiuStep0; stSimpl.
  constructor.
  trivial.

  intros s' X; subst s'; unfold cheneyDoAllocGuar; stSimpl; constructor.

  rename Hp into Hp0; intros A Hp; llei Hp.
  hyr Hp (contUpdateSt Hp 1).
  sconjLift2 Hp; hypReplace Hp (contUpdateSt Hp 1).
  sconjLift3 Hp; hypReplace Hp (contUpdateSt Hp (8 + rget s rfree)).
  sconjLift2 Hp; sconjLift3 Hp; assumption.

  intro r.
  destruct (regCompare r v0).
  subst r.
  stSimpl; reflexivity.
  destruct (regCompare r rfree).
  subst; stSimpl; reflexivity.
  destruct (regCompare r ra); subst; stSimpl; reflexivity.

Qed.


Lemma cheneyLoopHeaderOk : 
  iseqOkS P cheneyLoopHeaderPre cheneyLoopHeaderGuar cheneyLoopHeader.
Proof.
  unfold cheneyLoopHeader.
  iseqWeakSolver Psimpl.
  unfold cheneyLoopHeaderPre.
  intros s Hp.

  assert (Hp2 : exists x,
    s |||- 12 + rget s rgcInfo |-> x **
      (fun h : Heap.heap => cheneyLoopPre (h, rfileOf s))).
  lle Hp; exists x; auto.
  clear Hp; destruct Hp2 as [x Hp].
  rewrite (lwStep Hp); auto.
  rewrite (swStep Hp); auto.
  stSimpl.

  assert (Hp' : cheneyLoopPre
     (uhset (rset s rroot x) (12 + rget s rgcInfo) (rget s rtoEnd))).
  unfold cheneyLoopPre in Hp |- * ; unfold cheneyLoopStOk in Hp |- * .
  stSimpl.
  destruct (sconjElim Hp) as [h1 [h2 [J [Hp1 Hp2]]]]; clear Hp.
  destruct Hp2 as [phi [G [W [M [F [Geq [StLt [Al Hp2]]]]]]]].
  exists phi; exists G; exists W; exists M; exists F.
  repeat (constructor; [assumption | idtac]).
  clear Geq StLt Al.
  unfold scanFieldStOk in Hp2 |- * .
  destruct Hp2 as [WF [D [Feq [Imp [LeLe Hp2]]]]].
  repeat (constructor; [auto | idtac]).
  clear WF D Feq Imp LeLe.
  unfold hpredTrue in Hp2.
  replace (Heap.heap_canon (heapOf (Heap.heap_canon h2, rfileOf s))) with
    (Heap.heap_canon h2) in Hp2.
  rewrite <- hpredTrueEq in Hp2.
  assert (Hp := sconjIntro Hp1 Hp2 J); clear J Hp1 Hp2.
  destruct s as [h r].
  replace (heapOf (uhset (rset (h, r) rroot x) 
    (12 + rget (h, r) rgcInfo) (rget (h, r) rtoEnd))) with
    (Heap.uset h (12 + rget (h, r) rgcInfo) (rget (h, r) rtoEnd)) by auto.
  replace (heapOf (h, r)) with h in Hp by auto.
  hypReplace Hp (contUpdateSt Hp (rget (h, r) rtoEnd)).
  hypReplace Hp (sconjComm Hp); linLogElim Hp.
  scm.
  simpl; unfold Heap.heap_canon.
  exact (Heap.HeapMap.map_canonEqvEq (Heap.HeapMap.map_canonEqv h2)).

  constructor.
  assumption.

  destruct Hp' as [phi [G [W [M [F Hp']]]]].
  intros s' Guar'; constructor.

  intros y A Hp2.
  assert (R := contContEq Hp2 Hp); subst.
  econstructor.
  constructor.
  2:constructor.
  3:eauto.
  hypReplace Hp2 (contUpdateSt Hp2 (rget s rtoEnd)).
  assumption.
  intro r; stSimpl; reflexivity.

  destruct Guar' as [_ [Req RA]].
  constructor.
  2:rewrite RA; stSimpl; reflexivity.
  intros r I; simpl in I; rewrite <- Req.
  destruct I; stSimpl; reflexivity.
  intuition; regNotInListSolver.

Qed.
