(* partial map implemented as a function *)

Section FnMap.

Variable A B : Set.
Variable beqEq : A -> A -> bool.

Definition map := A -> option B.

Definition empty_map : map := fun _ => None.

Definition map_get (m:map) (a:A) : option B := m a.

Definition map_add (m:map) (a:A) (b:B) : map :=
  fun a' =>
    if beqEq a a' then
      Some b
    else
      m a'.

Definition map_remove m x z :=
  if beqEq x z then
    (None (A := B))
  else
    m z.


Definition map_union (m1 m2 : map) : map :=
  fun a =>
    match m1 a with
      | Some b => Some b
      | None => m2 a
    end.


Lemma mapUnionEq : forall m x v,
  m x = Some v -> forall m',
    map_union m m' x = Some v.
Proof.
  unfold map_union.
  intros m x v E m'.
  destruct (m x).
  assumption.
  discriminate E.
Qed.


Definition mapEqv (m1 m2 : map) :=
  forall a, m1 a = m2 a.

Definition mapSub (m1 m2 : map) :=
  forall x y, m1 x = Some y -> m2 x = Some y.


Lemma mapEqvSub : forall m1 m2,
  mapEqv m1 m2 ->
  mapSub m1 m2.
Proof.
  intros m1 m2 E x y R.
  rewrite <- (E x); assumption.
Qed.


Lemma mapSubSound : forall m x y,
  m x = Some y ->
  forall m', mapSub m m' ->
    m' x = Some y.
Proof.
  intros phi x y R phi' S.
  exact (S x y R).
Qed.


Lemma mapUnionSub : forall s s',
  mapSub s (map_union s s').
Proof.
  intros s s' x y E.
  unfold map_union.
  rewrite E.
  reflexivity.
Qed.


(** setoids *)

Require Import Setoid.

Lemma mapEqvRefl : forall f, mapEqv f f.
Proof.
  intros f x.
  reflexivity.
Qed.

Lemma mapEqvSymm : forall f f', mapEqv f f' -> mapEqv f' f.
Proof.
  intros f f' M x.
  rewrite (M x).
  reflexivity.
Qed.

Lemma mapEqvTrans : forall f f' f'', 
  mapEqv f f' -> mapEqv f' f'' -> mapEqv f f''.
Proof.
  intros f f' f'' E1 E2 x.
  rewrite (E1 x); rewrite (E2 x).
  reflexivity.
Qed.


End FnMap.


Implicit Arguments map_add [A B].
Implicit Arguments map_union [A B].
Implicit Arguments mapUnionEq [A B].
Implicit Arguments mapEqv [A B].
Implicit Arguments mapSub [A B].
Implicit Arguments mapEqvSub [A B].
Implicit Arguments mapSubSound [A B].



(* this doesn't seem to be working if i show it in
   general, so let's fix it to the case we care about *)

Definition mapnn := map nat nat.
Definition mapEqvnn := mapEqv (A := nat) (B := nat).

Require Import EqNat.

Definition map_addnn : mapnn -> nat -> nat -> mapnn :=
  map_add (B := nat) beq_nat.

Definition map_unionnn : mapnn -> mapnn -> mapnn :=
  map_union (A := nat) (B := nat).

Definition mapSubnn : mapnn -> mapnn -> Prop :=
  mapSub (A := nat) (B := nat).


Lemma FnMapSetoid : Setoid_Theory mapnn mapEqvnn.
Proof.
  unfold mapnn; unfold mapEqvnn.
  constructor.
  apply mapEqvRefl.
  apply mapEqvSymm.
  apply mapEqvTrans.
Qed.


Add Setoid mapnn mapEqvnn FnMapSetoid as fnmap_mor.



Add Morphism map_addnn : map_addMorph.
Proof.
  unfold map_addnn; unfold map_add.
  intros m m' E a b x.
  unfold mapEqvnn in E; rewrite E; reflexivity.
Qed.


Add Morphism map_unionnn : map_unionMorph.
Proof.
  unfold map_unionnn; unfold map_union.
  unfold mapEqvnn; intros m1 m1' E1 m2 m2' E2 x.
  rewrite E1.
  rewrite E2.
  reflexivity.
Qed.


Add Morphism mapSubnn with signature mapEqvnn ==> mapEqvnn ==> iff as mapSubMorph.
Proof.
  unfold mapEqvnn, mapSubnn, mapSub.
  intros m1 m1' E1 m2 m2' E2; constructor; intros F x y E.
  rewrite <- E2; apply F; rewrite E1; auto.
  rewrite E2; apply F; rewrite <- E1; auto.
Qed.


