Require Import NatListSet.
Require Import Heap.
Require Import List.
Require Import OptUtil.
Require Import NatUtil.
Require Import LinLog.
Require Import StateTheory.

Import HeapMap.


Set Implicit Arguments.

(* properties of the domains of heaps described by lin log predicates *)


Lemma notInMapDomNone : forall h x,
  ~set_In x (map_dom h) ->
  map_get h x = None.
Proof.
  intros h x I.
  induction h; simpl.
  reflexivity.
  destruct a as [y v].
  simpl in I.
  destruct (POnat.POnat.eqDec x y).
  elimtype False; apply I; left; auto.
  apply IHh.
  intro; apply I; right; assumption.
Qed.


Lemma notInHeapDomNone : forall h x,
  ~set_In x (heap_dom h) ->
  get h x = None.
Proof.
  intros h x.
  unfold heap_dom; unfold get; intro NI.
  assert (X : ~exists y, x = 4 * y /\ set_In y (map_dom h)).
  intros [y [E I]].
  apply NI; clear NI; eapply set_map_intro; eauto.
  clear NI.
  assert (Y := isPtrOk x).
  destruct (isPtr x).
  2:reflexivity.
  apply notInMapDomNone.
  intro I; apply X; clear X.
  exists (div4 x).
  destruct (ptrP_elim Y) as [y E]; subst x.
  rewrite div44eq; rewrite div44eq in I. 
  intuition.
Qed.


Lemma contDom : forall h x y,
  h ||- x |-> y ->
    set_eqv (heap_dom h) (set_add x empty_set).
Proof.
  intros h x y Hp.
  destruct (contElim Hp) as [Px heq]; clear Hp.
  assert (X := heapDom_eqv heq).
  eapply set_eqvTrans; eauto.
  clear X.
  unfold singleton_heap; unfold heap_dom.
  rewrite singletonDom.
  simpl.
  replace (div4 x + (div4 x + (div4 x + (div4 x + 0)))) with (4 * div4 x).
  2:reflexivity.
  destruct (ptrP_elim Px) as [k E]; clear Px heq.
  subst x.
  rewrite div44eq.
  apply set_eqvRefl.
Qed.


Lemma contDom' : forall h x,
  h ||- x |->_ ->
    set_eqv (heap_dom h) (set_add x empty_set).
Proof.
  intros h x Hp.
  linLogElim Hp.
  eapply contDom; eauto.
Qed.


Lemma inMapInHeap : forall x m,
  set_In x (map_dom m) ->
  set_In (4 * x) (heap_dom m).
Proof.
  intros x m I; induction m.
  assumption.
  destruct a as [y v].
  simpl in I; destruct I.
  subst x.
  simpl; auto.
  hypReplace IHm (IHm H); clear H.
  right; assumption.
Qed.


Lemma inHeapInMap : forall x m,
  set_In (4 * x) (heap_dom m) ->
  set_In x (map_dom m).
Proof.
  intros x m I; induction m.
  assumption.
  destruct a as [y v].
  assert (D : y = x \/ set_In (4 * x) (heap_dom m)).
  simpl in I; intuition.
  left; unfold POnat.POnat.A in * |- *; omega.
  clear I; destruct D; [left | right]; auto.
Qed.
  

Lemma heap_get_In_dom : forall a m,
  match get m a with
    | Some _ => set_In a (heap_dom m)
    | None => ~set_In a (heap_dom m)
  end.
Proof.
  intros a m.
  unfold get.
  assert (Q := isPtrOk a).
  assert (X := map_mem_In_dom (div4 a) m).
  destruct (isPtr a).
  destruct (ptrP_elim Q) as [k H]; subst a; clear Q.
  rewrite div44eq; rewrite div44eq in X.
  destruct (map_get m k).
  apply inMapInHeap; assumption.
  intro I; apply X.
  apply inHeapInMap; assumption.

  intro I; apply Q; clear X Q.
  unfold heap_dom in I.
  destruct (set_map_elim I) as [b [Feq X]].
  rewrite Feq.
  eapply ptrP_intro; eauto.

Qed.


Lemma heap_joinDom : forall ha hb h,
  heap_join ha hb h ->
  set_eqv (heap_dom h) (heap_dom ha |_| heap_dom hb).
Proof.
  intros ha hb h J.
  intros x; constructor; intro I.

  assert (NE : get h x <> None).
  intro.
  assert (X := heap_get_In_dom x h).
  rewrite H in X.
  auto.

  unfold get in NE.
  assert (X := isPtrOk x).
  destruct (isPtr x).
  2:elimtype False; auto.
  destruct (ptrP_elim X) as [k E]; subst x.
  destruct (joinCoversDom J NE);
    unfold MapDef.get in H.

  apply set_union_intro1; apply inMapInHeap.
  rewrite div44eq in H.
  assert (Y := map_mem_In_dom k ha).
  destruct (map_get ha k).
  assumption.
  elimtype False; auto.

  apply set_union_intro2; apply inMapInHeap.
  rewrite div44eq in H.
  assert (Y := map_mem_In_dom k hb).
  destruct (map_get hb k).
  assumption.
  elimtype False; auto.

  (* other direction *)
  destruct (set_union_elim I) as [I' | I']; clear I.

  (* element in ha *)
  assert (X := heap_get_In_dom x ha).
  unfold get in X.
  assert (Y := isPtrOk x).
  destruct (isPtr x).
  2:elimtype False; auto.
  destruct (ptrP_elim Y) as [y E]; subst x; clear Y.
  rewrite div44eq in X.
  assert (Y := mapJoinSoundLeft J); hypReplace Y (Y y).
  unfold MapDef.get in Y; destruct (map_get ha y).
  assert (map_get h y = Some b).
  eauto.
  clear Y.
  assert (Z := heap_get_In_dom (4 * y) h).
  unfold get in Z; rewrite div44eq in Z.
  rewrite H in Z.
  rewrite isPtrEq in Z.
  assumption.
  eapply ptrP_intro; eauto.
  intuition.
 
  (* element in hb *)
  assert (X := heap_get_In_dom x hb).
  unfold get in X.
  assert (Y := isPtrOk x).
  destruct (isPtr x).
  2:elimtype False; auto.
  destruct (ptrP_elim Y) as [y E]; subst x; clear Y.
  rewrite div44eq in X.
  assert (Y := mapJoinSoundRight J); hypReplace Y (Y y).
  unfold MapDef.get in Y; destruct (map_get hb y).
  assert (map_get h y = Some b).
  eauto.
  clear Y.
  assert (Z := heap_get_In_dom (4 * y) h).
  unfold get in Z; rewrite div44eq in Z.
  rewrite H in Z.
  rewrite isPtrEq in Z.
  assumption.
  eapply ptrP_intro; eauto.
  intuition.
 
Qed.


Lemma sconjDom : forall h A B,
  h ||- A ** B ->
    forall DomA DomB,
      (forall h', h' ||- A -> set_eqv (heap_dom h') DomA) ->
      (forall h', h' ||- B -> set_eqv (heap_dom h') DomB) ->
      set_eqv (heap_dom h) (DomA |_| DomB).
Proof.
  intros h A B Hp DomA DomB D1 D2.
  destruct (sconjElim Hp) as [ha [hb [J [HA HB]]]]; clear Hp.
  hypReplace D1 (D1 _ HA); clear HA.
  hypReplace D2 (D2 _ HB); clear HB.
  assert (X := heap_joinDom J).
  eapply set_eqvTrans; eauto.
  intro x; constructor; intro I; destruct (set_union_elim I); clear I.
  apply set_union_intro1; rewrite <- D1; assumption.
  apply set_union_intro2; rewrite <- D2; assumption.
  rewrite D1; apply set_union_intro1; auto.
  rewrite D2; apply set_union_intro2; auto.
Qed.


Lemma fasDom : forall h (A : nat -> hpred) s,
  h ||- fas A s ->
    forall DomA,
      (forall x h', h' ||- A x -> set_eqv (heap_dom h') (DomA x)) ->
      set_eqv (heap_dom h) (set_mmap DomA s).
Proof.
  intros h A s Hp DomA D.
  unfold fas in Hp.

  assert (E : set_eqv (heap_dom h) (set_mmap DomA (set_canon s))).
  assert (X := set_canonCanonP s).
  generalize dependent h.
  generalize dependent (set_canon s); clear s.
  induction 1; simpl; intros h Hp.

  assert (X := unitElim Hp); clear Hp D DomA A.
  assert (Y := heapDom_eqv X).
  eapply set_eqvTrans; eauto.
  unfold heap_dom; simpl; apply set_eqvRefl.

  eapply sconjDom; eauto.

  rewrite E; apply set_mmap_eqv; apply set_eqvSymm; apply set_canonEqv.

Qed.


Lemma heap_domEqvElim : forall h1 h,
  subHeap h1 h -> forall h2,
  subHeap h2 h ->
  set_eqv (heap_dom h1) (heap_dom h2) ->
  heap_eqv h1 h2.
Proof.
  intros h1 h S1 h2 S2 E x.

  assert (E' : Heap.get h1 (4 * x) = Heap.get h2 (4 * x)).
  hypReplace S1 (S1 (4 * x)).
  hypReplace S2 (S2 (4 * x)).

  assert (X1 := heap_get_In_dom (4 * x) h1).
  assert (X2 := heap_get_In_dom (4 * x) h2).

  destruct (get h1 (4 * x)); destruct (get h2 (4 * x)).
  assert (n = n0).
  assert (Some n = Some n0).
  rewrite <- S1.
  apply S2.
  reflexivity.
  reflexivity.
  injection H; auto.
  subst n0.
  reflexivity.

  elf; apply X2; rewrite <- E; auto.
  elf; apply X1; rewrite E; auto.
  
  reflexivity.

  unfold get in E'.
  rewrite isPtrEq in E'.
  rewrite div44eq in E'; assumption.
  eapply ptrP_intro; reflexivity.

Qed.


Lemma inHeapInMap' : forall y m,
  set_In y (heap_dom m) ->
  exists x,
    y = 4 * x /\
    set_In x (map_dom m).
Proof.
  intros y m I; induction m.

  elimtype False; exact I.

  destruct a as [x v].
  assert (D : 4 * x = y \/ set_In y (heap_dom m)).
  simpl in I; intuition.
  clear I; destruct D.
  exists x; constructor.
  symmetry; assumption.
  left; reflexivity.
  destruct (IHm H) as [x' [Xeq Xi]].
  exists x'; constructor.
  assumption.
  right; assumption.

Qed.


Lemma map_eqvSubsetDom : forall h h',
  map_eqv h h' ->
  set_subset (map_dom h) (map_dom h').
Proof.
  intros h h' E x I.

  assert (E' :  MapDef.get h x <> None -> MapDef.get h' x <> None).
  hypReplace E (E x); intro E'.
  destruct (MapDef.get h x).
  rewrite <- E; discriminate.
  elimtype False; apply E'; reflexivity.

  clear E; rename E' into E.
  generalize dependent x.
  generalize dependent h'.

  induction h; intros.

  elimtype False; exact I.

  destruct a as [q v].
  simpl in I; destruct I.

  subst q.
  clear IHh.

  assert (E' : MapDef.get h' x <> None).
  apply E; simpl.
  destruct (POnat.POnat.eqDec x x).
  discriminate.
  elimtype False; apply n; reflexivity.

  clear E; rename E' into E.
  induction h'.
  elimtype False; apply E; reflexivity.
  destruct a as [x' v']; simpl in E.
  destruct (POnat.POnat.eqDec x x').
  subst x'; left; reflexivity.
  right; apply IHh'; assumption.

  apply IHh.
  assumption.
  intro E'; simpl in E; apply E; clear E.
  destruct (POnat.POnat.eqDec x q).
  discriminate.
  assumption.

Qed.


Lemma map_eqvDomEqv : forall h h',
  map_eqv h h' ->
  set_eqv (map_dom h) (map_dom h').
Proof.
  intros h h' E; constructor; intro I.
  exact (map_eqvSubsetDom E _ I).
  assert (E' : map_eqv h' h).
  intros x; symmetry; apply E.
  exact (map_eqvSubsetDom E' _ I).
Qed.


Lemma heap_eqvDomEqv : forall h h',
  heap_eqv h h' ->
  set_eqv (heap_dom h) (heap_dom h').
Proof.
  intros h h' E; constructor; intros I.
  destruct (inHeapInMap' _ _ I) as [y [Yeq Yin]]; clear I.
  subst a.
  apply inMapInHeap.
  rewrite <- (map_eqvDomEqv E); auto.
  destruct (inHeapInMap' _ _ I) as [y [Yeq Yin]]; clear I.
  subst a.
  apply inMapInHeap.
  rewrite (map_eqvDomEqv E); auto.
Qed.
Implicit Arguments heap_eqvDomEqv [h h'].


Lemma heap_removeDomIntro : forall x h,
  set_In x (heap_dom h) -> forall a,
  x <> a ->
  ptrP a ->
  set_In x (heap_dom (heap_remove h a)).
Proof.
  intros x h I a Pa NE.
  destruct (inHeapInMap' _ _ I) as [y [Xeq Yin]]; clear I.
  subst x.
  apply inMapInHeap.
  induction h.

  exact Yin.

  destruct a0 as [b v].
  unfold heap_remove; simpl.

  simpl in Yin; destruct Yin.
  subst b.
  destruct (POnat.POnat.eqDec (div4 a) y).
  subst y.
  elimtype False; apply Pa; clear IHh Pa.
  destruct (ptrP_elim NE) as [k E]; subst a.
  rewrite div44eq; reflexivity.
  left; reflexivity.

  hypReplace IHh (IHh H); clear H.
  destruct (POnat.POnat.eqDec (div4 a) b).
  assumption.
  simpl; right; assumption.

Qed.


Lemma heap_removeDomElim : forall x h a,
  set_In x (heap_dom (heap_remove h a)) ->
  ptrP a ->
  set_In x (heap_dom h) /\ x <> a.
Proof.
  unfold heap_remove.
  intros x h a I Pa.
  destruct (ptrP_elim Pa) as [b E]; subst a; clear Pa.
  rewrite div44eq in I.
  destruct (inHeapInMap' _ _ I) as [y [Xeq Yin]]; clear I.
  subst x.

  assert (H :  set_In y (map_dom h) /\ y <> b).
  induction h.

  elimtype False; exact Yin.

  destruct a as [a v].
  simpl.
  simpl in Yin.
  destruct (POnat.POnat.eqDec b a).
  subst b.
  hypReplace IHh (IHh Yin); clear Yin.
  destruct IHh as [I NE].
  constructor.
  right; assumption.
  assumption.

  simpl in Yin; destruct Yin.
  subst a.
  constructor.
  left; reflexivity.
  auto.
  destruct (IHh H) as [I NE]; clear IHh H.
  constructor; auto.

  destruct H as [I NE].
  constructor.
  apply inMapInHeap; assumption.
  omega.

Qed.


Lemma singleton_mapSplitJoin : forall h (v x : nat),
  map_get h x = Some v ->
  heap_join (singleton_map x v) (map_remove h x) h.
Proof.
  intros h v x E.
  constructor.
  unfold MapDef.A; unfold POnat.POnat.A; intros x'.
  simpl; compare x x'; intro NE.

  subst x'.
  destruct (POnat.POnat.eqDec x x).
  clear e.
  unfold MapDef.get; rewrite E.
  rewrite (map_getRemove x h).
  constructor.
  elimtype False; auto.

  destruct (POnat.POnat.eqDec x' x).
  elimtype False; auto.
  clear n.
  rewrite (map_getRemoveNE NE).
  unfold MapDef.get.
  destruct (map_get h x'); constructor.

Qed.


Lemma singleton_heapSplitJoin : forall h a v,
  get h a = Some v ->
  heap_join (singleton_heap a v) (heap_remove h a) h.
Proof.
  unfold get; unfold singleton_heap; unfold heap_remove.
  intros h a v.
  assert (Pa := isPtrOk a); destruct (isPtr a).
  2:discriminate 1.
  destruct (ptrP_elim Pa) as [k x]; subst a; clear Pa.
  rewrite div44eq.
  apply singleton_mapSplitJoin.
Qed.


Lemma heapDomToCont : forall X h,
  set_eqv X (Heap.heap_dom h) ->
  h ||- foralls x in X, x |->_.
Proof.
  intros X h Eq.
  unfold fas.
  assert (C : set_canonP (set_canon X)).
  apply set_canonCanonP.

  assert (Eq' : set_eqv (set_canon X) (heap_dom h)).
  setoid_replace (set_canon X) with X.
  auto.
  rewrite <- set_canonEqv; apply set_eqvRefl.

  clear Eq; rename Eq' into Eq.
  generalize dependent h.
  generalize dependent (set_canon X); clear X.
  induction 1.

  (* base case *)
  simpl; intros h Eq; destruct h.
  apply unitIntro.
  destruct p as [x v]; elimtype False; hypReplace Eq (Eq (4 * x)).
  destruct Eq as [_ Q]; apply Q.
  left; reflexivity.
  
  (* inductive case *)
  intros h Eq.
  simpl.
  assert (X : exists v, get h a = Some v).
  assert (X := heap_get_In_dom a h).
  destruct (get h a).
  econstructor; reflexivity.
  elimtype False; apply X; clear X.
  destruct (Eq a) as [Q _]; apply Q; left; reflexivity.

  destruct X as [v E].

  assert (Pa : ptrP a).
  unfold get in E; assert (U := isPtrOk a); destruct (isPtr a).
  assumption.
  discriminate E.

  lli; instantiate (1 := v).
  eapply sconjIntro.
  instantiate (1 := singleton_heap a v).
  apply contIntro.
  assumption.
  instantiate (1 := heap_remove h a).
  apply IHC.
  intros x; destruct (Eq x) as [Eq1 Eq2]; clear Eq.
  constructor; intro I.
  apply heap_removeDomIntro.
  apply Eq1; right; assumption.
  hypReplace H (H _ I).
  unfold POnat.POnat.LT in H; omega.
  assumption.

  destruct (heap_removeDomElim _ _ I Pa) as [I' NE].
  hypReplace I' (Eq2 I'); clear Eq2; simpl in I'.
  destruct I'.
  elimtype False; auto.
  assumption.

  apply singleton_heapSplitJoin; assumption.

Qed.


Lemma sconjDomSubL : forall h A B,
  h ||- A ** B ->
    forall DomA,
      (forall h', h' ||- A -> set_subset DomA (heap_dom h')) ->
      set_subset DomA (heap_dom h).
Proof.
  intros h A B Hp DomA D.
  destruct (sconjElim Hp) as [ha [hb [J [HA HB]]]]; clear Hp.
  hypReplace D (D _ HA); clear HA.
  apply set_subsetTrans with (heap_dom ha).
  assumption.
  intros x I.
  assert (I' : set_In x (heap_dom ha |_| heap_dom hb)).
  apply set_union_intro1; assumption.
  destruct (heap_joinDom J x); auto.
Qed.


Lemma sconjDomSubR : forall h A B,
  h ||- A ** B ->
    forall DomB,
      (forall h', h' ||- B -> set_subset DomB (heap_dom h')) ->
      set_subset DomB (heap_dom h).
Proof.
  intros h A B Hp DomB D.
  sconjLift2of2 Hp; eapply sconjDomSubL; eauto.
Qed.


Lemma heapSubDomToCont : forall X h,
  set_subset X (Heap.heap_dom h) ->
  h ||- (foralls x in X, x |->_) ** top.
Proof.
  intros X h Sub.
  unfold fas.
  assert (C : set_canonP (set_canon X)).
  apply set_canonCanonP.

  assert (Sub' : set_subset (set_canon X) (heap_dom h)).
  intros x I.
  apply Sub.
  rewrite <- set_canonEqv in I; auto.

  clear Sub; rename Sub' into Sub.
  generalize dependent h.
  generalize dependent (set_canon X); clear X.
  induction 1.

  (* base case *)
  intros; simpl; lli.
  
  (* inductive case *)
  intros h Sub.
  simpl; apply sconjAssocR.
  assert (X : exists v, get h a = Some v).
  assert (X := heap_get_In_dom a h).
  destruct (get h a).
  econstructor; reflexivity.
  elimtype False; apply X; clear X; apply Sub; left; reflexivity.
  destruct X as [v E].
  lli; instantiate (1 := v).
  eapply sconjIntro.
  instantiate (1 := singleton_heap a v).
  apply contIntro.
  unfold get in E.
  assert (U := isPtrOk a); destruct (isPtr a).
  assumption.
  discriminate E.

  instantiate (1 := heap_remove h a).
  apply IHC.
  intros x I.
  apply heap_removeDomIntro.
  apply Sub; right; assumption.
  hypReplace H (H _ I).
  unfold POnat.POnat.LT in H; omega.
  assert (Q := isPtrOk a); unfold get in E; destruct (isPtr a).
  assumption.
  discriminate E.
  apply singleton_heapSplitJoin; assumption.

Qed.


Lemma sconjDomDisj : forall h A B,
  h ||- A ** B ->
    forall DomA DomB,
      (forall h', h' ||- A -> set_eqv (heap_dom h') DomA) ->
      (forall h', h' ||- B -> set_eqv (heap_dom h') DomB) ->
      set_disj DomA DomB.
Proof.
  intros h A B Hp DomA DomB D1 D2.
  destruct (sconjElim Hp) as [ha [hb [J [HA HB]]]]; clear Hp.
  hypReplace D1 (D1 _ HA); clear HA.
  hypReplace D2 (D2 _ HB); clear HB.

  assert (h ||- (foralls x in DomA, x |->_) ** (foralls x in DomB, x |->_)).
  eapply sconjIntro.
  3:eauto.
  apply heapDomToCont; apply set_eqvSymm; assumption.
  apply heapDomToCont; apply set_eqvSymm; assumption.
  eapply fasFasDisj.
  instantiate (1 := unit).
  instantiate (3 := h).
  lli; eauto.

  cbv beta; intros h0 x0 Hp0; hypReplace Hp0 (unitIR Hp0); sconjMatch.
  cbv beta; intros h0 x0 Hp0; hypReplace Hp0 (unitIR Hp0); sconjMatch.

Qed.


Lemma objBufferDom : forall h s,
  (h ||- foralls x in s, x |->__) ->
    set_eqv (Heap.heap_dom h) (set_mmap (fun x => set_add x (set_add (x + 4) empty_set)) s).
Proof.
  intros h s Hp.
  eapply fasDom; eauto.
  clear h Hp s.
  intros x h Hp.
  setoid_replace (set_add x (set_add (x + 4) empty_set)) with
    ((set_add x empty_set) |_| (set_add (x + 4) empty_set)).
  simpl in Hp.
  eapply sconjDom; eauto.
  clear h Hp; intros h Hp; apply contDom'; assumption.
  clear h Hp; intros h Hp.
  replace (x + 4) with (4 + x).
  apply contDom'; assumption.
  omega.
  apply set_eqvRefl.
Qed.
Implicit Arguments objBufferDom [h s].


Lemma objsHeapEqvEqv : forall h S h1, 
  h ||- ((foralls x in S, x |->__) && heap_eqv h1) ** top -> forall h2,
  h ||- ((foralls x in S, x |->__) && heap_eqv h2) ** top ->
  heap_eqv h1 h2.
Proof.
  intros h S h1 Hp h2 Hp'.
  eapply heap_domEqvElim.
  instantiate (1 := h).
  destruct (aconjElim (sconjDistrib2 Hp)); eapply heap_eqvSubHeapL; eauto.
  destruct (aconjElim (sconjDistrib2 Hp')); eapply heap_eqvSubHeapL; eauto.
  assert (Hp1 := heapEqvImplL Hp); clear Hp.
  assert (Hp2 := heapEqvImplL Hp'); clear Hp'.
  
  assert (set_eqv (heap_dom h2)
    (set_mmap (fun x => set_union (set_add x empty_set) (set_add (4 + x) empty_set)) S)).
  eapply fasDom; eauto.
  cbv beta; intros x h' Hp.
  eapply sconjDom; eauto.
  intros; eapply contDom'; eauto.
  intros; eapply contDom'; eauto.

  assert (set_eqv (heap_dom h1)
    (set_mmap (fun x => set_union (set_add x empty_set) (set_add (4 + x) empty_set)) S)).
  eapply fasDom; eauto.
  cbv beta; intros x h' Hp.
  eapply sconjDom; eauto.
  intros; eapply contDom'; eauto.
  intros; eapply contDom'; eauto.

  eapply set_eqvTrans; eauto.
  apply set_eqvSymm; assumption.

Qed.
