Require Import List.
Require POtype.
Require CanonListSet.
Require MapJoin.


Module ListMap (Adef : POtype.POtype) (Bdef : POtype.OneType).

  Import Adef.
  Import Bdef.

  Set Implicit Arguments.


  (** ** Basic definitions *)

  Definition map := list (prod A B).

  Definition empty_map : map := nil.

  Definition map_add (m:map) (a:A) (b:B) : map := (a, b)::m.

  Fixpoint map_get (m:map) (a:A) {struct m} : option B :=
    match m with
      | nil => None
      | (a', b)::m' =>
	if eqDec a a' then
	  Some b
	else
	  map_get m' a
    end.

  Definition singleton_map : A -> B -> map := map_add empty_map.

  Fixpoint map_diff (m1 m2:map) {struct m1} : map :=
    match m1 with
    | nil => nil
    | (a1, v) :: m1' =>
      match map_get m2 a1 with
	| Some _ => map_diff m1' m2
	| None => map_add (map_diff m1' m2) a1 v
      end
    end.

  Definition map_union : map -> map -> map := app (A := prod A B).

  Definition map_mapRange (f:B->B) (m:map) : map :=
    List.map (fun (p:prod A B) => let (x, y) := p in (x, f y)) m.

  Definition map_mapDom (f:A->A) (m:map) : map :=
    List.map (fun (p:prod A B) => let (x, y) := p in (f x, y)) m.

  Fixpoint map_remove (m:map) (a:A) {struct m} : map :=
    match m with
      | nil => empty_map
      | (a', b)::m' => match (eqDec a a') with
                    | left _ => map_remove m' a
                    | right _ => (a', b)::(map_remove m' a)
                  end
    end.


  (** ** Join *)

  Module MapDef <: MapJoin.MapDef.
    Definition A := A.
    Definition B := B.
    Definition map := map.
    Definition get := map_get.
    Definition Aeq_dec := eqDec.
  End MapDef.

  Module ABJoin := MapJoin.M MapDef.
  Export ABJoin.

  Module CD <: CombineDef.
    Definition mapCombine (x y:map) : map := x ++ y.

    Lemma mapCombineOk : combineSemantics mapCombine.
    Proof.
      unfold combineSemantics.
      intros m1 m2 a.
      induction m1.
      reflexivity.
      destruct a0 as [a0 b0].
      simpl.
      destruct (eqDec a a0); auto.
    Qed.

  End CD.


  Module LMC := Combine (CD).
  Export LMC.

  Module EMD <: EmptyMapDef.
    Definition empty_map := empty_map.

    Lemma getEmptyNone : forall w,
      map_get empty_map w = None.
    Proof.
      reflexivity.
    Qed.

  End EMD.

  Module LME := JoinUnit (EMD).
  Export LME.


  Module MAD <: MapAddDef.
    Definition add := map_add.

    Lemma getAddEq : forall m a b,
      map_get (map_add m a b) a = Some b.
    Proof.
      intros.
      simpl.
      destruct (eqDec a a); intuition.
    Qed.

    Lemma getAddNeq : forall m a a' b,
      a <> a' ->
      map_get (map_add m a b) a' = map_get m a'.
    Proof.
      intros.
      simpl.
      destruct (eqDec a' a).
      elimtype False; auto.
      reflexivity.
    Qed.

  End MAD.

  Module MA := MapAdd (MAD).
  Export MA.



  (** ** Canonizing *)

  Module ASet := CanonListSet.M Adef.
  Import ASet.

  Definition map_dom (m:map) : set :=
    List.map (fun x => fst x) m.

  Definition map_getL (m:map) (a:A) :=
    match map_get m a with
      | Some b => (a, b)::nil
      | None => nil
    end.

  Definition map_canon (m:map) : map :=
    List.flat_map (map_getL m) (set_canon (map_dom m)).

  Definition map_canonP (m:map) : Prop :=
    set_canonP (map_dom m).

  Lemma map_mem_In_dom : forall a m,
    match map_get m a with
      | Some _ => set_In a (map_dom m)
      | None => ~set_In a (map_dom m)
    end.
  Proof.
    induction m; simpl.
    
    auto.
    
    destruct a0 as [a' b'].
    simpl.
    destruct (eqDec a a').
    left; auto.
    destruct (map_get m a).
    right; assumption.
    intro Z.
    destruct Z; auto.
  Qed.


  Lemma mapDomEqv : forall m1 m2,
    map_eqv m1 m2 ->
    set_eqv (map_dom m1) (map_dom m2).
  Proof.
    intros m1 m2 E a.
    assert (Ea := E a).
    assert (Z1 := map_mem_In_dom a m1).
    assert (Z2 := map_mem_In_dom a m2).
    unfold MapDef.get in Ea; rewrite Ea in Z1.
    destruct (map_get m2 a).
    constructor; auto.
    constructor; intro; elimtype False; auto.
  Qed.


  Theorem map_canonEqvEq : forall m1 m2,
    map_eqv m1 m2 ->
    map_canon m1 = map_canon m2.
  Proof.
    unfold map_canon.
    intros m1 m2 ME.
    assert (SE := mapDomEqv ME).
    rewrite (set_canonEqvEq SE).
    clear SE.
    generalize (set_canon (map_dom m2)).
    induction s.
    reflexivity.
    simpl.
    rewrite IHs.
    unfold map_getL.
    assert (R := ME a).
    unfold MapDef.get in R; rewrite R.
    destruct (map_get m2 a); reflexivity.
  Qed.


  Lemma empty_mapCanonP : map_canonP empty_map.
  Proof.
    constructor.
  Qed.

    
  Lemma canonStepSome : forall m a s b,
    map_get m a = Some b -> set_In a s ->
    map_get (flat_map (map_getL m) s) a = Some b.
  Proof.
    induction s; simpl.

    intuition.

    intros b Eq Z.
    assert (IH := IHs _ Eq); clear IHs.
    unfold map_getL at 1.
    destruct (eqDec a0 a).
    subst.
    destruct (map_get m a).
    simpl.
    destruct (eqDec a a).
    assumption.
    elimtype False; apply n; reflexivity.
    discriminate.
    
    destruct Z.
    elimtype False; apply n; assumption.
    destruct (map_get m a0).
    simpl.
    destruct (eqDec a a0).
    elimtype False; apply n; auto.
    auto.
    auto.
  Qed.


  Lemma canonStepNone : forall m a s,
    map_get m a = None ->
    map_get (flat_map (map_getL m) s) a = None.
  Proof.
    induction s.
    reflexivity.
    simpl.
    intro E.
    unfold map_getL at 1.
    destruct (eqDec a a0).
    subst.
    rewrite E.
    simpl.
    apply IHs; assumption.
    
    destruct (map_get m a0).
    simpl.
    destruct (eqDec a a0).
    elimtype False; auto.
    auto.
    auto.
  Qed.
  
  
  Lemma canonStep : forall m a s,
    (map_get m a <> None -> set_In a s) ->
    map_get (flat_map (map_getL m) s) a = map_get m a.
  Proof.
    intros m a s I.
    assert (ZS := canonStepSome m a s).
    assert (ZN := canonStepNone m a s).
    destruct (map_get m a).
    apply ZS; auto.
    apply I; discriminate.
    apply ZN; auto.
  Qed.
 
 
  Lemma notInDomIsNone : forall a m,
    ~set_In a (map_dom m) -> map_get m a = None.
  Proof.
    induction m.
    reflexivity.
    
    destruct a0 as [a0 b0].
    simpl.
    intros.
    destruct (eqDec a a0).
    subst.
    elimtype False; apply H; left; reflexivity.
    apply IHm.
    intro; apply H; auto.
  Qed.


  Lemma map_canonEqv : forall m,
    map_eqv m (map_canon m).
  Proof.
    intros m a.
    unfold MapDef.get; unfold map_canon.
    destruct (set_canonEqv (map_dom m) a) as [Q _].
    assert (Z := canonStep m a (set_canon (map_dom m))).
    symmetry.
    apply Z.
    clear Z.
    intro Ne.
    destruct (set_In_dec a (map_dom m)).
    auto.
    elimtype False; apply Ne; apply notInDomIsNone; auto.
  Qed.
  

  Lemma mapJoinCanon : forall ma mb m,
    mapJoin ma mb m ->
    mapJoin (map_canon ma) (map_canon mb) (map_canon m).
  Proof.
    intros ma mb h J.
    eapply mapJoinEqv.
    eauto.
    apply map_canonEqv.
    apply map_canonEqv.
    apply map_canonEqv.
  Qed.


  Lemma map_canonCanonP : forall m,
    map_canonP (map_canon m).
  Proof.
    unfold map_canonP; unfold map_canon.
    intro m.
    assert (Z := set_canonCanonP (map_dom m)).
    generalize dependent (set_canon (map_dom m)).
    assert (forall x, map_getL m x = nil \/
      exists y, map_getL m x = (x,y)::nil).
    unfold map_getL.
    intro x; destruct (map_get m x).
    right; econstructor; reflexivity.
    left; reflexivity.

    generalize dependent (map_getL m).
    clear m.
    induction s.
    auto.
    
    simpl.
    destruct (H a).
    rewrite H0; simpl.
    inversion_clear 1.
    apply IHs; assumption.
    destruct H0 as [y R].
    rewrite R; simpl.
    clear R y.
    inversion_clear 1.
    constructor.
    apply IHs.
    assumption.
    intros a' I.
    apply H1.
    clear IHs H0 H1 a.
    induction s.
    assumption.

    simpl; simpl in I.
    destruct (H a); clear H.
    rewrite H0 in I; simpl in I.
    right; apply IHs; assumption.
    destruct H0 as [y R]; rewrite R in I; simpl in I.
    destruct I.
    left; assumption.
    right; apply IHs; assumption.
  Qed.


  Lemma map_canonPEq : forall m,
    map_canonP m -> map_canon m = m.
  Proof.
    unfold map_canonP; unfold map_canon.
    intros m C.
    rewrite (set_canonCanonPEq C).
    induction m.
    reflexivity.
    destruct a as [a b]; simpl.
    unfold map_getL at 1; simpl.
    destruct (eqDec a a).
    clear e.
    simpl.
    inversion_clear C.
    assert (forall a', set_In a' (map_dom m) ->
      (map_getL ((a, b)::m) a' = map_getL m a')).
    intros.
    assert (Z := H0 _ H1).
    unfold map_getL at 1.
    simpl.
    destruct (eqDec a' a).
    leafSolve.
    unfold map_getL.
    reflexivity.
    replace (flat_map (map_getL ((a, b) :: m)) (map_dom m)) with
      (flat_map (map_getL m) (map_dom m)).
    rewrite IHm; auto.
    clear IHm H H0.
    generalize dependent (map_dom m).
    generalize dependent (map_getL ((a, b) :: m)).
    intros.
    induction s.
    reflexivity.
    simpl.
    simpl in H1.
    rewrite (H1 a0).
    rewrite IHs.
    reflexivity.
    intros; apply H1; auto.
    auto.
    elimtype False; apply n; reflexivity.
  Qed.


  Lemma map_canonCanonEq : forall m,
    map_canon (map_canon m) = map_canon m.
  Proof.
    intro m.
    apply map_canonPEq.
    apply map_canonCanonP.
  Qed.


  (** going from a set to a map *)
  Section SetToMap.
    Variable f : A -> B.

    Definition setToMap (s:set) : map := List.map (fun x => (x, f x)) s.

    Lemma setToMap_get : forall x s,
      set_In x s ->
      map_get (setToMap s) x = Some (f x).
    Proof.
      induction s; simpl.
      intuition.
      destruct 1.
      subst x; destruct (eqDec a a); intuition.
      destruct (eqDec x a).
      subst; reflexivity.
      eauto.
    Qed.


  End SetToMap.


  (** reasoning about singletons *)

  Lemma singletonGetEq : forall a b,
    map_get (singleton_map a b) a = Some b.
  Proof.
    intros.
    simpl.
    destruct (eqDec a a); intuition.
  Qed.


  Lemma singletonGetNeq : forall a a' b,
    a <> a' ->
    map_get (singleton_map a b) a' = None.
  Proof.
    intros.
    simpl.
    destruct (eqDec a' a).
    elimtype False; auto.
    reflexivity.
  Qed.


  Lemma singletonAddEqv : forall x v1 v2,
    map_eqv (map_add (singleton_map x v1) x v2) (singleton_map x v2).
  Proof.
    intros.
    unfold singleton_map; unfold map_add.
    unfold map_eqv.
    intros a.
    simpl.
    destruct (eqDec a x); reflexivity.
  Qed.


  Lemma singletonDom : forall x y,
    map_dom (singleton_map x y) = set_add x empty_set.
  Proof.
    intros.
    unfold singleton_map.
    reflexivity.
  Qed.

  (** reasoning about MapDiff *)

  Lemma map_domDiffEqv : forall m m',
    set_eqv (map_dom (map_diff m m'))
    (set_diff (map_dom m) (map_dom m')).
  Proof.
    intros m m'.
    induction m.

    simpl; apply set_eqvRefl.

    destruct a as [a v].
    simpl.
    assert (Y := map_mem_In_dom a m').
    assert (Z := set_memSound a (map_dom m')).
    destruct (map_get m' a).
    destruct (set_mem a (map_dom m')).
    assumption.
    elimtype False; exact (Z Y).
    destruct (set_mem a (map_dom m')).
    elimtype False; exact (Y Z).
    simpl.
    constructor; intro I.

    simpl in I; destruct I.
    subst a0.
    apply set_add_intro1; reflexivity.

    destruct (IHm a0) as [IHm' _]; clear IHm.
    apply set_add_intro2; auto.

    destruct (IHm a0) as [_ IHm']; clear IHm.
    simpl; destruct (set_add_elim I).
    left; symmetry; assumption.
    right; auto.

  Qed.


  Lemma map_get_diffNotInEq : forall x m',
    ~set_In x (map_dom m') ->
    forall m,
      map_get m x = map_get (map_diff m m') x.
  Proof.
    intros x m' NI.
    induction m; simpl.

    reflexivity.
    
    destruct a as [a va].
    destruct (eqDec x a).
    clear IHm.
    subst a.
    assert (X := map_mem_In_dom x m').
    destruct (map_get m' x).
    elimtype False; exact (NI X).
    simpl; destruct (eqDec x x).
    reflexivity.
    elimtype False; apply n; reflexivity.

    destruct (map_get m' a).
    assumption.
    simpl; destruct (eqDec x a).
    elimtype False; auto.
    assumption.
  Qed.


  Lemma map_get_diffInEq : forall x m',
    set_In x (map_dom m') ->
    forall m,
      map_get (map_diff m m') x = None.
  Proof.
    intros x m' NI.
    induction m; simpl.

    reflexivity.
    
    destruct a as [a va].
    assert (X := map_mem_In_dom a m').
    destruct (map_get m' a).
    assumption.

    simpl.
    destruct (eqDec x a).
    subst.
    elimtype False; exact (X NI).
    assumption.

  Qed.

  
  Lemma map_getRemove : forall x m,
    map_get (map_remove m x) x = None.
  Proof.
    induction m.
    reflexivity.
    destruct a as [x' v].
    simpl.
    destruct (eqDec x x').
    subst x'; assumption.
    simpl.
    destruct (eqDec x x').
    elimtype False; auto.
    assumption.
  Qed.


  Lemma map_getRemoveNE : forall x y,
    x <> y -> forall m,
    map_get (map_remove m x) y = map_get m y.
  Proof.
    intros x y NE; induction m.
    reflexivity.
    destruct a as [y' v]; simpl.
    destruct (eqDec x y').
    subst y'.
    destruct (eqDec y x).
    elimtype False; auto.
    assumption.
    simpl.
    destruct (eqDec y y'); auto.
  Qed.


  (* map_mapRange lemmas *)
  Lemma map_mapRangeGetEq : forall m x f,
    match map_get m x with
      | Some y => map_get (map_mapRange f m) x = Some (f y)
      | None => map_get (map_mapRange f m) x = None
    end.
  Proof.
    induction m; simpl; intros x f.
    reflexivity.
    destruct a as [a b].
    destruct (eqDec x a).
    reflexivity.
    apply IHm.
  Qed.


  Lemma map_mapRangeGetSome : forall m x y,
    map_get m x = Some y -> forall f,
    map_get (map_mapRange f m) x = Some (f y).
  Proof.
    intros m x y E f.
    assert (X := map_mapRangeGetEq m x f).
    rewrite E in X; assumption.
  Qed.


  Lemma map_mapRangeGetNone : forall m x,
    map_get m x = None -> forall f,
    map_get (map_mapRange f m) x = None.
  Proof.
    intros m x E f.
    assert (X := map_mapRangeGetEq m x f).
    rewrite E in X; assumption.
  Qed.


  Lemma map_mapRangeEqv : forall m m',
    map_eqv m m' -> forall f,
    map_eqv (map_mapRange f m) (map_mapRange f m').
  Proof.
    intros m m' E f a.
    assert (X := E a); clear E.
    assert (Y := map_mapRangeGetEq m a f).
    assert (Y' := map_mapRangeGetEq m' a f).
    unfold MapDef.get in * .
    destruct (map_get m a); destruct (map_get m' a);
      try (discriminate X).
    injection X; clear X; intro; subst b0.
    rewrite Y; rewrite Y'; reflexivity.
    rewrite Y; rewrite Y'; reflexivity.
  Qed.


  Lemma map_mapRangeSingleton : forall f x y,
    map_eqv (map_mapRange f (singleton_map x y)) (singleton_map x (f y)).
  Proof.
    intros f x y a; reflexivity.
  Qed.


  Lemma map_mapRangeJoin : forall ma mb m,
    mapJoin ma mb m -> forall f,
    mapJoin (map_mapRange f ma) (map_mapRange f mb) (map_mapRange f m).
  Proof.
    intros ma mb m [J] f.
    constructor; intro a; hypReplace J (J a);
      unfold MapDef.get in * .
    assert (Xa := map_mapRangeGetEq ma a f).
    assert (Xb := map_mapRangeGetEq mb a f).
    assert (X := map_mapRangeGetEq m a f).
    destruct (map_get ma a); destruct (map_get mb a); destruct (map_get m a);
      inversion J; subst; rewrite Xa; rewrite Xb; rewrite X; constructor.
  Qed.


  (* map_mapDom lemmas *)

  Definition domMapOk (f : A -> A) m :=
    forall x y, 
      set_In x (map_dom m) -> set_In y (map_dom m) ->
      x <> y ->
      f x <> f y.


  Lemma map_mapDomGetSome : forall f m x y,
    map_get (map_mapDom f m) x = Some y -> 
    exists z, map_get m z = Some y /\ f z = x.
  Proof.
    intro f; induction m; simpl; intros x y E.
    discriminate E.
    destruct a as [a b].
    destruct (eqDec x (f a)).
    injection E; clear E; intro; subst b; exists a; destruct (eqDec a a).
    intuition.
    elf; auto.
    
    destruct (IHm _ _ E) as [z [Zget Feq]]; exists z.
    subst x.
    constructor.
    2:reflexivity.
    destruct (eqDec z a).
    subst z.
    elf; auto.
    assumption.
  Qed.


  Lemma map_mapDomGetNone : forall f m x,
    map_get (map_mapDom f m) x = None -> 
    forall y, f y = x -> map_get m y = None.
  Proof.
    intro f; induction m; simpl; intros x E y E'.
    reflexivity.
    destruct a as [a b]; subst x.
    destruct (eqDec y a); destruct (eqDec (f y) (f a)).
    assumption.
    subst a; elf; apply n; reflexivity.
    discriminate E.
    apply (IHm (f y)); auto.
  Qed.


  Lemma map_mapDomGetEq : forall f m x,
    match map_get (map_mapDom f m) x with
      | Some z => exists y, f y = x /\ map_get m y = Some z
      | None => forall y, f y = x -> map_get m y = None
    end.
  Proof.
    intros f m x.
    assert (X := map_mapDomGetSome f m x).
    assert (Y := map_mapDomGetNone f m (x := x)).
    destruct (map_get (map_mapDom f m) x).

    destruct (X b); clear X Y.
    reflexivity.
    exists x0.
    intuition.

    intros y E; apply Y; auto.

  Qed.


  Lemma domMapOkEqv : forall f m m',
    domMapOk f m ->
    map_eqv m m' ->
    domMapOk f m'.
  Proof.
    intros f m m' DMO E x y Ix Iy NE.
    apply DMO.
    rewrite (mapDomEqv E); assumption.
    rewrite (mapDomEqv E); assumption.
    assumption.
  Qed.


  Lemma map_mapDomEqv : forall f m,
    domMapOk f m -> forall m',
    map_eqv m m' -> 
    map_eqv (map_mapDom f m) (map_mapDom f m').
  Proof.
    unfold MapDef.map; unfold map_eqv; unfold MapDef.A; 
      replace MapDef.get with map_get; [idtac | auto].
    intros f m F m' E x.
    assert (X := map_mapDomGetEq f m x).
    assert (X' := map_mapDomGetEq f m' x).
    destruct (map_get (map_mapDom f m) x).

    destruct X as [y [xEq Meq]]; rewrite <- Meq; rewrite (E y); subst x.
    destruct (map_get (map_mapDom f m') (f y)).
    destruct X' as [y' [x'Eq Meq']].
    destruct (eqDec y' y).
    subst y'; rewrite Meq'; reflexivity.
    elf; apply (F y' y).
    assert (X := map_mem_In_dom y' m'); rewrite Meq' in X;
      rewrite (mapDomEqv E); assumption.
    assert (X := map_mem_In_dom y m); rewrite Meq in X; assumption.
    assumption.
    assumption.
    auto.

    (* case: map_get is None *)
    destruct (map_get (map_mapDom f m') x); [idtac | reflexivity].
    destruct X' as [y [xEq Meq]]; hypReplace X (X _ xEq); clear xEq.
    rewrite E in X; rewrite X in Meq; discriminate Meq.

  Qed.


  Lemma map_mapDomSingleton : forall f x y,
    map_eqv (map_mapDom f (singleton_map x y)) (singleton_map (f x) y).
  Proof.
    intros f x y a; reflexivity.
  Qed.


  Lemma domMapOkSub : forall f m,
    domMapOk f m -> forall m',
    set_subset (map_dom m') (map_dom m) ->
    domMapOk f m'.
  Proof.
    intros f m F m' Sub x y Ix Iy NE.
    apply F; try (apply Sub); assumption.
  Qed.


  Lemma mapJoinDomSubL : forall ma mb m,
    mapJoin ma mb m ->
    set_subset (map_dom ma) (map_dom m).
  Proof.
    intros ma mb m [J] x Ix.
    hypReplace J (J x); unfold MapDef.get in J.
    assert (X := map_mem_In_dom x ma).
    assert (Y := map_mem_In_dom x m).
    destruct (map_get ma x); inversion J; clear J.
    subst; rewrite <- H2 in Y; assumption.
    rewrite <- H0 in Y; elf; auto.
    rewrite <- H0 in Y; elf; auto.
  Qed.


  Lemma mapJoinDomSubR : forall ma mb m,
    mapJoin ma mb m ->
    set_subset (map_dom mb) (map_dom m).
  Proof.
    intros ma mb m [J] x Ix.
    hypReplace J (J x); unfold MapDef.get in J.
    assert (X := map_mem_In_dom x mb).
    assert (Y := map_mem_In_dom x m).
    destruct (map_get mb x); inversion J; clear J.
    subst; rewrite <- H2 in Y; assumption.
    rewrite <- H in Y; elf; auto.
    rewrite <- H in Y; elf; auto.
  Qed.


  Lemma map_mapDomJoin : forall ma mb m,
    mapJoin ma mb m -> forall f,
      domMapOk f m ->
      mapJoin (map_mapDom f ma) (map_mapDom f mb) (map_mapDom f m).
  Proof.
    intros ma mb m J f F.
    assert (Subb := mapJoinDomSubR J).
    destruct J as [J].
    constructor; intro x;
      unfold MapDef.get in * ; unfold MapDef.map in * ; unfold MapDef.A in * .
    assert (Ea := map_mapDomGetEq f ma x).
    assert (Eb := map_mapDomGetEq f mb x).
    assert (E := map_mapDomGetEq f m x).

    destruct (map_get (map_mapDom f ma) x).

    destruct Ea as [y [xEq Ea]]; hypReplace J (J y); rewrite Ea in J; clear Ea.
    inversion J; clear J; subst v.
    destruct (map_get (map_mapDom f m) x).
    destruct E as [y' [xEq' E]].
    destruct (eqDec y y').

    subst y' x; clear xEq'.
    rewrite E in H2; injection H2; clear H2; intro; subst b0.
    destruct (map_get (map_mapDom f mb) (f y)).
    destruct Eb as [y' [Feq Meq]].
    destruct (eqDec y y').
    subst y; rewrite Meq in H; discriminate H.
    elf; apply (F y y'); auto.
    assert (X := map_mem_In_dom y m); rewrite E in X; assumption.
    apply Subb; assert (X := map_mem_In_dom y' mb); rewrite Meq in X; assumption.
    constructor.

    subst; elf; apply (F y y'); auto.
    assert (X := map_mem_In_dom y m); rewrite <- H2 in X; assumption.
    assert (X := map_mem_In_dom y' m); rewrite E in X; assumption.

    rewrite (E _ xEq) in H2; discriminate H2.

    destruct (map_get (map_mapDom f mb) x).
    destruct Eb as [y [Xeq Eb]]; hypReplace J (J y).
    rewrite (Ea _ Xeq) in J; rewrite Eb in J; clear Ea Eb.
    inversion J; clear J; subst v.
    destruct (map_get (map_mapDom f m) x).
    destruct E as [y' [Xeq' E]].
    destruct (eqDec y y').
    subst x y'.
    rewrite <- H0 in E; injection E; clear E; intro; subst b0; constructor.
    elf; apply (F y y'); subst x; auto.
    assert (X := map_mem_In_dom y m); rewrite <- H0 in X; assumption.
    assert (X := map_mem_In_dom y' m); rewrite E in X; assumption.
    rewrite (E _ Xeq) in H0; discriminate H0.

    destruct (map_get (map_mapDom f m) x).
    destruct E as [y [xeq E]].
    hypReplace J (J y); rewrite (Ea _ xeq) in J; rewrite (Eb _ xeq) in J;
      inversion J.
    rewrite E in H; discriminate H.
    constructor.
  Qed.


End ListMap.
