Require Import OptUtil.

Set Implicit Arguments.


Module Type MapDef.
  Parameter A : Set.
  Parameter B : Set.
  Parameter map : Set.
  Parameter get : map -> A -> option B.
  Hypothesis Aeq_dec : forall x y:A, {x = y} + {x <> y}.
End MapDef.


Module M (Mdef : MapDef).
  
  Import Mdef.

  Definition inDom (w:A) (m:map) := exists v, get m w = Some v.

  Definition map_eqv (m1 m2:map) : Prop :=
    (forall a, get m1 a = get m2 a).

  Lemma map_eqvRefl : forall m,
    map_eqv m m.
  Proof.
    unfold map_eqv.
    reflexivity.
  Qed.

  Lemma map_eqvSymm : forall m1 m2,
    map_eqv m1 m2 -> map_eqv m2 m1.
  Proof.
    unfold map_eqv.
    intuition.
  Qed.

  Lemma map_eqvTrans : forall m1 m2 m3,
    map_eqv m1 m2 -> map_eqv m2 m3 ->
    map_eqv m1 m3.
  Proof.
    unfold map_eqv.
    intros m1 m2 m3 E1 E2 a.
    rewrite E1.
    auto.
  Qed.

  Section DomProp.

    Variable w:A.
    Variable m:map.

    Lemma notInDomGetNone : ~inDom w m -> get m w = None.
    Proof.
      unfold inDom.
      intros.
      destruct (get m w).
      elimtype False.
      apply H.
      exists b.
      reflexivity.
      reflexivity.
    Qed.

    Lemma inDomGetSome : inDom w m -> exists v, get m w = Some v.
    Proof.
      auto.
    Qed.
    
    Lemma getNoneNotInDom : get m w = None -> ~inDom w m.
    Proof.
      intro G.
      intro I; destruct I as [v Eq].
      rewrite G in Eq.
      discriminate Eq.
    Qed.


    (* bad name *)
    Lemma domDisjoint : forall w',
      inDom w m -> ~inDom w' m -> w <> w'.
    Proof.
      unfold inDom.
      intros.
      destruct (Aeq_dec w w').
      subst.
      auto.
      auto.
    Qed.

  End DomProp.


  Inductive elemJoin : option B -> option B -> option B -> Prop :=
    | noneJoin : elemJoin None None None
    | leftJoin : forall v, elemJoin (Some v) None (Some v)
    | rightJoin : forall v, elemJoin None (Some v) (Some v).

  Inductive mapJoin (m1 m2 m : map) : Prop :=
    mji : (forall k, elemJoin (get m1 k) (get m2 k) (get m k)) ->
    mapJoin m1 m2 m.

  Lemma reflJoinRight : forall v,
    elemJoin None v v.
  Proof.
    destruct v.
    constructor.
    constructor.
  Qed.

  Lemma reflJoinLeft : forall v,
    elemJoin v None v.
  Proof.
    destruct v.
    constructor.
    constructor.
  Qed.

  Lemma mapJoinEqv : forall ma ma' mb mb' m m',
    mapJoin ma mb m ->
    map_eqv ma ma' -> map_eqv mb mb' -> map_eqv m m' ->
    mapJoin ma' mb' m'.
  Proof.
    intros ma ma' mb mb' m m' [J] Ea Eb E.
    constructor.
    intros k; assert (Jk := J k); clear J.
    destruct (Ea k); destruct (Eb k); destruct (E k).
    assumption.
  Qed.


  Lemma mapJoinJoinEqvR : forall ma mb m,
    mapJoin ma mb m -> forall mb',
    mapJoin ma mb' m ->
    map_eqv mb mb'.
  Proof.
    intros ma mb m [J] mb' [J'].
    intro a.
    assert (J2 := J a); clear J.
    assert (J2' := J' a); clear J'.
    inversion J2; clear J2;
      rewrite <- H0 in J2'; rewrite <- H2 in J2'; clear H0 H H2;
	inversion J2'; reflexivity.
  Qed.


  Lemma mapJoinJoinEqvRR : forall ma mb m,
    mapJoin ma mb m -> forall m',
    mapJoin ma mb m' ->
    map_eqv m m'.
  Proof.    
    intros ma mb m [J] m' [J']; intro a.
    assert (J2 := J a); assert (J2' := J' a); clear J J'.
    inversion J2; clear J2;
      rewrite <- H0 in J2'; rewrite <- H in J2'; clear H0 H H2;
	inversion J2'; reflexivity.
  Qed.


  Section MapJoinProps.
    
    Variable m1 : map.
    Variable m2 : map.
    Variable m3 : map.
    Variable J : mapJoin m1 m2 m3.

    Lemma mapJoinComm : mapJoin m2 m1 m3.
    Proof.
      destruct J as [F].
      constructor.
      intro k.
      destruct (F k); clear F.
      constructor.
      constructor.
      constructor.
    Qed.

    (* use dom stuff in proposition? *)
    Lemma mapJoinSoundNone : forall w,
      get m1 w = None -> get m2 w = None ->
      get m3 w = None.
    Proof.
      intros w N1 N2.
      destruct J as [J'].
      destruct (J' w); clear J'.
      reflexivity.
      assumption.
      assumption.
    Qed.


    Lemma mapJoinSoundLeft : forall w v,
      get m1 w = Some v -> get m3 w = Some v.
    Proof.
      destruct J.
      intros.
      assert (Jw := H w).
      rewrite H0 in Jw.
      inversion_clear Jw.
      reflexivity.
    Qed.

    Lemma joinedDomLargerL : forall w,
      inDom w m1 -> inDom w m3.
    Proof.
      intros w [v G].
      exists v.
      apply mapJoinSoundLeft.
      assumption.
    Qed.

    Lemma joinedMapsDisjointL : forall w,
      inDom w m1 -> ~inDom w m2.
    Proof.
      intros w G.
      apply getNoneNotInDom.
      unfold inDom in G.
      destruct J as [J'].
      destruct (J' w).
      reflexivity.
      reflexivity.
      destruct G as [v' G'].
      discriminate G'.
    Qed.

    Lemma optPredJoinLeft : forall w p,
      optPred (get m1 w) p -> optPred (get m3 w) p.
    Proof.
      intros w p G.
      assert (Z := mapJoinSoundLeft (w:=w)).
      destruct (get m1 w).
      assert (R : Some b = Some b).
      reflexivity.
      rewrite (Z _ R); assumption.
      contradiction.
    Qed.

  End MapJoinProps.

  Definition combineSemantics combine :=
    forall m1 m2 w,
      get (combine m1 m2) w =
      match get m1 w with
	| Some v => Some v
	| None => get m2 w
      end.

  Module Type CombineDef.
    Variable mapCombine : map -> map -> map.
    Variable mapCombineOk : combineSemantics mapCombine.
  End CombineDef.

  Module Combine (CD : CombineDef).
    Import CD.

    Theorem mapJoinAssoc : forall m1 m2 m3 m1a m1b,
      mapJoin m1a m1b m1 -> mapJoin m1 m2 m3 ->
      exists m1b2, mapJoin m1a m1b2 m3 /\ mapJoin m1b m2 m1b2.
    Proof.
      intros m1 m2 m3 m1a m1b J1 J.
      exists (mapCombine m1b m2).
      destruct J as [J'].
      destruct J1 as [J1].
      constructor.
      
      constructor.
      intro k.
      rewrite (mapCombineOk m1b m2 k).
      assert (Jk := J' k); clear J'.
      assert (J1k := J1 k); clear J1.
      destruct J1k.
      assumption.
      assumption.
      inversion_clear Jk.
      constructor.

      constructor.
      intro k.
      rewrite (mapCombineOk m1b m2 k).
      assert (Jk := J' k); clear J'.
      assert (J1k := J1 k); clear J1.
      destruct (get m1b k).
      inversion J1k.
      rewrite <- H2 in Jk.
      inversion_clear Jk.
      constructor.
      
      destruct (get m2 k).
      constructor.
      constructor.

    Qed.

  End Combine.


  Section SymProps.

    Variable m1 : map.
    Variable m2 : map.
    Variable m3 : map.
    Variable J : mapJoin m1 m2 m3.
    Notation J' := (mapJoinComm J).

    Lemma mapJoinSoundRight : forall w v,
      get m2 w = Some v -> get m3 w = Some v.
    Proof.
      intros.
      eapply mapJoinSoundLeft.
      apply J'.
      assumption.
    Qed.

    Lemma joinedDomLargerR : forall w,
      inDom w m2 -> inDom w m3.
    Proof.
      intros.
      eapply joinedDomLargerL.
      apply J'.
      assumption.
    Qed.

    Lemma joinedMapsDisjointR : forall w,
      inDom w m2 -> ~inDom w m1.
    Proof.
      intros.
      eapply joinedMapsDisjointL.
      apply J'.
      apply H.
    Qed.

  End SymProps.

  
  Lemma joinCoversDom : forall m1 m2 m3,
    mapJoin m1 m2 m3 -> forall x,
    get m3 x <> None ->
    get m1 x <> None \/ get m2 x <> None.
  Proof.
    intros m1 m2 m3 [J] x G.
    assert (Jk := J x); clear J.
    destruct (get m3 x).
    clear G; inversion Jk.
    left; discriminate.
    right; discriminate.
    elimtype False; auto.
  Qed.


  Module Type EmptyMapDef.
    Variable empty_map : map.
    Hypothesis getEmptyNone : forall w, get empty_map w = None.
  End EmptyMapDef.

  Module JoinUnit (ED : EmptyMapDef).
    Import ED.

    Theorem joinUnitIntro : forall m,
      mapJoin empty_map m m.
    Proof.
      intro m.
      constructor.
      intro k.
      rewrite getEmptyNone.
      destruct (get m k).
      constructor.
      constructor.
    Qed.

    Theorem joinUnitLeft : forall m m',
      mapJoin empty_map m m' -> 
      map_eqv m m'.
    Proof.
      destruct 1.
      intro k.
      assert (Zk := getEmptyNone k).
      destruct (H k).
      reflexivity.
      discriminate Zk.
      reflexivity.
    Qed.

    Theorem joinUnitRight : forall m m',
      mapJoin m empty_map m' ->
      map_eqv m m'.
    Proof.
      destruct 1.
      intro k.
      assert (Zk := getEmptyNone k).
      destruct (H k).
      reflexivity.
      reflexivity.
      discriminate Zk.
    Qed.

  End JoinUnit.


  (* this is pretty incomplete at the moment *)

  Module Type MapAddDef.
    Variable add : map -> A -> B -> map.
    Variable getAddEq : forall m a b,
      get (add m a b) a = Some b.
    Variable getAddNeq : forall m a a' b,
      a <> a' ->
      get (add m a b) a' = get m a'.
  End MapAddDef.

  Module MapAdd (MAD : MapAddDef).
    Import MAD.

    (* can define singletons in here, too.  well, with empty added. *)

    Lemma map_addJoin : forall a v v' ma mb m,
      get ma a = Some v ->
      mapJoin ma mb m ->
      mapJoin (add ma a v') mb (add m a v').
    Proof.
      intros a v v' ma mb m Eq J.
      
      constructor.
      intro k.
      destruct (Aeq_dec k a).
      subst.
      rewrite getAddEq; rewrite getAddEq. 
      replace (get mb a) with (None (A := B)).
      constructor.
      assert (X := joinedMapsDisjointL J (w:=a)).
      unfold inDom in X.
      rewrite Eq in X.
      destruct (get mb a).
      elimtype False.
      apply X.
      econstructor; reflexivity.
      econstructor; reflexivity.
      reflexivity.
      rewrite getAddNeq.
      rewrite getAddNeq.
      destruct J as [J]; apply J.
      auto.
      auto.
    Qed.

    Lemma map_eqvAdd : forall m1 m2 x v,
      map_eqv m1 m2 ->
      map_eqv (add m1 x v) (add m2 x v).
    Proof.
      unfold map_eqv.
      intros m1 m2 x v H.
      intro a.
      destruct (Aeq_dec a x).
      subst; rewrite getAddEq; rewrite getAddEq; reflexivity.
      rewrite getAddNeq; [rewrite getAddNeq; auto | auto].
    Qed.


  End MapAdd.


End M.
