Require CanonListSet.
Require POnat.
Require Import NatUtil.
Require Import Arith.


Set Implicit Arguments.


Ltac hypReplace H y :=
  let HT := fresh in
    (assert (HT := y); clear H; rename HT into H).


Module NatListSetM := CanonListSet.M POnat.POnat.

Export NatListSetM.

Implicit Arguments set_map_elim [a f s].


Infix "|_|" := set_union (at level 60, right associativity).

Ltac sui1 := apply set_union_intro1.
Ltac sui2 := apply set_union_intro2.


(* {k | 0 <= k < y } *)
Fixpoint setZeroRange (y:nat) : set :=
  match y with
    | O => empty_set
    | S n' => set_add n' (setZeroRange n')
  end.


Lemma setZeroRangeIntro : forall a b,
  a < b ->
  set_In a (setZeroRange b).
Proof.
  induction b.
  intro; elimtype False; omega.
  simpl.
  intro LE.
  inversion_clear LE.
  auto.
  right; auto.
Qed.


Lemma setZeroRangeElim : forall a b,
  set_In a (setZeroRange b) ->
  a < b.
Proof.
  induction b.
  simpl; intuition.
  simpl.
  destruct 1.
  subst; auto.
  assert (Q := IHb H).
  auto.
Qed.


Lemma zeroInSetZeroRange : forall n,
  setZeroRange n = empty_set \/ set_In 0 (setZeroRange n).
Proof.
  destruct n.
  left; reflexivity.
  right.
  apply setZeroRangeIntro.
  omega.
Qed.


(* { k | x <= k < y } *)
Definition setRange (x y:nat) : set :=
  set_diff (setZeroRange y) (setZeroRange x).


Lemma setRangeIntro : forall a c b,
  b <= a < c ->
  set_In a (setRange b c).
Proof.
  unfold setRange.
  intros a c b [LE LT].
  apply set_diff_intro.
  apply setZeroRangeIntro; assumption.
  intro.
  assert (D := setZeroRangeElim _ _ H).
  omega.
Qed.


Lemma setRangeElim : forall a c b,
  set_In a (setRange b c) ->
  b <= a < c.
Proof.
  unfold setRange.
  intros a c b I.
  destruct (set_diff_elim I) as [H1 H2]; clear I.
  constructor.
  assert (Q : ~a < b).
  intro; apply H2.
  apply setZeroRangeIntro.
  assumption.
  omega.
  apply setZeroRangeElim; auto.
Qed.
Implicit Arguments setRangeElim [a c b].


Lemma setRangeAddMax : forall m n,
  m <= n ->
  set_eqv (setRange m (S n)) (set_add n (setRange m n)).
Proof.
  intros.
  unfold set_eqv.
  intro a.
  compare a n.
  intros; subst.
  constructor; intro.
  left; reflexivity.
  apply setRangeIntro.
  omega.
  intro Neq; constructor; intro.
  right.
  apply setRangeIntro.
  assert (Q := setRangeElim H0).
  unfold POnat.POnat.A in a.
  assert (a < n \/ a = n).
  omega.
  destruct H1.
  omega.
  intuition.
  destruct H0.
  subst; intuition.
  apply setRangeIntro.
  assert (Q := setRangeElim H0).
  omega.
Qed.


Lemma set_eqvEmptySetRange : forall x,
  set_eqv empty_set (setRange x x).
Proof.
  intros x z; constructor; intro I.
  elimtype False; simpl in I; assumption.
  simpl.
  hypReplace I (setRangeElim I).
  omega.
Qed.


(** sets of addresses *)

(** {regMin + 4 * k | 0 <= k < (regMax - regMin) / 4 } *)
Definition addrRange (regMin regMax: nat) :=
  set_map (fun k => regMin + 4 * k)
    (setZeroRange (div4 (regMax - regMin))).


Lemma addrRangeIntro : forall a x y k,
  4 + a <= y -> a = x + 4 * k ->
  set_In a (addrRange x y).
Proof.
  intros.
  subst.
  unfold addrRange.
  eapply set_map_intro.
  reflexivity.
  apply setZeroRangeIntro.
  assert (4 * k + 4 <= y - x).
  omega.
  clear H.
  generalize dependent k.
  generalize dependent (y - x); clear y x.
  apply (ind_4 (fun n => forall k, 4 * k + 4 <= n -> k < div4 n)); 
    intros; try omega.

  rewrite div4plus.

  destruct k.
  omega.
  assert (k < div4 n).
  apply H.
  omega.
  omega.
Qed.


Lemma addrRangeElim : forall a x y,
  set_In a (addrRange x y) ->
  4 + a <= y /\ exists k, a = x + 4 * k.
Proof.
  unfold POnat.POnat.A.
  intros a x y I.
  hypReplace I (set_map_elim I).
  destruct I as [b [bEq I]].
  subst.
  constructor.
  hypReplace I (setZeroRangeElim _ _ I).
  assert (4 * b + 4 <= y - x).
  2:omega.
  generalize dependent b.
  generalize dependent (y - x); clear x y.
  apply (ind_4 (fun n => forall b,  b < div4 n -> 4 * b + 4 <= n));
    try (unfold div4; simpl; intros b H; omega).
  intros a IH b.
  rewrite div4plus.
  intros H.
  destruct b.
  simpl; omega.
  assert (4 * b + 4 <= a).
  apply IH.
  omega.
  omega.
  exists b.
  omega.
Qed.

Implicit Arguments addrRangeElim [a x y].


Lemma addrRangeRemoveMin : forall m n,
  set_eqv (set_remove m (addrRange m n)) (addrRange (m + 4) n).
Proof.
  split; intro Ha. 
  destruct (set_remove_elim _ _ _ Ha).
  decompose [and ex] (addrRangeElim H0).
  assert (x <> 0).
  intro e.
  subst x.
  subst a.
  apply H.
  auto.
  apply addrRangeIntro with (k := x - 1);
    omega.
  decompose [and ex] (addrRangeElim Ha).
  apply set_remove_intro.
  apply addrRangeIntro with (k := x + 1);
    omega.
  omega.
Qed.


Lemma addrRangeRemoveMax : forall x y,
  aligned4 x y ->
   set_eqv (set_remove (y - 4) (addrRange x y)) (addrRange x (y - 4)).
Proof.
  intros x y [k E] z.
  constructor; intro I'.
  destruct (set_remove_elim _ _ _ I') as [I1 I2]; clear I'.
  destruct (addrRangeElim I2) as [I3 [k' I4]]; clear I2.
  compare x y.
  intro; elimtype False; omega.
  intro Ne.
  eapply addrRangeIntro.
  unfold POnat.POnat.A in * |- *; omega.
  apply I4.

  destruct (addrRangeElim I') as [I2 [n I3]]; clear I'.
  apply set_remove_intro.
  2:omega.
  subst.
  eapply addrRangeIntro.
  omega.
  reflexivity.

Qed.


Lemma set_eqvEmptyAddrRange : forall x,
  set_eqv empty_set (addrRange x x).
Proof.
  intro x; constructor; intro X; elimtype False.
  apply X.
  destruct (addrRangeElim X) as [H [k H']]; clear X.
  omega.
Qed.


Lemma minInAddrRange : forall x y,
  set_eqv (addrRange x y) empty_set \/ set_In x (addrRange x y).
Proof.
  intros x y.
  destruct (set_In_dec x (addrRange x y)).
  right; assumption.
  left.
  intro z; constructor; intro I.
  simpl.
  destruct (addrRangeElim I) as [R1 [k R2]]; clear I.
  apply n.
  subst.
  eapply addrRangeIntro.
  omega.
  instantiate (1 := 0).
  omega.
  simpl in I; elimtype False; assumption.
Qed.


Lemma addrAddrSetDisj : forall x y x' y',
  ptrP x -> ptrP x' ->
  ptrP y -> ptrP y' ->
  set_disj (addrRange x y) (addrRange x' y') ->
  set_disj (addrRange x y) (setRange x' y').
Proof.
  intros x y x' y' Px Px' Py Py' D.
  destruct (minInAddrRange x y) as [Ex | Ix].

  setoid_rewrite Ex.
  apply set_disj_intro; simpl; intros z X _; assumption.
  
  destruct (set_In_dec x' (addrRange x' y')) as [Ix' | Ix'].

  destruct (ptrP_elim Px) as [k E]; clear Px.
  destruct (ptrP_elim Px') as [k' E']; clear Px'.

  assert (IE : (x < x' /\ y < x' + 4) \/ (y' < x + 4 /\ y' < y)).
  assert (DJ : x < x' \/ x' <= x).
  omega.
  destruct DJ.

  left.
  constructor.
  assumption.

  assert (LTE : k <= k').
  omega.
  destruct (LTeq LTE) as [k'' R]; clear LTE H.
  destruct (addrRangeElim Ix) as [E1 [k2 E2]]; clear Ix.
  destruct (set_In_dec x' (addrRange x y)) as [X' | X'].
  elimtype False; exact (set_disj_elim D _ X' Ix').
  clear D.
  assert (X'' : ~(4 + x' <= y)).
  intro; apply X'.
  eapply addrRangeIntro.
  assumption.
  instantiate (1 := k'').
  omega.
  omega.

  right.
  destruct (addrRangeElim Ix) as [Z [k3 Z']].
  destruct (LTeq H) as [k'' R]; clear H.
  subst.
  destruct (ptrPlusPtr R) as [q Qeq].
  assert (DJ : y < y' \/ y' <= y).
  omega.
  destruct DJ as [Lt | Le].
  elimtype False.
  apply (set_disj_elim D _ Ix); clear D.
  eapply addrRangeIntro.
  omega.
  subst.
  rewrite <- R.
  reflexivity.

  destruct (set_In_dec (4 * k) (addrRange (4 * k') y')) as [X | X].
  elimtype False; exact (set_disj_elim D _ Ix X).
  clear D.
  assert (X' : ~(4 + (4 * k) <= y')).
  intro; apply X.
  eapply addrRangeIntro.
  assumption.
  instantiate (1 := q).
  omega.
  omega.
  
  apply set_disj_intro; intros z I1 I2.
  destruct (addrRangeElim I1) as [I3 [k4 I4]]; clear I1.
  hypReplace I2 (setRangeElim I2).
  clear Ix Ix' D.
  destruct IE as [[Lt1 Lt2] | [Lt1 Lt2]].
  omega.
  unfold POnat.POnat.A in * .
  destruct I2 as [_ I2].
  destruct (ptrP_elim Py) as [ky Ey]; clear Py.
  destruct (ptrP_elim Py') as [ky' Ey']; clear Py'.
  omega.
  
  clear D.
  assert (Lt : y' < 4 + x').
  assert (Lt : ~(4 + x' <= y')).
  intro Lt; apply Ix'; clear Ix'.
  eapply addrRangeIntro.
  assumption.
  instantiate (1 := 0); auto.
  omega.

  clear Ix Ix'.
  apply set_disj_intro; intros z Iz Iz'.
  hypReplace Iz' (setRangeElim Iz').
  destruct (addrRangeElim Iz) as [_ [k I]]; clear Iz.
  destruct (ptrP_elim Px) as [kx Ex]; clear Px.
  destruct (ptrP_elim Px') as [kx' Ex']; clear Px'.
  destruct (ptrP_elim Py') as [ky' Ey']; clear Py'.
  subst.
  replace (4 * kx + 4 * k) with (4 * (kx + k)) in Iz'.
  destruct (LTeq (x:=kx') (y:=kx + k)) as [k2 E].
  omega.
  rewrite <- E in Iz'; clear E.
  clear kx k y Py.
  omega.
  omega.

Qed.



(** sets of object pointers *)

Definition rangeObjs (regMin regMax: nat) :=
  set_map (fun k => regMin + 8 * k)
    (setZeroRange (div8 (regMax - regMin))).


Lemma rangeObjsElim : forall a x y,
  set_In a (rangeObjs x y) ->
  8 + a <= y /\ exists k, a = x + 8 * k.
Proof.
  unfold POnat.POnat.A.
  unfold rangeObjs.
  intros a x y I.
  hypReplace I (set_map_elim I).
  destruct I as [b [bEq I]].
  subst.
  constructor.
  hypReplace I (setZeroRangeElim _ _ I).
  assert (8 * b + 8 <= y - x).
  2:omega.
  generalize dependent b.
  generalize dependent (y - x); clear x y.
  apply (ind_8 (fun n => forall b,  b < div8 n -> 8 * b + 8 <= n));
    try (unfold div8; simpl; intros b H; omega).
  intros a IH b.
  rewrite div8plus.
  intros H.
  destruct b.
  simpl; omega.
  assert (8 * b + 8 <= a).
  apply IH.
  omega.
  omega.
  exists b.
  omega.
Qed.

Implicit Arguments rangeObjsElim [a x y].


Lemma rangeObjsIntro : forall a x y k,
  8 + a <= y -> a = x + 8 * k ->
  set_In a (rangeObjs x y).
Proof.
  intros.
  subst.
  unfold rangeObjs.
  eapply set_map_intro.
  reflexivity.
  apply setZeroRangeIntro.
  assert (8 * k + 8 <= y - x).
  omega.
  clear H.
  generalize dependent k.
  generalize dependent (y - x); clear y x.
  apply (ind_8 (fun n => forall k, 8 * k + 8 <= n -> k < div8 n)); 
    intros; try omega.

  rewrite div8plus.

  destruct k.
  omega.
  assert (k < div8 n).
  apply H.
  omega.
  omega.
Qed.


(* set_In elimination tactic *)
Ltac set_InElim :=
  match goal with
    | H : set_In ?a ?s |- set_In ?a ?s => exact H
    | H : set_In ?a ?s |- set_In ?a _ =>
      (match s with
	 | set_add _ _ => destruct (set_add_elim H)
	 | set_remove _ _ => destruct (set_remove_elim _ _ _ H)
	 | set_union _ _ => destruct (set_union_elim H)
	 | set_diff _ _ => destruct (set_diff_elim H)
	 | set_insert _ _ => destruct (set_insert_elim _ _ _ H)
	 | set_canon _ => 
	   let H2 := fresh in assert (H2 := set_canon_elim _ _ H)
	 | setZeroRange _ => 
	   let H2 := fresh in assert (H2 := setZeroRangeElim _ _ H)
	 | setRange _ _ => 
	   let H2 := fresh in assert (H2 := setRangeElim H)
	 | addrRangeElim _ _ =>
	   let H2 := fresh in
	   let k := fresh "k" in
	   let H3 := fresh in
	     destruct (addrRangeElim H) as [H2 [k H3]]
	 | rangeObjs _ _ =>
	   let H2 := fresh in
	   let k := fresh "k" in
	   let H3 := fresh in
	     destruct (rangeObjsElim H) as [H2 [k H3]]
      end; clear H; try set_InElim)
  end.



Lemma rangeSplitDisj : forall x y z, 
  set_disj (rangeObjs x y) (rangeObjs y z).
Proof.
  intros x y z.
  apply set_disj_intro.
  intros a I1 I2.
  destruct (rangeObjsElim I1) as [H1 [k1 H1']]; clear I1.
  destruct (rangeObjsElim I2) as [H2 [k2 H2']]; clear I2.
  omega.
Qed.



Lemma rangeObjsTransEqv : forall x y z,
  aligned8 x y -> y <= z ->
  set_eqv (set_union (rangeObjs x y) (rangeObjs y z)) (rangeObjs x z).
Proof.
  unfold set_eqv; unfold POnat.POnat.A.
  intros x y z AL LT a.
  constructor; intro H.

  hypReplace H (set_union_elim H); destruct H.
  hypReplace H (rangeObjsElim H).
  destruct H as [H [k Eq]].
  subst.
  eapply rangeObjsIntro.
  omega.
  reflexivity.

  hypReplace H (rangeObjsElim H).
  destruct H as [H [k Eq]].
  subst.
  destruct AL as [n X].
  eapply rangeObjsIntro.
  assumption.
  rewrite X.
  assert (x + 8 * n + 8 * k = x + 8 * (n + k)).
  omega.
  eauto.

  hypReplace H (rangeObjsElim H).
  destruct H as [H [k Eq]].
  assert (a < y \/ a >= y).
  omega.
  subst.
  destruct AL as [n X].
  subst.
  destruct H0.
  apply set_union_intro1.
  eapply rangeObjsIntro.
  omega.
  reflexivity.
  apply set_union_intro2.
  eapply rangeObjsIntro.
  omega.
  assert (x + 8 * k = x + 8 * n + 8 * (k - n)).
  omega.
  eauto.
Qed.


Lemma rangeRemoveEqv : forall x y,
  set_eqv (set_remove x (rangeObjs x y)) (rangeObjs (8 + x) y).
Proof.
  intros x y.
  unfold POnat.POnat.A in x.
  unfold rangeObjs.
  intro a; unfold POnat.POnat.A in a.
  constructor; intro I.

  assert (M := set_remove_elim1 _ _ _ I).
  assert (N := set_remove_elim2 _ I).
  clear I.
  destruct (set_map_elim M) as [a' [Eq I]]; clear M.

  destruct a'.

  elimtype False; apply N.
  simpl in Eq.
  replace (x + 0) with x in Eq.
  assumption.
  auto.

  unfold POnat.POnat.A.

  apply (set_map_intro (a:=a) a').
  subst.
  assert (x + 8 * S a' = 4 + 4 + x + 8 * a').
  omega.
  assumption.

  apply setZeroRangeIntro.
  hypReplace I (setZeroRangeElim _ _ I).
  subst; clear N.
  replace (y - (8 + x)) with ((y - x) - 8).
  2:omega.
  generalize dependent (y - x).
  clear x y.
  intros.

  replace (div8 (n - 8)) with (div8 n - 1).
  omega.
  clear a' I.
  symmetry; apply div8sub.

  destruct (set_map_elim I) as [a' [Ea I']]; clear I.
  hypReplace I' (setZeroRangeElim _ _ I').
  apply set_remove_intro.
  apply (set_map_intro (a:=a) (S a')).
  unfold POnat.POnat.A in a'.
  assert (a = x + 8 * S a'); [omega | assumption].
  apply setZeroRangeIntro.
  replace (y - (8 + x)) with ((y - x) - 8) in I'.
  2:omega.
  rewrite div8sub in I'.
  omega.
  omega.
Qed.


Lemma minInObjsRange : forall x y,
  rangeObjs x y = empty_set \/ set_In x (rangeObjs x y).
Proof.
  unfold rangeObjs.
  intros.
  destruct (zeroInSetZeroRange (div8 (y - x))); [left | right].
  rewrite H; reflexivity.
  eapply set_map_intro.
  2:eauto.
  simpl.
  auto.
Qed.


Lemma set_eqvEmptyRangeObjs : forall x,
  set_eqv empty_set (rangeObjs x x).
Proof.
  intro x; constructor; intro X; elimtype False.
  apply X.
  destruct (rangeObjsElim X) as [EqX [y EqY]]; omega.
Qed.


Lemma rangeObjsEndSplit : forall x y,
  aligned8 x y -> x <> y ->
  set_eqv (rangeObjs x y) (set_add (y - 8) (rangeObjs x (y - 8))).
Proof.
  intros x y [k R] Ne; subst.
  destruct k.
  elimtype False; apply Ne; omega.
  clear Ne.
  constructor; intro H.

  destruct (rangeObjsElim H) as [H1 [n H2]]; clear H.
  subst.
  assert (N : n = k \/ n < k).
  omega.
  clear H1; destruct N.
  subst.
  apply set_add_intro1; unfold POnat.POnat.A; omega.
  apply set_add_intro2.
  eapply rangeObjsIntro.
  omega.
  reflexivity.

  destruct (set_add_elim H); clear H.
  subst.
  eapply rangeObjsIntro.
  omega.
  instantiate (1 := k).
  omega.
 
  destruct (rangeObjsElim H0) as [H [n H']]; clear H0.
  subst.
  eapply rangeObjsIntro.
  omega.
  reflexivity.

Qed.


Lemma rangeObjsCard : forall x y,
  aligned8 x y ->
  8 * set_card (rangeObjs x y) = y - x .
Proof.
  intros x y [k E].
  subst.
  replace (x + 8 * k) with (8 * k + x).
  2:omega.
  replace (8 * k + x - x) with (8 * k).
  2:omega.
  induction k.

  assert (EE := set_eqvEmptyRangeObjs x).
  replace (8 * 0 + x) with x.
  rewrite <- (set_cardMorph EE).
  reflexivity.
  reflexivity.

  assert (aligned8 x (8 * S k + x)).
  exists (S k); omega.
  assert (Ne : x <> 8 * S k + x).
  omega.
  assert (EE := rangeObjsEndSplit H Ne); clear H Ne.
  rewrite (set_cardMorph EE); clear EE.
  rewrite <- set_cardInsertAddEq.
  rewrite set_insertCard2.
  replace (8 * S k + x - 8) with (8 * k + x).
  omega.
  omega.
  clear IHk; intros H.
  destruct (rangeObjsElim H) as [H1 [n H2]]; clear H.
  omega.

Qed.


Lemma rangeObjsEndRemove : forall x y,
  aligned8 x y -> x <> y ->
   set_eqv (set_remove (y - 8) (rangeObjs x y)) (rangeObjs x (y - 8)).
Proof.
  intros x y [k E] Ne z.
  destruct k.
  elimtype False; omega.

  assert (I : set_In (y-8) (rangeObjs x y)).
  eapply rangeObjsIntro.
  omega.
  instantiate (1 := k).
  omega.

  constructor; intro I'.
  destruct (set_remove_elim _ _ _ I') as [I1 I2]; clear I'.
  destruct (rangeObjsElim I2) as [I3 [k' I4]]; clear I2.
  eapply rangeObjsIntro.
  subst.
  replace (x + 8 * S k - 8) with (x + 8 * k) in I1.
  replace (8 + (x + 8 * k')) with (x + 8 * S k') in I3.
  assert (LT : k' < k).
  assert (LTE : k' <= k).
  omega.
  compare k' k; intro R.
  subst.
  elimtype False; apply I1; reflexivity.
  omega.
  omega.
  omega.
  omega.
  apply I4.
  
  destruct (rangeObjsElim I') as [I2 [n I3]]; clear I'.
  apply set_remove_intro.
  2:omega.
  subst.
  eapply rangeObjsIntro.
  omega.
  reflexivity.

Qed.


Lemma rangeObjsStartAdd : forall x y,
  aligned8 (8 + x) y ->
  set_eqv (set_add x (rangeObjs (8 + x) y)) (rangeObjs x y).
Proof.
  intros x y [k0 Al]; constructor; intro I; set_InElim.

  subst a.
  eapply rangeObjsIntro.
  omega.
  instantiate (1 := 0); auto.

  eapply rangeObjsIntro.
  assumption.
  instantiate (1 := S k).
  omega.

  destruct k; [left | right].

  unfold POnat.POnat.A; omega.
 
  eapply rangeObjsIntro.
  assumption.
  instantiate (1 := k).
  omega.

Qed.
Implicit Arguments rangeObjsStartAdd [x y].


Lemma rangeObjsEndAdd : forall x y,
  aligned8 x y ->
  set_eqv (set_add y (rangeObjs x y)) (rangeObjs x (8 + y)).
Proof.
  intros x y [k E]; constructor; intro I.

  destruct (set_add_elim I) as [I' | I']; clear I.
  subst a; eapply rangeObjsIntro; eauto.
  destruct (rangeObjsElim I') as [H [k' H']]; clear I'.
  subst a; eapply rangeObjsIntro; eauto; omega.
  
  destruct (rangeObjsElim I) as [H [k' H']]; clear I.
  compare a y; intro X; [apply set_add_intro1 | apply set_add_intro2].
  assumption.

  (* omega apparently isn't smart enough, so we have to help it along *)
  assert (LE : a <= y).
  omega.
  destruct (le_lt_or_eq _ _ LE); clear LE.
  eapply rangeObjsIntro.
  omega.
  eauto.
  elimtype False; auto.

Qed.


Lemma addrRangeSetRangeSubset : forall x y,
  set_subset (addrRange x y) (setRange x y).
Proof.
  intros x y a I.
  destruct (addrRangeElim I) as [I1 [k I2]]; clear I.
  subst.
  apply setRangeIntro.
  omega.
Qed.


Lemma objRangeAddrSetSubset : forall x y,
  set_subset (rangeObjs x y) (addrRange x y).
Proof.
  intros x y z I.
  destruct (rangeObjsElim I) as [A [k B]].
  subst.
  eapply addrRangeIntro.
  omega.
  instantiate (1 := 2 * k).
  omega.
Qed.


Lemma objRangeSetRangeSubset : forall x y,
  set_subset (rangeObjs x y) (setRange x y).
Proof.
  intros x y a I.
  apply addrRangeSetRangeSubset.
  apply objRangeAddrSetSubset; assumption.
Qed.


Lemma rangeObjsNegElim1 : forall a x y,
  ~set_In a (rangeObjs x y) ->
  aligned8 x y ->
  (exists k', y = a + 8 * (S k')) ->
  a < x.
Proof.
  intros a x y NI [k al] [k' al'].
  assert (D : a < x \/ a >= x).
  omega.
  destruct D.
  assumption.
  elf; unfold POnat.POnat.A in * .
  elf; apply NI; clear NI; eapply rangeObjsIntro.
  omega.
  instantiate (1 := k - S k').
  omega.
Qed.
Implicit Arguments rangeObjsNegElim1 [a x y].


Lemma rangeObjsNegElim2 : forall a x y,
  ~set_In a (rangeObjs x y) ->
  aligned8 x y ->
  aligned8 x a ->
  a >= y.
Proof.
  intros a x y NI [k al] [k' al'].
  assert (D : a >= y \/ a < y).
  omega.
  destruct D as [D | D].
  assumption.
  elf; unfold POnat.POnat.A in * .
  subst; apply NI; clear NI; eapply rangeObjsIntro.
  omega.
  reflexivity.
Qed.
Implicit Arguments rangeObjsNegElim2 [a x y].


Lemma rangeSplit : forall s x y, 
  set_disj s (rangeObjs x y) -> 
  x <= y ->
  forall x', 
    set_eqv (set_union s (rangeObjs x y)) (rangeObjs x' y) ->
    aligned8 x' x ->
    set_eqv s (rangeObjs x' x).
Proof.
  intros s x y D Lt x' E Al a.
  constructor; intro I.
  destruct Al as [k Ek].

  assert (I' : set_In a (rangeObjs x' y)).
  apply (set_eqv_elim1 E).
  apply set_union_intro1; assumption.
  destruct (rangeObjsElim I') as [E2 [k2 E2']]; clear I'.
  subst.

  assert (k2 < k).
  assert (X := set_disj_elim D (x' + 8 * k2) I).
  assert (~(k <= k2)).
  intro Le.
  apply X.
  eapply rangeObjsIntro.
  omega.
  instantiate (1 := k2 - k).
  omega.
  omega.
  
  eapply rangeObjsIntro.
  omega.
  reflexivity.

  assert (H : x <= y \/ y < x).
  omega.
  destruct H.
  
  assert (E' := rangeObjsTransEqv Al H).
  setoid_replace s with (rangeObjs x' x).
  assumption.
  clear a I.
  apply set_eqvSymm.
  eapply (unionDisjEqv (s1 := rangeObjs x' x) (s1' := s) (s2 := rangeObjs x y)).
  setoid_rewrite E'.
  setoid_rewrite E.
  apply set_eqvRefl.
  apply rangeSplitDisj.
  assumption.
  
  elimtype False; omega.

Qed.


Lemma rangeSplit' : forall s x y, 
  set_disj (rangeObjs x y) s -> forall z,
    set_eqv (rangeObjs x y |_| s) (rangeObjs x z) ->
    aligned8 x y ->
    set_eqv s (rangeObjs y z).
Proof.
  intros s x y D z E Axy; constructor; intro I.

  assert (I' : set_In a (rangeObjs x z)).
  rewrite <- E; apply set_union_intro2; assumption.
  set_InElim.
  assert (NI : ~set_In a (rangeObjs x y)).
  intro I''.
  apply (set_disj_elim D a); assumption.
  clear D.
  assert (GE : a >= y).
  eapply rangeObjsNegElim2; eauto.
  exists k; assumption.
  clear NI.
  destruct Axy; subst.
  apply rangeObjsIntro with (k - x0); omega.

  assert (I' : set_In a (rangeObjs x y |_| s)).
  rewrite E.
  set_InElim.
  subst.
  destruct Axy as [k' Axy].
  subst.
  apply rangeObjsIntro with (k' + k); omega.
  destruct (set_union_elim I') as [I'' | I'']; clear I'.
  set_InElim; elf; omega.
  assumption.
Qed.
Implicit Arguments rangeSplit' [s x y].


Add Morphism set_map : set_mapMorph.
Proof.
  exact set_map_eqv.
Qed.

