Require Import LinLog.
Require Import List.
Require Import NatUtil.

(* second idea: flattenator aka associator *)


Set Implicit Arguments.


Inductive ntree : Set :=
  | node : ntree -> ntree -> ntree
  | leaf : nat -> ntree.

Coercion leaf : nat >-> ntree. 

Definition nlist1 : nat -> list nat := fun x => x::nil.
Coercion nlist1 : nat >-> list.
Definition tlist1 : ntree -> list ntree := fun x => x::nil.
Coercion tlist1 : ntree >-> list.


Fixpoint ltree (l:list ntree) : ntree :=
  match l with
    | x::y::nil => node x y
    | x::l' => node x (ltree l')
    | nil => leaf 0
  end.


Fixpoint treeFlatten (tr:ntree) {struct tr} : list nat :=
  match tr with
    | node tr1 tr2 => treeFlatten tr1 ++ treeFlatten tr2
    | leaf x => x
  end.

Ltac assocConjG' T isLeft :=
  match eval compute in T with
    | leaf _ => idtac
    | node ?T1 ?T2 => 
      assocConjG' T1 true;
      eapply sconjImplR;
	[idtac |
	  let h := fresh "h" in
	  let Hp := fresh "Hp" in
	    (intros h Hp; assocConjG' T2 isLeft; apply Hp)];
      match eval compute in isLeft with
	| true => apply sconjAssocL
	| false => idtac
      end
  end.


Ltac assocConjG T := assocConjG' T false.


(* this won't work if there are uninstantiated existential variables *)
Ltac assocConj' Hp T isLeft :=
  match eval compute in T with
    | leaf _ => idtac
    | node ?T1 ?T2 => 
      assocConj' Hp T1 true;
      let h1 := fresh "h" in
      let h2 := fresh "h" in
      let J := fresh "J" in
      let Hp1 := fresh "H" in
      let Hp2 := fresh "H" in
	(destruct (sconjElim Hp) as [h1 [h2 [J [Hp1 Hp2]]]];
	  assocConj' Hp2 T2 isLeft;
	  hyr Hp (sconjIntro Hp1 Hp2 J);
	  clear h1 h2 J Hp1 Hp2);
      match eval compute in isLeft with
	| true => hyr Hp (sconjAssocR Hp)
	| false => idtac
      end
  end.

Ltac assocConj Hp T := assocConj' Hp T false.


Notation "[| A , B |]" := (node A B) (at level  80).

(*
Lemma TestAssocG : forall h A B C D E,
  h ||- A ** ((B ** C) ** D) ** E ->
  h ||- A ** B ** C ** D ** E.
Proof.
  intros.
  assocConjG ([|1, [|[|[|2, 3|], 4|], 5|]|]).
  assumption.
Qed.


Lemma TestAssocG2 : forall h A B C D,
  h ||- (A ** B ** C) ** D ->
  h ||- A ** B ** C ** D.
Proof.
  intros.
  assocConjG ([| [| 1, [|2, 3|] |], 4|]).
  assumption.
Qed.


Lemma TestAssoc : forall h A B C D E,
  h ||- A ** B ** C ** D ** E ->
  h ||- A ** ((B ** C) ** D) ** E.
Proof.
  intros.
  assocConj H ([|1, [|[|[|2, 3|], 4|], 5|]|]).
  assumption.
Qed.


Lemma TestAssoc2 : forall h A B C D,
  h ||- A ** B ** C ** D ->
  h ||- (A ** B ** C) ** D.
Proof.
  intros h A B C D Hp.
  assocConj Hp ([| [| 1, [|2, 3|] |], 4|]).
  assumption.
Qed.


Lemma TestAssoc3 : forall h A B C D,
  h ||- A ** B ** C ** D ->
  h ||- A ** B ** C ** D.
Proof.
  intros h A B C D Hp.
  assocConj Hp ([| 1, [|2, [|3, 4|]|]|]).
  assumption.
Qed.
*)


(* We define a tactic permConG for permuting a series of
   separating conjunctions in the goal.

   Example: 

   permConG (1::3::2::nil)

   will transform a goal of the form

   h ||- A ** B ** C

   to

   h ||- A ** C ** B

*)


(* First, we define permutations on lists of natural numbers and prove
   a few properties of these permutations.

   Second, we define a tactic to use these permutations to alter heap
   predicate proofs.

   Next, we define a duplicate-preserving list sort, and show that a
   list is a permutation of its sorted version.

   Finally, we combine the above to get permConG.



   Note: All of the lemmas in this file are ended with Defined instead
   of Qed.  This makes their definitions transparent, to allow analysis
   by tactics.

*)



(* definition of Permutation basically taken from
   Coq.Lists from the latest version *)

(* Permutation l l' holds if l is a permutation of l' *)
Inductive Permutation : list nat -> list nat -> Prop :=
  | perm_nil: Permutation nil nil
  | perm_skip: forall x l l',
    Permutation l l' -> 
    Permutation (x::l) (x::l')
  | perm_swap: forall x y l,
    Permutation (y::x::l) (x::y::l)
  | perm_trans: forall l l' l'',
    Permutation l l' -> 
    Permutation l' l'' -> 
    Permutation l l''.

Hint Constructors Permutation.


(* basic properties of Permutation *)

Lemma PermutationRefl : forall l,
  Permutation l l.
Proof.
  induction l; auto.
Defined.

Lemma PermutationSymm : forall l l',
  Permutation l l' ->
  Permutation l' l.
Proof.
  induction 1; auto.
  apply perm_trans with l'; assumption.
Defined.


Lemma noNilPerm : forall l,
  Permutation nil l ->
  l = nil.
Proof.
  assert (forall l l', Permutation l l' -> l = nil -> l' = nil).
  induction 1.
  reflexivity.
  discriminate 1.
  discriminate 1.
  auto.
  intros; eauto.
Defined.


(* Given a derivation P : Permutation l l', and pretending that the
current heap predicate goal is really l', follow P to permute the goal
to a heap predicate equivalent of l.

A Permutation describes a series of operations that can be used to
transition one list to another.  The actual contents of the list don't
actually matter: we can use the set of operations to transition
between "isomorphic lists": if l1 is isomorphic to l1', and l2 is
isomorphic to l2', and we know H : Permutation l1 l2, then we can use
the steps in H to transition from l1' to l2'.

We can think of a series of heap predicates separated by separating
conjections as a list.  For instance, we can think of A ** (B ** (C **
D)) as a list [A, B, C, D].

We combine these ideas to create a tactic that analyzes a Permutation
term describing the transition from one list of nats to another to
transition from one heap predicate to another.

To do this, we must be able to translate each list operation into a
heap predicate operation.

The first two cases involve lists that do not change, so we do not
change the predicate.

For the skip case, we want to go from a proof of h ||- A ** B to a
proof of h ||- A ** B', using a subterm of our Permutation derivation
to transition from B to B'.  The problem is that we don't know what B'
is yet, so we must use eapply, which lets us delay the choice.  A
tricky point here is that we must make sure we do not let eauto
prematurely instantiate the existential variables created by eapply.

For swap, we are either dealing with h ||- A ** B ** C or h ||- A **
B.  Each case requires its own lemma.  In the list case these are the
same, but we have no nil for our "lists" of heap predicates.

Finally, for the transitive case, we simply apply one transition, then
the other.  Because we are transitioning from the second argument of
Permutation to the first we apply the second subterm first.

*)


Ltac permConjG' P :=
  match eval compute in P with
    | perm_nil => idtac
    | perm_skip _ perm_nil => idtac
    | perm_skip _ ?P' => 
      eapply sconjImplR;
	[idtac | 
	  let h := fresh "h" in
	  let Hp := fresh "Hp" in
	    (intros h Hp; permConjG' P'; apply Hp)]
    | perm_swap _ _ _ => first [apply sconjGLift2' | apply sconjComm]
    | perm_trans ?P1 ?P2 => permConjG' P1; permConjG' P2
  end.


(* won't work with uninstantiated existential variables *)
Ltac permConj' Hp P :=
  match eval compute in P with
    | perm_nil => idtac
    | perm_skip _ perm_nil => idtac
    | perm_skip _ ?P' =>
      (let h1 := fresh "h" in
      let h2 := fresh "h" in
      let J := fresh "J" in
      let Hp1 := fresh "Hp" in
      let Hp2 := fresh "Hp" in
	(destruct (sconjElim Hp) as [h1 [h2 [J [Hp1 Hp2]]]];
	 permConj' Hp2 P';
	 hyr Hp (sconjIntro Hp1 Hp2 J);
	 clear h1 h2 J Hp1 Hp2))
    | perm_swap _ _ _ => 
      first [sconjLift2 Hp | sconjLift2of2 Hp]
    | perm_trans ?P1 ?P2 => permConj' Hp P1; permConj' Hp P2
  end.


(* Now we define a duplicate preserving list sort.  This will allow us
   to have a create a nice interface to permConjG'. *)

(* duplicate preserving ordered insertion *)
Fixpoint list_insert (a:nat) (l:list nat) {struct l} : set :=
  match l with
    | nil => (a::nil)
    | a'::l' =>
      if blt_nat a a' then
	a::l
      else
        a'::(list_insert a l')
  end.

(* duplicate preserving list sort *)
Fixpoint list_sort (l:list nat) : set := 
  match l with
    | nil => nil
    | x::l' => list_insert x (list_sort l')
  end.


Lemma PermInsert' : forall a l,
  Permutation (list_insert a l) (a :: l).
Proof.
  induction l; simpl.
  auto.
  destruct (blt_nat a a0).
  apply PermutationRefl.
  apply perm_trans with (a0::a::l); auto.
Defined.


Lemma PermInsert : forall l l',
  Permutation l l' -> forall a,
  Permutation (list_insert a l) (a :: l').
Proof.
  induction l.

  intros l' H a; replace l' with (nil (A := nat)).
  simpl; auto.
  symmetry; apply noNilPerm; assumption.

  simpl.
  intros l' H a'.
  destruct (blt_nat a' a).
  auto.

  assert (X : exists x, x = a::l).
  exists (a::l); reflexivity.

  destruct X as [x E]; rewrite <- E in H.
  generalize dependent a; generalize dependent l.
  induction H; simpl; intros.

  discriminate E.

  injection E; clear E; intros; subst a l0.
  assert (X := perm_skip x (IHl _ H a')); clear IHl.
  apply perm_trans with (x::a'::l'); auto.

  injection E; clear E; intros; subst l0 a.
  simpl.
  destruct (blt_nat a' x); clear IHl.
  apply perm_trans with (a'::y::x::l); auto.

  apply perm_trans with (x::a'::y::l); auto.
  apply perm_trans with (x::y::a'::l); auto.
  apply perm_trans with (y::x::a'::l); auto.
  constructor; constructor; apply PermInsert'.

  subst l; rename l0 into l.
  assert (X := fun H => IHPermutation1 l H a); 
    clear IHPermutation2 IHPermutation1.
  apply perm_trans with (a'::l'); auto.
Defined.
  

(* A list is a permutation of its sorted version. *)
Lemma CanonPerm : forall l,
  Permutation (list_sort l) l.
Proof.
  induction l; simpl.
  auto.
  apply PermInsert; assumption.
Defined.



(* The argument l to permConjG is treated as the desired order,
  where the current order is simply [1, 2, ...].  We use CanonPerm
  to automatically derive the Permutation relating the two, and thus
  the necessary series of operations for permConjG' to perform. *)

Ltac permConjG l := permConjG' (CanonPerm l).

Ltac permConj Hp l := permConj' Hp (CanonPerm l).


(* permx is nice for debugging permConjG, as it
   removes the stuff ignored by the tactic. *)
Inductive permx : Prop :=
  | permx_nil
  | permx_skip : permx -> permx
  | permx_swap
  | permx_trans : permx -> permx -> permx.

Fixpoint permxify (l1 l2:list nat) (P:Permutation l1 l2) {struct P} : permx :=
  match P with
    | perm_nil => permx_nil
    | perm_skip _ _ _ P' => permx_skip (permxify P')
    | perm_swap _ _ _ => permx_swap
    | perm_trans _ _ _ P' P'' =>
      permx_trans (permxify P') (permxify P'')
  end.


(*
Lemma TestPermG : forall h A B C D E F G,
  h ||- G ** F ** E ** D ** C ** A ** B ->
  h ||- A ** B ** C ** D ** E ** F ** G.
Proof.
  intros.
  permConjG (7::6::5::4::3::1::2::nil).
  assumption.
Qed.


Lemma TestPerm : forall h A B C D E F G,
  h ||- A ** B ** C ** D ** E ** F ** G ->
  h ||- G ** F ** E ** D ** C ** A ** B.
Proof.
  intros.
  permConj H (7::6::5::4::3::1::2::nil).
  assumption.
Qed.
*)


(* combining the two: *)

Ltac aspConjG T := 
  permConjG (treeFlatten T);
  assocConjG T.

Ltac aspConj Hp T := 
  permConj Hp (treeFlatten T);
  assocConj Hp T.


(*
Lemma TestAspG : forall h A B C D E F G,
  h ||- G ** F ** ((E ** D) ** C) ** A ** B ->
  h ||- A ** B ** C ** D ** E ** F ** G.
Proof.
  intros.
  aspConjG ([|7, [|6, [|[|[|5, 4|], 3|], [|1, 2|]|]|]|]).
  assumption.
Qed.


Lemma TestAsp : forall h A B C D E F G,
  h ||- A ** B ** C ** D ** E ** F ** G ->
  h ||- G ** F ** ((E ** D) ** C) ** A ** B.
Proof.
  intros.
  aspConj H ([|7, [|6, [|[|[|5, 4|], 3|], [|1, 2|]|]|]|]).
  assumption.
Qed.
*)




