Require Export LinLog.
Require Import StateTheory.
Require Import List.
Require Import WeakSCAP.
Require Import SpecUtil.
Require Export GC.


Set Implicit Arguments.


(* the abstract state view and associated operations *)


(* This specifies what the mutator and GC must agree on *)
Module Type MG_INTERFACE_DEF.

  (* root register managed by GC *)
  Parameter rroot : rnum.

  (* private register used for GC data *)
  Parameter rgcInfo : rnum.

  (* gcGuar specifies the effect of the GC on the state view *)
  Parameter gcGuar : state -> state -> list rnum -> Prop.

  (* holds on registers preserved by the GC and barriers *)
  Parameter presReg : rnum -> Prop.

  (* holds on atomic values, ie values that are not object pointers. *)
  Parameter atomP : nat -> Prop.

  (* default field value *)
  Parameter NULL : nat.

  (* upper bound on the domain of the code heap used by the GC *)
  Parameter gcPhiDom : nset.

  (* labels for the functions implementing the GC *)
  Parameter allocLbl : nat.
  Parameter read1Lbl : nat.
  Parameter read2Lbl : nat.
  Parameter write1Lbl : nat.
  Parameter write2Lbl : nat.


End MG_INTERFACE_DEF.


Coercion rfileOf : state >-> rfile.


Section GCSpecs.

  Variable repr : state -> rfile -> list rnum -> hpred.

  Variable rroot : rnum.

  Variable gcGuar : state -> state -> list rnum -> Prop.

  Variable presReg : rnum -> Prop.

  Variable NULL : nat.

  (* read barrier specification *)
  Definition readPre k s :=
    exists vs,
      s |||- repr vs s (rroot::nil) ** top /\
      vs |||- k + rget vs rroot |->_ ** top.

  Definition readGuar k s s' :=
    (forall vs A rr,
	s |||- repr vs s rr ** A ->
	In rroot rr ->
	(forall r, In r rr -> r = rroot \/ calleeSaved r) -> (* was: ~In v0 rr *)
	exists vs',
	  gcGuar vs vs' rr /\
	  s' |||- repr vs' s' (v0::rr) ** A /\
	  vs' |||- k + rget vs' rroot |-> rget vs' v0 ** top) /\
    regPres (fun r => presReg r /\ r <> v0) s s'.

(* this fits more with what i've been thinking of it as
  Definition readGuar' k s s' :=
    (forall vs A rr,
	s |||- repr vs s rr ** A ->
	In rroot rr ->
	(forall r, In r rr -> r = rroot \/ calleeSaved r) -> (* was: ~In v0 rr *)
	exists vs', exists x,
	  gcGuar vs vs' rr /\
	  s' |||- repr (rset vs' v0 x) s' (v0::rr) ** A /\
	  vs' |||- k + rget vs' rroot |-> x ** top) /\
    regPres (fun r => presReg r /\ r <> v0) s s'.*)

  Definition read1Spec := bspeci (readPre 0) (readGuar 0).
  Definition read2Spec := bspeci (readPre 4) (readGuar 4).
    

  (* specification for the write barrier *)
  Definition writePre k s :=
    exists vs,
      s |||- repr vs s (rroot::a0::nil) ** top /\
      vs |||- k + rget vs rroot |->_ ** top.

  (* performing the view update before the "gc step" ensures that
     we use the reachability of the updated view *)
  Definition writeGuar k s s' :=
    (forall vs A rr,
      s |||- repr vs s rr ** A ->
      incl (rroot::a0::nil) rr ->
      exists vs',
	gcGuar (uhset vs (k + rget vs rroot) (rget vs a0)) vs' rr /\
	s' |||- repr vs' s' rr ** A) /\
    regPres presReg s s'.
  (* this is a little broken: it should look like readGuar, with respect
     to fixing rr, if we are to support real write barriers. *)

  Definition write1Spec := bspeci (writePre 0) (writeGuar 0).
  Definition write2Spec := bspeci (writePre 4) (writeGuar 4).


  (* spec for the allocator *)
  Definition allocPre s := 
    exists vs, s |||- repr vs s (rroot::nil) ** top.

  Definition allocGuar s s' :=
    (forall vs A,
      s |||- repr vs s (rroot::nil) ** A ->
      exists vs', exists x,
	gcGuar vs vs' (rroot::nil) /\
	let vs'' := rset (uhset (uhset vs' x NULL) (4 + x) NULL) v0 x in
	s' |||- repr vs'' s' (v0::rroot::nil) ** A /\
	forall B, vs' |||- B -> vs'' |||- B ** x |->, (NULL::NULL::nil)) /\
    regPres presReg s s'.

  Definition allocSpec := bspeci allocPre allocGuar.

  (* this is wrong because it doesn't handle all the possible cases of overlap*)
  (*Definition allocGuarWrong s s' :=
    (forall vs A,
      s |||- repr vs s (rroot::nil) ** A ->
      exists vs', exists objs, exists x,
	gcGuar vs vs' (rroot::nil) /\
	vs' |||- (foralls z in objs, z |->__) /\
	~set_In x objs /\
	s' |||- repr (rset (uhset (uhset vs' x NULL) (4 + x) NULL) v0 x) s' 
	  (v0::rroot::nil) ** A) /\
    regPres presReg s s'.*)


End GCSpecs.


Notation specWeaker :=
  (fun (sp sp':bspec) =>
    let (p, g) := sp in
    let (p', g') := sp' in
      (forall s, p s -> p' s) /\ 
      (forall s s', g' s s' -> g s s')).


Module Type VIEW_GC.

  Declare Module I : MG_INTERFACE_DEF.

  Parameter repr : state -> rfile -> list rnum -> hpred.
  (* repr has to be strong enough to imply the root is pointing at
     an object if it is a valid pointer in hv.  I think that is reasonable. *)


  (* atomic values are represented as themselves *)
  Parameter atomRegEq : forall h vs rf rr A r,
    h ||- repr vs rf rr ** A ->  (* drop the ** A to be consistent with setNonRootOk *)
    In r rr ->
    I.atomP (RFile.get rf r) \/ I.atomP (rget vs r) ->
    RFile.get rf r = rget vs r.

  (* non root registers can be changed without affecting repr *)
  Parameter setNonRootOk : forall h vs rf rr r v,
    h ||- repr vs rf rr ->
    ~In r (I.rgcInfo::rr) ->
    h ||- repr vs (RFile.set rf r v) rr.

  (* root registers can be copied *)
  Parameter rootCopyOk : forall h vs rf rr rs rd,
    h ||- repr vs rf rr ->
    In rs rr ->
    ~In rd (zero::I.rgcInfo::nil) ->
    h ||- repr (rset vs rd (rget vs rs)) (RFile.set rf rd (RFile.get rf rs)) (rd::rr).

  (* root registers can be dropped *)
  Parameter rootWeaken : forall h vs rf rr rr',
    h ||- repr vs rf rr ->
    incl rr' rr ->
    h ||- repr vs rf rr'.

  (* an atomic register can be treated as a root register *)
  Parameter addAtomRoot : forall h vs rf rr A r,
    h ||- repr vs rf rr ** A ->  (* drop the ** A to be consistent with setNonRootOk *)
    ~In r (zero::I.rgcInfo::rr) ->
    I.atomP (RFile.get rf r) ->
    h ||- repr (rset vs r (RFile.get rf r)) rf (r::rr) ** A.


  Parameter phi : chspec.

  Parameter gcPhiDomOk : forall f,
    phi f <> NoneT _ -> set_In f I.gcPhiDom.

  Parameter read1PhiOk :
    optPredT (phi I.read1Lbl)
      (specWeaker (read1Spec repr I.rroot I.gcGuar I.presReg)).
  Parameter read2PhiOk :
    optPredT (phi I.read2Lbl)
      (specWeaker (read2Spec repr I.rroot I.gcGuar I.presReg)).
  Parameter write1PhiOk : 
    optPredT (phi I.write1Lbl)
      (specWeaker (write1Spec repr I.rroot I.gcGuar I.presReg)).
  Parameter write2PhiOk : 
    optPredT (phi I.write2Lbl)
      (specWeaker (write2Spec repr I.rroot I.gcGuar I.presReg)).
  Parameter allocPhiOk : 
    optPredT (phi I.allocLbl)
      (specWeaker (allocSpec repr I.rroot I.gcGuar I.presReg I.NULL)).

  Parameter GCC : cdHeap.
  Parameter GCCok : cdHeapOkSC phi GCC phi.
  
End VIEW_GC.


Module Type MG_SUB_INTERFACE.

  Declare Module I : MG_INTERFACE_DEF.
  Declare Module I' : MG_INTERFACE_DEF.

  Parameter rootEq : I.rroot = I'.rroot.

  Parameter gcInfoEq : I.rgcInfo = I'.rgcInfo.

  Parameter nullEq : I.NULL = I'.NULL.

  Parameter gcGuarSub : forall s s' rr,
    I.gcGuar s s' rr -> I'.gcGuar s s' rr.

  Parameter presRegSub : forall r,
    I'.presReg r -> I.presReg r.

  Parameter atomPSub : forall n,
    I'.atomP n -> I.atomP n.

  Parameter gcPhiDomSub : set_subset I.gcPhiDom I'.gcPhiDom.

  Parameter allocLblEq : I.allocLbl = I'.allocLbl.
  Parameter read1LblEq : I.read1Lbl = I'.read1Lbl.
  Parameter read2LblEq : I.read2Lbl = I'.read2Lbl.
  Parameter write1LblEq : I.write1Lbl = I'.write1Lbl.
  Parameter write2LblEq : I.write2Lbl = I'.write2Lbl.


End MG_SUB_INTERFACE.


Module CoerceGC 
  (G : VIEW_GC)
  (S : MG_SUB_INTERFACE with Module I := G.I) (*<: VIEW_GC with Module I := S.I'*) .

  Module I := S.I'.

  Definition repr := G.repr.

  Lemma atomRegEq : forall h vs rf rr A r,
    h ||- repr vs rf rr ** A ->
    In r rr ->
    S.I'.atomP (RFile.get rf r) \/ S.I'.atomP (rget vs r) ->
    RFile.get rf r = rget vs r.
  Proof.
    intros; eapply G.atomRegEq; eauto.
    destruct H1; [left | right]; apply S.atomPSub; assumption.
  Qed.

  Lemma setNonRootOk : forall h vs rf rr r v,
    h ||- repr vs rf rr ->
    ~In r (S.I'.rgcInfo::rr) ->
    h ||- repr vs (RFile.set rf r v) rr.
  Proof.
    intros.
    rewrite <- S.gcInfoEq in H0.
    apply G.setNonRootOk; assumption.
  Qed.

  Lemma rootCopyOk : forall h vs rf rr rs rd,
    h ||- repr vs rf rr ->
    In rs rr ->
    ~In rd (zero::S.I'.rgcInfo::nil) ->
    h ||- repr (rset vs rd (rget vs rs)) (RFile.set rf rd (RFile.get rf rs)) (rd::rr).
  Proof.
    intros.
    rewrite <- S.gcInfoEq in H1.
    apply G.rootCopyOk; assumption.
  Qed.

  Definition rootWeaken := G.rootWeaken.

  Lemma addAtomRoot : forall h vs rf rr A r,
    h ||- repr vs rf rr ** A ->
    ~In r (zero::S.I'.rgcInfo::rr) ->
    I.atomP (RFile.get rf r) ->
    h ||- repr (rset vs r (RFile.get rf r)) rf (r::rr) ** A.
  Proof.
    intros.
    rewrite <- S.gcInfoEq in H0.
    eapply G.addAtomRoot; eauto.
    apply S.atomPSub; assumption.
  Qed.

  Definition phi := G.phi.

  Lemma gcPhiDomOk : forall f,
    phi f <> NoneT _ -> set_In f I.gcPhiDom.
  Proof.
    intros f Ne; apply S.gcPhiDomSub; apply G.gcPhiDomOk; assumption.
  Qed.


  Lemma read1PhiOk :
    optPredT (G.phi S.I'.read1Lbl)
      (specWeaker (read1Spec repr S.I'.rroot S.I'.gcGuar S.I'.presReg)).
  Proof.
    cbv beta.
    assert (X := G.read1PhiOk); cbv beta in X.
    rewrite <- S.read1LblEq.
    replace S.I.read1Lbl with G.I.read1Lbl; [idtac | reflexivity].
    destruct (G.phi G.I.read1Lbl).
    destruct b as [p g].
    simpl; simpl in X.
    destruct X as [X Y]; constructor.

    intros s Pok; apply (X s); clear X Y.
    rewrite <- S.rootEq in Pok.
    destruct Pok as [vs [Hp Vhp]].
    exists vs; constructor; assumption.

    intros s s' G; assert (Z := Y _ _ G); clear X Y G.
    rewrite <- S.rootEq.
    destruct Z as [X Y]; constructor.
    intros vs A rr Hp I NI; hypReplace X (X vs A rr Hp I NI).
    destruct X as [vs' [G [Hp' Vhp]]]; exists vs'; constructor.
    apply S.gcGuarSub; assumption.
    constructor; assumption.

    intros r [A B]; apply Y; constructor.
    apply S.presRegSub; assumption.
    assumption.

    assumption.

  Qed.


  Lemma read2PhiOk :
    optPredT (G.phi S.I'.read2Lbl)
      (specWeaker (read2Spec repr S.I'.rroot S.I'.gcGuar S.I'.presReg)).
  Proof.
    cbv beta.
    assert (X := G.read2PhiOk); cbv beta in X.
    rewrite <- S.read2LblEq.
    replace S.I.read2Lbl with G.I.read2Lbl; [idtac | reflexivity].
    destruct (G.phi G.I.read2Lbl).
    destruct b as [p g].
    simpl; simpl in X.
    destruct X as [X Y]; constructor.

    intros s Pok; apply (X s); clear X Y.
    rewrite <- S.rootEq in Pok.
    destruct Pok as [vs [Hp Vhp]].
    exists vs; constructor; assumption.

    intros s s' G; assert (Z := Y _ _ G); clear X Y G.
    rewrite <- S.rootEq.
    destruct Z as [X Y]; constructor.
    intros vs A rr Hp I NI; hypReplace X (X vs A rr Hp I NI).
    destruct X as [vs' [G [Hp' Vhp]]]; exists vs'; constructor.
    apply S.gcGuarSub; assumption.
    constructor; assumption.

    intros r [A B]; apply Y; constructor.
    apply S.presRegSub; assumption.
    assumption.

    assumption.

  Qed.


  Lemma write1PhiOk :
    optPredT (G.phi S.I'.write1Lbl)
      (specWeaker (write1Spec repr S.I'.rroot S.I'.gcGuar S.I'.presReg)).
  Proof.
    cbv beta.
    assert (X := G.write1PhiOk); cbv beta in X.
    rewrite <- S.write1LblEq.
    replace S.I.write1Lbl with G.I.write1Lbl; [idtac | reflexivity].
    destruct (G.phi G.I.write1Lbl).
    destruct b as [p g].
    simpl; simpl in X.
    destruct X as [X Y]; constructor.

    intros s Pok; apply (X s); clear X Y.
    rewrite <- S.rootEq in Pok.
    destruct Pok as [vs [Hp Vhp]].
    exists vs; constructor; assumption.

    intros s s' G; assert (Z := Y _ _ G); clear X Y G.
    rewrite <- S.rootEq.
    destruct Z as [X Y]; constructor.
    intros vs A rr Hp I; hypReplace X (X vs A rr Hp I).
    destruct X as [vs' [G Hp']]; exists vs'; constructor.
    apply S.gcGuarSub; assumption.
    assumption.

    intros r Z; apply Y; apply S.presRegSub; assumption.

    assumption.

  Qed.


  Lemma write2PhiOk :
    optPredT (G.phi S.I'.write2Lbl)
      (specWeaker (write2Spec repr S.I'.rroot S.I'.gcGuar S.I'.presReg)).
  Proof.
    cbv beta.
    assert (X := G.write2PhiOk); cbv beta in X.
    rewrite <- S.write2LblEq.
    replace S.I.write2Lbl with G.I.write2Lbl; [idtac | reflexivity].
    destruct (G.phi G.I.write2Lbl).
    destruct b as [p g].
    simpl; simpl in X.
    destruct X as [X Y]; constructor.

    intros s Pok; apply (X s); clear X Y.
    rewrite <- S.rootEq in Pok.
    destruct Pok as [vs [Hp Vhp]].
    exists vs; constructor; assumption.

    intros s s' G; assert (Z := Y _ _ G); clear X Y G.
    rewrite <- S.rootEq.
    destruct Z as [X Y]; constructor.
    intros vs A rr Hp I; hypReplace X (X vs A rr Hp I).
    destruct X as [vs' [G Hp']]; exists vs'; constructor.
    apply S.gcGuarSub; assumption.
    assumption.

    intros r Z; apply Y; apply S.presRegSub; assumption.

    assumption.

  Qed.


  Lemma allocPhiOk :
    optPredT (G.phi S.I'.allocLbl)
      (specWeaker (allocSpec repr S.I'.rroot S.I'.gcGuar S.I'.presReg S.I'.NULL)).
  Proof.
    cbv beta.
    assert (X := G.allocPhiOk); cbv beta in X.
    rewrite <- S.allocLblEq.
    replace S.I.allocLbl with G.I.allocLbl; [idtac | reflexivity].
    destruct (G.phi G.I.allocLbl).
    destruct b as [p g].
    simpl; simpl in X.
    destruct X as [X Y]; constructor.

    intros s Pok; apply (X s); clear X Y.
    rewrite <- S.rootEq in Pok.
    destruct Pok as [vs Hp].
    exists vs; assumption.

    intros s s' G; assert (Z := Y _ _ G); clear X Y G.
    rewrite <- S.rootEq.
    destruct Z as [X Y]; constructor.
    intros vs A Hp; hypReplace X (X vs A Hp).
    destruct X as [vs' [x [G [Vhp New]]]];
      exists vs'; exists x; constructor.
    apply S.gcGuarSub; assumption.
    rewrite <- S.nullEq.
    intuition.

    intros r Z; apply Y; apply S.presRegSub; assumption.

    assumption.

  Qed.


  Definition GCC := G.GCC.

  Definition GCCok := G.GCCok.


End CoerceGC.



Module Type BASIC_MGI_ARG.
  Parameter gcGuar : state -> state -> list rnum -> Prop.
End BASIC_MGI_ARG.


(* A basic MGI that defines the simple parts *)

Module BasicMGI (BMA : BASIC_MGI_ARG) <: MG_INTERFACE_DEF.

  Definition rroot := t0.
  Definition rgcInfo := s7. (* was : t1 *)
  Definition gcGuar := BMA.gcGuar.
  Definition presReg := calleeSaved.
  Definition atomP x := atomP x.
  Definition gcPhiDom := setRange 10 200.
  Definition allocLbl := 10.
  Definition read1Lbl := 20.
  Definition read2Lbl := 30.
  Definition write1Lbl := 40.
  Definition write2Lbl := 50.
  Definition NULL := 1.

End BasicMGI.


(*
Notation CHENEYREAD1 := 10.
Notation CHENEYREAD2 := 12.
Notation CHENEYWRITE1 := 14.
Notation CHENEYWRITE2 := 16.
Definition CHENEYALLOC := 18.
*)





Module BasicGCGuar <: BASIC_MGI_ARG.
  Definition gcGuar (s s':state) (rr:list rnum) := 
    (forall objs h,
      s |||- (minObjHp (List.map (rget s) rr) objs && Heap.heap_eqv h) ** top ->
      s' |||- Heap.heap_eqv h ** top) /\
    forall r, In r rr -> rget s r = rget s' r.

End BasicGCGuar.

Module BasicGCMGI := BasicMGI (BasicGCGuar).


Lemma basicGcGuarRefl : forall vs rr,
  BasicGCGuar.gcGuar vs vs rr.
Proof.
  intros vs rr.
  constructor.
  intros objs h HpOk.
  destruct (aconjElim (sconjDistrib2 HpOk)) as [_ X].
  assumption.
  reflexivity.
Qed.
